# -*- coding: utf-8 -*-

import os
import argparse
import psutil

import pickle

from experimentlogger.experiment_logger import ExperimentLogger
from prolothar_common.experiments.stopwatch import Stopwatch

from prolothar_rule_mining.rule_miner.dataset_splitting_miner import DatasetSplittingMiner
from prolothar_rule_mining.rule_miner.multilabel.boomer import Boomer
from prolothar_rule_mining.rule_miner.multilabel.boomer.pruning import NoPruning
from prolothar_rule_mining.rule_miner.multilabel.boomer.pruning import NaiveF1ScorePruning
from prolothar_rule_mining.rule_miner.data_to_sequence.multiset_orderer import MultiLabelSequencePredictor
from prolothar_rule_mining.rule_miner.data_to_sequence.multiset_orderer import LpOrdering

from experiments.rule_mining.dataset_loader import DatasetLoader
from experiments.rule_mining.noise_generator import NoiseGenerator
from experiments.rule_mining.experiment_utils import plot_and_save_true_model

class BoomerRunner:

    def __init__(self):
        self.__logger: ExperimentLogger = None
        self.__dataset_loader = DatasetLoader()
        self.__noise_generator = NoiseGenerator()

    def run_from_cli(self):
        args = self.__parse_cli_arguments()

        self.__logger = ExperimentLogger(
            os.path.join('experiments', 'rule_mining', 'boomer', 'results'),
            args.dataset)
        self.__logger.log(str(args))

        dataset, true_model = self.__dataset_loader.read_dataset(args, 0)
        dataset,_ = dataset.split(0.2)
        self.__noise_generator.apply(dataset, args)
        self.__logger.log('dataset contains %d instances' % len(dataset))

        if true_model is not None:
            plot_and_save_true_model(true_model, dataset, self.__logger)

        rule_miner = MultiLabelSequencePredictor(
            DatasetSplittingMiner(
                lambda x_train, x_val: Boomer(
                    pruning=self.__create_pruning(args, x_val),
                    nr_of_threads=args.nr_of_cpus).mine_rules(x_train),
                hold_out_set_ratio=args.valset_proportion,
                random_seed=args.random_seed),
            LpOrdering()
        )

        stopwatch = Stopwatch()
        stopwatch.start()

        rules = rule_miner.mine_rules(dataset)

        self.__logger.log('')
        self.__logger.log('elapsed time: %r' % stopwatch.get_elapsed_time())

        with self.__logger.open_binary_file('multilabel_rules.pickle') as f:
            pickle.dump(rules.get_multilabel_rule(), f)

        with self.__logger.open_text_file(f'rules.txt') as f:
            for rule in rules.get_multilabel_rule():
                f.write(f'{rule}\n')

    def __create_pruning(self, args, valset):
        if args.pruning == 'none':
            return NoPruning()
        elif args.pruning == 'naivef1':
            return NaiveF1ScorePruning(valset)

    def __parse_cli_arguments(self):
        parser = argparse.ArgumentParser('Boomer rule mining on a single dataset')

        parser.add_argument(
                '--trainset_size', type=int,
                help=('nr of instances in the trainset. default means maximal size'),
                required=False, default=-1)

        parser.add_argument(
                '--valset_proportion', type=float,
                help='proportion in (0,1) of the validation set',
                required=False, default=0.2)

        parser.add_argument(
                '--pruning', type=str,
                help='pruning strategy',
                required=False, default='none', choices=['none', 'naivef1'])

        parser.add_argument(
                '--nr_of_cpus', type=int,
                help=('number of cpus to use for parallel computations if '
                      'applicable. Default is to use all available cpus'),
                required=False, default=len(psutil.Process().cpu_affinity()))

        self.__dataset_loader.add_cli_arguments(parser)
        self.__noise_generator.add_cli_arguments(parser)

        return parser.parse_args()

if __name__ == '__main__':
    BoomerRunner().run_from_cli()
