# -*- coding: utf-8 -*-

import os
import argparse
import psutil

import pickle

from experimentlogger.experiment_logger import ExperimentLogger
from prolothar_common.experiments.stopwatch import Stopwatch

from prolothar_rule_mining.rule_miner.classification.myra import CAntMinerPB
from prolothar_rule_mining.rule_miner.classification.rce import ReliableRuleMiner
from prolothar_rule_mining.rule_miner.classification.rce.strategy.beam_search import BeamSearch as ConditionBeamSearch
from prolothar_rule_mining.rule_miner.classification.classy import Classy
from prolothar_rule_mining.rule_miner.classification.fdd_rule_miner import FddRuleMiner
from prolothar_rule_mining.rule_miner.data_to_sequence.event_classifier import EventClassifier

from experiments.rule_mining.dataset_loader import DatasetLoader
from experiments.rule_mining.noise_generator import NoiseGenerator
from experiments.rule_mining.experiment_utils import plot_and_save_true_model
from experiments.rule_mining.experiment_utils import create_computation_engine

class ClsRunner:

    def __init__(self):
        self.__logger: ExperimentLogger = None
        self.__dataset_loader = DatasetLoader()
        self.__noise_generator = NoiseGenerator()

    def run_from_cli(self):
        args = self.__parse_cli_arguments()

        self.__logger = ExperimentLogger(
            os.path.join('experiments', 'rule_mining', 'cls', 'results'),
            args.dataset)
        self.__logger.log(str(args))

        dataset, true_model = self.__dataset_loader.read_dataset(args, 0)
        trainset,testset = dataset.split(
            args.testset_proportion, random_seed=args.random_seed)
        if args.trainset_size > 0:
            trainset = trainset.random_subset(args.trainset_size,
                                              random_seed=args.random_seed)
        self.__noise_generator.apply(trainset, args)
        self.__logger.log('dataset contains %d instances' % len(dataset))
        self.__logger.log('trainset contains %d instances' % len(trainset))
        self.__logger.log('testset contains %d instances' % len(testset))

        if true_model is not None:
            plot_and_save_true_model(true_model, trainset, self.__logger)

        event_classifier = EventClassifier(
            classification_miner=self.__create_classification_miner(args),
            computation_engine=create_computation_engine(args)
        )

        stopwatch = Stopwatch()
        stopwatch.start()

        rules = event_classifier.mine_rules(trainset)

        self.__logger.log('')
        self.__logger.log('elapsed time: %r' % stopwatch.get_elapsed_time())

        with self.__logger.open_binary_file('mined_model.pickle') as f:
            pickle.dump(rules, f)

        for i,rule in enumerate(rules.get_classifier_list()):
            with self.__logger.open_text_file(f'rule_{i}.txt') as f:
                f.write(str(rule))

    def __create_classification_miner(self, args):
        classifier = {
            'Classy': Classy(),
            'CAntMinerPB': CAntMinerPB(nr_of_workers=1),
            'ReliableRuleMiner': ReliableRuleMiner(
                logger=None if args.nr_of_cpus > 1 else self.__logger.log,
                search_strategy=ConditionBeamSearch(
                    max_nr_of_base_candidates=args.max_nr_of_base_conditions,
                    beam_width=args.beam_width_condition_search))
        }[args.classifier]

        if args.feature_selector == 'FDD':
            classifier = FddRuleMiner(classifier)
        elif args.feature_selector is not None:
            raise NotImplementedError(args.feature_selector)

        return classifier

    def __parse_cli_arguments(self):
        parser = argparse.ArgumentParser('EventClassifier rule mining on a single dataset')

        parser.add_argument(
                '--trainset_size', type=int,
                help=('nr of instances in the trainset. default means maximal size'),
                required=False, default=-1)

        parser.add_argument(
                '--testset_proportion', type=float,
                help='proportion in (0,1) of the testset',
                required=False, default=0.2)

        parser.add_argument(
                '--classifier', type=str,
                help=('classifier algorithm used at split points'),
                required=False, default='ReliableRuleMiner',
                choices=['ReliableRuleMiner', 'Classy', 'CAntMinerPB'])

        parser.add_argument(
                '--max_nr_of_base_conditions', type=int,
                help=('if classifier is ReliableRuleMiner, the number of base '
                      'candidates to keep for extension during search. '
                      'default means no limitation. the limitation is useful '
                      'if some categorical attributes have many different values '
                      'such as an ID.'),
                required=False, default=-1)

        parser.add_argument(
                '--beam_width_condition_search', type=int,
                help=('if classifier is ReliableRuleMiner, the width parameter '
                      'for the beam search.'),
                required=False, default=5)

        parser.add_argument(
                '--feature_selector', type=str,
                help=('feature selector run before the classifier at split points'),
                required=False, default=None,
                choices=['FDD'])

        parser.add_argument(
                '--nr_of_cpus', type=int,
                help=('number of cpus to use for parallel computations if '
                      'applicable. Default is to use all available cpus'),
                required=False, default=len(psutil.Process().cpu_affinity()))

        parser.add_argument(
                '--parallel_engine', type=str,
                help='name of the computation engine for parallel computations',
                required=False, default='multiprocessing',
                choices=['ray', 'multiprocessing'])

        self.__dataset_loader.add_cli_arguments(parser)
        self.__noise_generator.add_cli_arguments(parser)

        return parser.parse_args()

if __name__ == '__main__':
    ClsRunner().run_from_cli()
