# -*- coding: utf-8 -*-

from typing import Union, List, Tuple
from abc import ABC, abstractmethod
import os
import itertools
import argparse
from collections import Counter
import psutil
import numpy as np
from runstats import Statistics

from experimentlogger.experiment_logger import ExperimentLogger
from prolothar_common.experiments.result_collector import ResultCollector
from prolothar_common.experiments.stopwatch import Stopwatch
import matplotlib.pyplot as plt
from strsimpy.levenshtein import Levenshtein
from statistics import harmonic_mean

from prolothar_common.models.dataset import TargetSequenceDataset
from prolothar_rule_mining.models.event_flow_graph.cover import CoverComputer

from prolothar_rule_mining.rule_miner.data_to_sequence.rules import Rule
from prolothar_rule_mining.rule_miner.data_to_sequence.rules.event_flow_graph_rule import EventFlowGraphRule

from experiments.rule_mining.dataset_loader import DatasetLoader
from experiments.rule_mining.noise_generator import NoiseGenerator
from experiments.rule_mining.comparison.method_list import MethodList, alignment_finder_factory_score

METHOD_NAME_BEST_EXPECTED = 'Best expected'

class RuleMiningExperiment(ABC):

    def __init__(self, x_name: str):
        self.__result_collector = ResultCollector(x_name)
        self.__stopwatch = Stopwatch()
        self._logger: ExperimentLogger = None
        self.__dataset_loader = DatasetLoader()
        self.__noise_generator = NoiseGenerator()

    def run_from_cli(self):
        args = self.__parse_cli_arguments()

        self._logger = ExperimentLogger(
            'experiments/rule_mining/comparison/results', 'comparison %s' % args.dataset)
        os.mkdir(os.path.join(
            self._logger.get_path_to_experiment_directory(), 'datasets'))

        self._logger.log(str(args))

        self._run_experiments(args)

        self.__result_collector.save_plots(
            self._logger.get_path_to_experiment_directory())

    @abstractmethod
    def _run_experiments(self, args):
        pass

    def _run_experiment_for_n_instances(self, n: int, x: Union[float, int], args):
        self._logger.log('')
        if x != n:
            self._logger.log(f'x: {x}')
        self._logger.log('size of trainset: %d' % n)

        for i in range(args.nr_of_repetitions):
            self._logger.log('repetition %d' % i)
            dataset,true_model = self.__dataset_loader.read_dataset(args, i)
            trainset,testset = dataset.split(
                args.testset_proportion, random_seed=args.random_seed)
            self._apply_noise(self.__noise_generator, trainset, args, x)
            if i == 0:
                self._logger.log('size of whole dataset: %d' % len(dataset))
                self._logger.log('size of alphabet: %d' % len(dataset.get_set_of_sequence_symbols()))
                self._logger.log('avg. sequence length: %r' % np.mean([len(i.get_target_sequence()) for i in dataset]))
                self._logger.log('number of attributes: %d' % dataset.get_nr_of_attributes())
                self._logger.log('unique sequences: %d' % len(dataset.get_sequences_ordered_by_frequency()))
                self._logger.log('size of testset: %d' % len(testset))
            trainset = trainset.random_subset(n, random_seed=args.random_seed)
            self.__plot_data_set_statistics(dataset, i)
            for miner_name, miner, color, marker in MethodList().get_method_list(args, trainset, true_model):
                if miner_name in args.miners:
                    self.__run_and_evaluate_miner(
                        miner, miner_name, trainset, testset, x)
                    self.__result_collector.set_color(miner_name, color)
                    self.__result_collector.set_marker(miner_name, marker)
            if true_model is not None:
                self._logger.log('TrueModel')
                self._evaluate_rules(
                    'TrueModel', true_model, trainset, testset, x)
                self.__result_collector.set_color('TrueModel', ResultCollector.colors.GOLD)
                self.__result_collector.set_marker('TrueModel', ResultCollector.markers.POINT)
            self.__add_best_expected_performance(trainset, testset, x)

    def __parse_cli_arguments(self):
        parser = argparse.ArgumentParser(
            'Data2Seq rule mining experiments')

        parser.add_argument(
                '--miners', type=str, nargs='+',
                help='a list of rule miners for comparison',
                required=True)

        parser.add_argument(
                '--testset_proportion', type=float,
                help='proportion in (0,1) of the testset',
                required=False, default=0.2)

        parser.add_argument(
                '--max_java_heapspace_in_mb', type=int,
                help='maximum size of java heap space (e.g. for CAntMinerPB)',
                required=False, default=-1)

        parser.add_argument(
                '--nr_of_cpus', type=int,
                help=('number of cpus to use for parallel computations if '
                      'applicable. Default is to use all available cpus'),
                required=False, default=len(psutil.Process().cpu_affinity()))

        parser.add_argument(
                '--parallel_engine', type=str,
                help=('computation engine for parallel computation if applicable'),
                required=False, default='ray', choices=['multiprocessing', 'ray'])

        self._add_cli_arguments(parser)

        self.__dataset_loader.add_cli_arguments(parser)
        self.__noise_generator.add_cli_arguments(parser, nargs=self._nargs_for_noise_levels())

        return parser.parse_args()

    @abstractmethod
    def _add_cli_arguments(self, parser: argparse.ArgumentParser):
        pass

    @abstractmethod
    def _nargs_for_noise_levels(self) -> Union[str, None]:
        pass

    @abstractmethod
    def _apply_noise(
            self, noise_generator: NoiseGenerator, trainset: TargetSequenceDataset,
            args, x):
        pass

    def __add_best_expected_performance(
            self, trainset: TargetSequenceDataset, testset: TargetSequenceDataset,
            x: Union[float, int]):
        unique_sequences_in_trainset = set(
            instance.get_target_sequence() for instance in trainset)

        nr_of_known_sequences = 0
        for instance in testset:
            if instance.get_target_sequence() in unique_sequences_in_trainset:
                nr_of_known_sequences += 1

        accuracy = nr_of_known_sequences / len(testset)

        self.__result_collector.add_result(
            METHOD_NAME_BEST_EXPECTED, 'Testset Accuracy on class', x, accuracy)
        self.__result_collector.set_color(METHOD_NAME_BEST_EXPECTED, ResultCollector.colors.BLACK)
        self.__result_collector.set_marker(METHOD_NAME_BEST_EXPECTED, ResultCollector.markers.TRIANGLE_DOWN)

    def __run_and_evaluate_miner(
            self, miner, miner_name: str, trainset: TargetSequenceDataset,
            testset: TargetSequenceDataset, x: Union[float, int]):
        try:
            self._logger.log(miner_name)
            self.__stopwatch.start()
            self._evaluate_rules(
                miner_name, miner.mine_rules(trainset),
                trainset, testset, x)
        except Exception as e:
            self._logger.log('Ignore run of %s because of %r' % (miner_name, e))

    def _evaluate_rules(
            self, miner_name: str, rules: Rule,
            trainset: TargetSequenceDataset, testset: TargetSequenceDataset,
            x: Union[float, int]):
        if miner_name not in ['TrueModel', 'RandomClassifier']:
            self.__result_collector.add_result(
                miner_name, 'Time in seconds', x,
                self.__stopwatch.get_elapsed_time().total_seconds())

        if miner_name != 'RandomClassifier':
            self.__evaluate_model_complexity(
                miner_name, rules, x, trainset, testset)

        self.__evaluate_on_subdataset(
            miner_name, rules, trainset, 'Trainset', x)
        self.__evaluate_on_subdataset(
            miner_name, rules, testset, 'Testset', x)

    def __evaluate_model_complexity(
            self, miner_name: str, rules: Rule, x: Union[int, float],
            trainset: TargetSequenceDataset, testset: TargetSequenceDataset):
        try:
            nr_of_terms = rules.count_nr_of_terms()
            self._logger.log('number of terms: %d' % nr_of_terms)
            self.__result_collector.add_result(
                miner_name, 'Number of terms', x, nr_of_terms)
        except (AttributeError, NotImplementedError) as e:
            pass
        try:
            if isinstance(rules, EventFlowGraphRule):
                eventflow_graph = rules.get_event_flow_graph()
            elif isinstance(rules, Rule):
                eventflow_graph = rules.to_eventflow_graph()
            else:
                raise NotImplementedError()
            nr_of_decision_points = sum(
                len(node.children) - 1 for node in itertools.chain(
                    eventflow_graph.nodes(), [eventflow_graph.source]))
            self._logger.log('number of decision branches: %d' % nr_of_decision_points)
            self.__result_collector.add_result(
                miner_name, 'Number of decision branches', x, nr_of_decision_points)

            mdl_of_model = eventflow_graph.compute_mdl(testset.get_set_of_sequence_symbols())
            cover_computer = CoverComputer(
                eventflow_graph, alignment_finder=alignment_finder_factory_score(eventflow_graph))
            mdl_data_trainset = cover_computer.compute_cover(trainset).compute_mdl()
            mdl_data_testset = cover_computer.compute_cover(testset).compute_mdl()
            self.__result_collector.add_result(
                miner_name, 'EventFlowGraph MDL Model', x, mdl_of_model)
            self.__result_collector.add_result(
                miner_name, 'EventFlowGraph MDL Data Trainset',
                x, mdl_data_trainset)
            self.__result_collector.add_result(
                miner_name, 'EventFlowGraph MDL Data Testset',
                x, mdl_data_testset)
            self.__result_collector.add_result(
                miner_name, 'EventFlowGraph MDL Total Trainset',
                x, mdl_of_model + mdl_data_trainset)
            self.__result_collector.add_result(
                miner_name, 'EventFlowGraph MDL Total Testset',
                x, mdl_of_model + mdl_data_testset)
        except (AttributeError, NotImplementedError) as e:
            pass
        try:
            self.__result_collector.add_result(
                miner_name, 'Number of nodes',
                x, rules.get_nr_of_nodes())
        except AttributeError:
            pass

    def __evaluate_on_subdataset(
            self, miner_name: str, rules: Rule,
            subdataset: TargetSequenceDataset, subdataset_name: str,
            x: Union[float, int]):
        predictions = [tuple(rules.execute(instance)) for instance in subdataset]

        self.__evaluate_sequence_prediction_metrics(
            miner_name, predictions, subdataset, subdataset_name, x)

        self.__evaluate_order_invariant_multi_label_classification_metrics(
            miner_name, predictions, subdataset, subdataset_name, x)

        self.__evaluate_order_invariant_multiset_metrics(
            miner_name, predictions, subdataset, subdataset_name, x)

    def __evaluate_sequence_prediction_metrics(
            self, miner_name: str, predictions: List[Tuple[str]], subdataset: TargetSequenceDataset,
            subdataset_name: str, x: Union[int, float]):
        completely_correct_sequences = 0
        levenshtein_distance = 0
        normalized_levenshtein_similarity = 0
        for predicted_sequence, instance in zip(predictions, subdataset):
            if predicted_sequence == instance.get_target_sequence():
                completely_correct_sequences += 1
            single_instance_levenshtein_distance = Levenshtein().distance(
                predicted_sequence, instance.get_target_sequence())
            levenshtein_distance += single_instance_levenshtein_distance
            try:
                normalized_levenshtein_similarity += 1 - (single_instance_levenshtein_distance / max(
                    len(predicted_sequence), len(instance.get_target_sequence())))
            except ZeroDivisionError:
                #both predicted and actual sequence are empty => similarity = 1
                normalized_levenshtein_similarity += 1

        accuracy = completely_correct_sequences / len(subdataset)
        normalized_levenshtein_similarity /= len(subdataset)

        self._logger.log('%s accuracy on class: %.2f%%' % (
            subdataset_name, accuracy * 100))
        self.__result_collector.add_result(
            miner_name, '%s Accuracy on class' % subdataset_name,
            x, accuracy)

        self._logger.log('%s levenshtein distance: %d' % (
            subdataset_name, levenshtein_distance))
        self.__result_collector.add_result(
            miner_name, '%s Levenshtein' % subdataset_name,
            x, levenshtein_distance)
        self.__result_collector.add_result(
            miner_name, '%s Normalized Levenshtein Similarity' % subdataset_name,
            x, normalized_levenshtein_similarity)

    def __evaluate_order_invariant_multi_label_classification_metrics(
            self, miner_name: str, predictions: List[Tuple[str]], subdataset: TargetSequenceDataset,
            subdataset_name: str, x: Union[int, float]):
        true_positives = 0
        false_positives = 0
        false_negatives = 0
        for predicted_sequence, instance in zip(predictions, subdataset):
            predicted_events = set(predicted_sequence)
            actual_events = instance.get_symbols()
            true_positives += len(predicted_events.intersection(actual_events))
            false_positives += len(predicted_events.difference(actual_events))
            false_negatives += len(actual_events.difference(predicted_events))
        try:
            micro_precision = true_positives / (true_positives + false_positives)
        except ZeroDivisionError:
            micro_precision = 1
        if true_positives + false_negatives == 0:
            micro_recall = 1
        else:
            micro_recall = true_positives / (true_positives + false_negatives)
        micro_f1 = harmonic_mean([micro_precision, micro_recall])
        self._logger.log('%s precision on events: %.2f%%' % (
            subdataset_name, micro_precision * 100))
        self.__result_collector.add_result(
            miner_name, '%s Precision on events' % subdataset_name,
            x, micro_precision)
        self._logger.log('%s recall on events: %.2f%%' % (
            subdataset_name, micro_recall * 100))
        self.__result_collector.add_result(
            miner_name, '%s Recall on events' % subdataset_name,
            x, micro_recall)
        self._logger.log('%s F1-score on events: %.2f%%' % (
            subdataset_name, micro_f1 * 100))
        self.__result_collector.add_result(
            miner_name, '%s F1-score on events' % subdataset_name,
            x, micro_f1)

    def __evaluate_order_invariant_multiset_metrics(
            self, miner_name: str, predictions: List[Tuple[str]],
            subdataset: TargetSequenceDataset,
            subdataset_name: str, x: Union[int, float]):
        #https://arxiv.org/pdf/1711.05246.pdf
        #Loss Functions for Multiset Prediction
        multiset_equality = 0
        macro_precision_statistics = Statistics()
        macro_recall_statistics = Statistics()
        for predicted_sequence, instance in zip(predictions, subdataset):
            counter_predicted = Counter(predicted_sequence)
            counter_actual = Counter(instance.get_target_sequence())
            if counter_predicted == counter_actual:
                multiset_equality += 1
            try:
                macro_precision_statistics.push(
                    1 - sum(
                        max(counter_predicted[e] - counter_actual[e], 0)
                        for e in subdataset.get_set_of_sequence_symbols()
                    ) / sum(counter_predicted.values())
                )
            except ZeroDivisionError:
                macro_precision_statistics.push(1)

            try:
                macro_recall_statistics.push(
                    1 - sum(
                        max(counter_actual[e] - counter_predicted[e], 0)
                        for e in subdataset.get_set_of_sequence_symbols()
                    ) / sum(counter_actual.values())
                )
            except ZeroDivisionError:
                macro_recall_statistics.push(1)

        multiset_equality = multiset_equality / len(subdataset)
        macro_precision = macro_precision_statistics.mean()
        macro_recall = macro_recall_statistics.mean()
        macro_f1 = harmonic_mean([macro_precision, macro_recall])
        self.__result_collector.add_result(
            miner_name, '%s Multiset Equality on events' % subdataset_name,
            x, multiset_equality)
        self.__result_collector.add_result(
            miner_name, '%s Multiset Precision on events' % subdataset_name,
            x, macro_precision)
        self.__result_collector.add_result(
            miner_name, '%s Multiset Recall on events' % subdataset_name,
            x, macro_recall)
        self.__result_collector.add_result(
            miner_name, '%s Multiset F1-score on events' % subdataset_name,
            x, macro_f1)

    def __plot_data_set_statistics(self, dataset: TargetSequenceDataset, dataset_nr: int):
        nr_of_unique_sequences_counter = Counter()
        sequence_length_counter = Counter()
        for instance in dataset:
            nr_of_unique_sequences_counter[instance.get_target_sequence()] += 1
            sequence_length_counter[len(instance.get_target_sequence())] += 1

        fig = plt.figure()
        plt.bar(
            list(range(len(nr_of_unique_sequences_counter))),
            list(sorted(nr_of_unique_sequences_counter.values())), 1)
        fig.savefig(os.path.join(
            self._logger.get_path_to_experiment_directory(), 'datasets',
            'count_unique_sequences_%d_%d_sequences.png' % (
                dataset_nr, len(dataset))))
        plt.close(fig)

        fig = plt.figure()
        x, y = zip(*sorted(sequence_length_counter.items()))
        plt.bar(x, y, 1)
        fig.savefig(os.path.join(
            self._logger.get_path_to_experiment_directory(), 'datasets',
            'count_sequence_lengths_%d_%d_sequences.png' % (
                dataset_nr, len(dataset))))
        plt.close(fig)
