from typing import Union, Iterable, Tuple, Any

import itertools
from collections import Counter

from dhparallel.engine.multiprocess.multiprocess import MultiprocessComputationEngine
from dhparallel.engine.ray.ray import RayComputationEngine

from prolothar_common.experiments.result_collector import ResultCollector
from prolothar_common.models.dataset import TargetSequenceDataset
from prolothar_common.models.dataset import MultiLabelDataset
from prolothar_common.models.dataset.instance import MultiLabelInstance
from prolothar_common.models.labeled_transition_system.state_representation import set_abstraction
from prolothar_common.models.labeled_transition_system.state_representation import list_abstraction
from prolothar_common.models.labeled_transition_system.state_representation import last_abstraction
from prolothar_common.models.labeled_transition_system.state_representation import last_n_abstraction
from prolothar_common.func_tools import do_nothing

from prolothar_rule_mining.models.converter.sequence_to_multilabel_dataset_converter import SequenceToMultiLabelDatasetConverter
from prolothar_rule_mining.models.event_flow_graph import EventFlowGraph

from prolothar_rule_mining.rule_miner.dataset_splitting_miner import DatasetSplittingMiner

from prolothar_rule_mining.rule_miner.classification.myra import CAntMinerPB
from prolothar_rule_mining.rule_miner.classification.classy import Classy
from prolothar_rule_mining.rule_miner.classification.knn import KnnClassifier
from prolothar_rule_mining.rule_miner.classification.fdd_rule_miner import FddRuleMiner
from prolothar_rule_mining.rule_miner.classification.rce import ReliableRuleMiner
from prolothar_rule_mining.rule_miner.classification.rce.strategy.beam_search import BeamSearch as ConditionBeamSearch
from prolothar_rule_mining.rule_miner.classification.naive_bayes import CategoricalNaiveBayesClassifier

try:
    from prolothar_rule_mining.rule_miner.multilabel.boomer import Boomer
    from prolothar_rule_mining.rule_miner.multilabel.boomer.pruning import NaiveF1ScorePruning
except ImportError as e:
    print('warning: Boomer is not compiled and will not be available during experiments')
from prolothar_rule_mining.rule_miner.multilabel.mlrbc import Mlrbc

from prolothar_rule_mining.rule_miner.data_to_sequence.rules import Rule
from prolothar_rule_mining.rule_miner.data_to_sequence.rules.append_rule import AppendRule
from prolothar_rule_mining.rule_miner.data_to_sequence.rules.list_of_rules import ListOfRules

from prolothar_rule_mining.rule_miner.data_to_sequence.multiset_orderer import MultiLabelSequencePredictor
from prolothar_rule_mining.rule_miner.data_to_sequence.multiset_orderer import LpOrdering
from prolothar_rule_mining.rule_miner.data_to_sequence.event_classifier import EventClassifier
from prolothar_rule_mining.rule_miner.data_to_sequence.classification_miner import ClassificationMiner
from prolothar_rule_mining.rule_miner.data_to_sequence.data_to_sequence_lstm import DataToSequenceLstm, DataToSequenceAdversarialLstm
from prolothar_rule_mining.rule_miner.data_to_sequence.window_classifier import WindowClassifier
from prolothar_rule_mining.rule_miner.data_to_sequence.dats import Dats
from prolothar_rule_mining.rule_miner.data_to_sequence.rudi.rudi import RuDi
from prolothar_rule_mining.rule_miner.data_to_sequence.rudi.event_flow_graph_miner import SequenceBottomUpEventFlowGraphMiner
from prolothar_rule_mining.rule_miner.data_to_sequence.rudi.event_flow_graph_miner import SequenceTopDownEventFlowGraphMiner
from prolothar_rule_mining.rule_miner.data_to_sequence.rudi.event_flow_graph_miner import SequenceAlignmentNoMdl
from prolothar_rule_mining.rule_miner.data_to_sequence.rudi.event_flow_graph_miner import OracleEventFlowGraphMiner

from prolothar_rule_mining.models.event_flow_graph.alignment.petrinet import PetrinetAligner
from prolothar_rule_mining.models.event_flow_graph.alignment.a_star import AStar
from prolothar_rule_mining.models.event_flow_graph.alignment.heuristics import NullHeuristic
from prolothar_rule_mining.models.event_flow_graph.alignment.beam_search import BeamSearch
from prolothar_rule_mining.models.event_flow_graph.alignment.greedy_beam_search import GreedyBeamSearch
from prolothar_rule_mining.models.event_flow_graph.alignment.heuristics import ReachabilityHeuristic
from prolothar_rule_mining.models.event_flow_graph.router.learning import OracleRouterLearner
from prolothar_rule_mining.models.event_flow_graph.router.learning import RuleClassifierRouterLearner

from experiments.rule_mining.experiment_utils import create_computation_engine

def alignment_finder_factory_score(graph):
    return BeamSearch(graph, ReachabilityHeuristic(graph), 10)

def alignment_finder_factory_model_extension(graph):
    return PetrinetAligner(graph)

def rule_to_eventflow_graph(rule: Rule) -> EventFlowGraph:
    try:
        return rule.get_event_flow_graph()
    except AttributeError:
        return rule.to_eventflow_graph()

class RandomClassifier():
    """
    simulates lower bound performance by predicting the most frequent sequence
    in the training set
    """
    def mine_rules(self, trainset: TargetSequenceDataset) -> Rule:
        counter = Counter()
        for instance in trainset:
            counter[instance.get_target_sequence()] += 1
        rules = ListOfRules()
        for event in max(counter, key=counter.get):
            rules.append_rule(AppendRule(event))
        return rules

    def is_whitebox(self) -> bool:
        return False

    def is_output_order_invariant(self) -> bool:
        return False

class MethodList:

    def get_method_list(
            self, args, trainset: TargetSequenceDataset,
            true_model: Union[None, Rule]) -> Iterable[Tuple[str, Any, str, str]]:
        return itertools.chain(
            self.__create_method_list_for_cross_method_comparison(args, trainset),
            self.__create_method_list_for_lstm_comparison(args, trainset),
            self.__create_method_list_for_gerd_beta_comparison(args, true_model),
            self.__create_method_list_for_beam_width_in_rule_miner(args, true_model),
            self.__create_method_list_for_beam_width_in_cover(args, true_model),
            self.__create_method_list_for_beam_width_in_rule_miner_no_oracle(args, true_model),
            self.__create_method_list_rudi_topdown_vs_bottumup(args, true_model),
            self.__create_method_list_rudi_patience_topdown(args, true_model),
            self.__create_method_list_rudi_patience_bottomup(args, true_model),
            self.__create_method_list_rudi_bottumup_discard_failed_candidates_parameter(args, true_model),
            self.__create_method_list_for_rudi_single_vs_multiprocessing_comparison(args, true_model),
            self.__create_method_list_rudi_different_classifiers(args, true_model),
            self.__create_method_list_for_alignment_comparison(args, true_model),
            self.__create_method_list_bottom_up_discard_parameter(args, true_model),
            self.__create_method_list_window_classifier_varying_classifier(args, true_model),
            self.__create_method_list_window_classifier_window_size(args, true_model),
            self.__create_method_list_for_boomer_hyperparameter_comparison(args, true_model),
            self.__create_method_list_for_dats_abstraction_comparison(args),
            self.__create_method_list_for_dats_alpha_comparison(args),
            self.__create_method_list_for_dats_min_frequency_comparison(args),
            self.__create_method_list_for_dats_min_frequency_comparison_catnb(args),
            self.__create_method_list_for_mlrbc_hyperparameter_comparison(args),
            self.__create_method_list_for_event_classifiers_comparison(args)
        )

    def __create_method_list_for_cross_method_comparison(
            self, args, trainset: TargetSequenceDataset):
        methods = [
            ('DATS', Dats(
                state_representation_function=list_abstraction,
                computation_engine=create_computation_engine(args)),
                ResultCollector.colors.RED, ResultCollector.markers.CIRCLE),
            ('DATS_CatNB', Dats(
                state_representation_function=list_abstraction,
                classifier_type=CategoricalNaiveBayesClassifier,
                computation_engine=create_computation_engine(args)),
                ResultCollector.colors.RED, ResultCollector.markers.CIRCLE),
            ('RandomClassifier', RandomClassifier(),
                ResultCollector.colors.BLACK, ResultCollector.markers.TRIANGLE_UP),
            ('ConSequence', RuDi(
                logger=do_nothing,
                event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                    do_nothing, patience=10,
                    alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                    alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                    logger=do_nothing,
                    search_strategy=ConditionBeamSearch(beam_width=1))),
                computation_engine=create_computation_engine(args)),
                ResultCollector.colors.GREEN, ResultCollector.markers.X),
            ('LSTM', DataToSequenceLstm(
                trainset, epochs=1000, verbose=False,
                activation_candidates=['tanh'], min_delta=0.001,
                optimizer_candidates=['Adam'], layers_candidates=[(64,64), (256,256)],
                dropout_candidates=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5], round_limit=5),
                ResultCollector.colors.BLUE, ResultCollector.markers.PIXEL),
            ('GAN-LSTM', DataToSequenceAdversarialLstm(
                trainset, epochs=1000, verbose=False,
                activation_candidates=['tanh'],
                optimizer_candidates=['Adam'], layers_candidates=[(64,64), (256,256)],
                dropout_candidates=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5], round_limit=5,
                generator_exclusive_epochs=10),
                ResultCollector.colors.LIGHT_BLUE, ResultCollector.markers.POINT),
            ('KNN', ClassificationMiner(KnnClassifier()),
                ResultCollector.colors.ORANGE, ResultCollector.markers.DIAMOND),
            ('PosCl', EventClassifier(
                ReliableRuleMiner(logger=None),
                computation_engine=create_computation_engine(args)),
                ResultCollector.colors.BROWN, ResultCollector.markers.CIRCLE),
        ]
        try:
            methods.append(('BoomerLP', MultiLabelSequencePredictor(
                DatasetSplittingMiner(
                    lambda x_train, x_val: Boomer(
                        pruning=NaiveF1ScorePruning(x_val),
                        nr_of_threads=args.nr_of_cpus).mine_rules(x_train),
                    random_seed=args.random_seed),
                LpOrdering()),
                ResultCollector.colors.BLUISH_PURPLE, ResultCollector.markers.PENTAGON))
        except NameError:
            pass
        return methods

    def __create_method_list_for_beam_width_in_cover(
            self, args, trainset: TargetSequenceDataset):
        return [
                    ('ConSequence_CoverBeamWidth_1', RuDi(
                        logger=do_nothing,
                        event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                            do_nothing, patience=10,
                            alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                            alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 1)),
                        router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                            logger=do_nothing,
                            search_strategy=ConditionBeamSearch(beam_width=1))),
                        computation_engine=create_computation_engine(args)),
                     ResultCollector.colors.RED, ResultCollector.markers.X),
                    ('ConSequence_CoverBeamWidth_5', RuDi(
                        logger=do_nothing,
                        event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                            do_nothing, patience=10,
                            alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                            alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 5)),
                        router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                            logger=do_nothing,
                            search_strategy=ConditionBeamSearch(beam_width=1))),
                        computation_engine=create_computation_engine(args)),
                     ResultCollector.colors.BLUE, ResultCollector.markers.POINT),
                    ('ConSequence_CoverBeamWidth_10', RuDi(
                        logger=do_nothing,
                        event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                            do_nothing, patience=10,
                            alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                            alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                        router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                            logger=do_nothing,
                            search_strategy=ConditionBeamSearch(beam_width=1))),
                        computation_engine=create_computation_engine(args)),
                     ResultCollector.colors.GREEN, ResultCollector.markers.SQUARE),
                    ('ConSequence_CoverBeamWidth_15', RuDi(
                        logger=do_nothing,
                        event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                            do_nothing, patience=10,
                            alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                            alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 15)),
                        router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                            logger=do_nothing,
                            search_strategy=ConditionBeamSearch(beam_width=1))),
                        computation_engine=create_computation_engine(args)),
                     ResultCollector.colors.ORANGE, ResultCollector.markers.HEXAGON),
                    ('ConSequence_CoverBeamWidth_20', RuDi(
                        logger=do_nothing,
                        event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                            do_nothing, patience=10,
                            alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                            alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 20)),
                        router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                            logger=do_nothing,
                            search_strategy=ConditionBeamSearch(beam_width=1))),
                        computation_engine=create_computation_engine(args)),
                     ResultCollector.colors.LIGHT_BROWN, ResultCollector.markers.PENTAGON),
                    ('ConSequence_CoverBeamWidth_50', RuDi(
                        logger=do_nothing,
                        event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                            do_nothing, patience=10,
                            alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                            alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 50)),
                        router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                            logger=do_nothing,
                            search_strategy=ConditionBeamSearch(beam_width=1))),
                        computation_engine=create_computation_engine(args)),
                     ResultCollector.colors.GREY, ResultCollector.markers.POINT),
            ]

    def __create_method_list_for_rudi_single_vs_multiprocessing_comparison(
            self, args, trainset: TargetSequenceDataset):
        return [
            ('ConSequence_1_CPU', RuDi(
                logger=do_nothing,
                event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                    do_nothing, patience=10,
                    alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                    alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                    logger=do_nothing,
                    search_strategy=ConditionBeamSearch(beam_width=1)))
                ), ResultCollector.colors.GREEN, ResultCollector.markers.X),
            ('ConSequence_2_CPUs', RuDi(
                logger=do_nothing,
                event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                    do_nothing, patience=10,
                    alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                    alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                    logger=do_nothing,
                    search_strategy=ConditionBeamSearch(beam_width=1))),
                computation_engine=MultiprocessComputationEngine(2)
                ), ResultCollector.colors.BLUE, ResultCollector.markers.X),
            ('ConSequence_4_CPUs', RuDi(
                logger=do_nothing,
                event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                    do_nothing, patience=10,
                    alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                    alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                    logger=do_nothing,
                    search_strategy=ConditionBeamSearch(beam_width=1))),
                computation_engine=MultiprocessComputationEngine(4)
                ), ResultCollector.colors.ORANGE, ResultCollector.markers.X),
            ('ConSequence_8_CPUs', RuDi(
                logger=do_nothing,
                event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                    do_nothing, patience=10,
                    alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                    alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                    logger=do_nothing,
                    search_strategy=ConditionBeamSearch(beam_width=1))),
                computation_engine=MultiprocessComputationEngine(8)
                ), ResultCollector.colors.RED, ResultCollector.markers.X),
            ('ConSequence_Ray', RuDi(
                logger=do_nothing,
                event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                    do_nothing, patience=10,
                    alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                    alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                    logger=do_nothing,
                    search_strategy=ConditionBeamSearch(beam_width=1))),
                computation_engine=RayComputationEngine()
                ), ResultCollector.colors.CYAN, ResultCollector.markers.PENTAGON),
        ]

    def __create_method_list_rudi_topdown_vs_bottumup(
            self, args, trainset: TargetSequenceDataset):
        return [
                    ('ConSequence_BottomUp', RuDi(logger=do_nothing,
                                   event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                       do_nothing, patience=10,
                                       alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                       alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                                   router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                       logger=do_nothing,
                                       search_strategy=ConditionBeamSearch(beam_width=1)))),
                     ResultCollector.colors.BLUE, ResultCollector.markers.SQUARE),
                    ('ConSequence_TopDown', RuDi(logger=do_nothing,
                                   event_flow_graph_miner=SequenceTopDownEventFlowGraphMiner(
                                       do_nothing, patience=10,
                                       alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                                   router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                       logger=do_nothing,
                                       search_strategy=ConditionBeamSearch(beam_width=1)))),
                     ResultCollector.colors.RED, ResultCollector.markers.CIRCLE),
                    ('ConSequence_TopDown_Fast', RuDi(logger=do_nothing,
                                   event_flow_graph_miner=SequenceTopDownEventFlowGraphMiner(
                                       do_nothing, patience=10, exact_mdl=False,
                                       alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                                   router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                       logger=do_nothing,
                                       search_strategy=ConditionBeamSearch(beam_width=1)))),
                     ResultCollector.colors.GREEN, ResultCollector.markers.PLUS),
            ]

    def __create_method_list_rudi_bottumup_discard_failed_candidates_parameter(
            self, args, trainset: TargetSequenceDataset):
        return [
                    ('ConSequence_BottomUp_Discard', RuDi(logger=do_nothing,
                                   event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                       do_nothing, patience=10, discard_failed_candidates=True,
                                       alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                       alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                                   router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                       logger=do_nothing,
                                       search_strategy=ConditionBeamSearch(beam_width=1)))),
                     ResultCollector.colors.BLUE, ResultCollector.markers.SQUARE),
                    ('ConSequence_BottomUp_Keep', RuDi(logger=do_nothing,
                                   event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                       do_nothing, patience=10, discard_failed_candidates=True,
                                       alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                       alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                                   router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                       logger=do_nothing,
                                       search_strategy=ConditionBeamSearch(beam_width=1)))),
                     ResultCollector.colors.RED, ResultCollector.markers.X),
            ]

    def __create_method_list_rudi_patience_topdown(
            self, args, trainset: TargetSequenceDataset):
        return [
                    ('ConSequence_TopDown_Patience5', RuDi(logger=do_nothing,
                                   event_flow_graph_miner=SequenceTopDownEventFlowGraphMiner(
                                       do_nothing, patience=5,
                                       alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                                   router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                       logger=do_nothing,
                                       search_strategy=ConditionBeamSearch(beam_width=1)))),
                     ResultCollector.colors.RED, ResultCollector.markers.PENTAGON),
                    ('ConSequence_TopDown_Patience10', RuDi(logger=do_nothing,
                                   event_flow_graph_miner=SequenceTopDownEventFlowGraphMiner(
                                       do_nothing, patience=10,
                                       alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                                   router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                       logger=do_nothing,
                                       search_strategy=ConditionBeamSearch(beam_width=1)))),
                     ResultCollector.colors.BLUE, ResultCollector.markers.PLUS),
                    ('ConSequence_TopDown_Patience25', RuDi(logger=do_nothing,
                                   event_flow_graph_miner=SequenceTopDownEventFlowGraphMiner(
                                       do_nothing, patience=25,
                                       alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                                   router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                       logger=do_nothing,
                                       search_strategy=ConditionBeamSearch(beam_width=1)))),
                     ResultCollector.colors.GREEN, ResultCollector.markers.SQUARE),
                    ('ConSequence_TopDown_Patience50', RuDi(logger=do_nothing,
                                   event_flow_graph_miner=SequenceTopDownEventFlowGraphMiner(
                                       do_nothing, patience=50,
                                       alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                                   router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                       logger=do_nothing,
                                       search_strategy=ConditionBeamSearch(beam_width=1)))),
                     ResultCollector.colors.ORANGE, ResultCollector.markers.CIRCLE),
            ]

    def __create_method_list_rudi_patience_bottomup(
            self, args, trainset: TargetSequenceDataset):
        return [
            ('ConSequence_BottomUp_Patience5', RuDi(logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1)))),
                ResultCollector.colors.RED, ResultCollector.markers.PENTAGON),
            ('ConSequence_BottomUp_Patience10', RuDi(logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1)))),
                ResultCollector.colors.BLUE, ResultCollector.markers.PLUS),
            ('ConSequence_BottomUp_Patience25', RuDi(logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1)))),
                ResultCollector.colors.GREEN, ResultCollector.markers.SQUARE),
            ('ConSequence_BottomUp_Patience50', RuDi(logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1)))),
                ResultCollector.colors.ORANGE, ResultCollector.markers.CIRCLE),
        ]

    def __create_method_list_rudi_different_classifiers(
            self, args, trainset: TargetSequenceDataset):
        return [
            ('ConSequence_Gerd', RuDi(logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1)))),
                ResultCollector.colors.BLUE, ResultCollector.markers.PENTAGON),
            ('ConSequence_Classy', RuDi(logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(Classy())),
             ResultCollector.colors.ORANGE, ResultCollector.markers.X),
            ('ConSequence_CAntMinerPB', RuDi(logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(
                                CAntMinerPB(
                                    random_seed=args.random_seed,
                                    nr_of_workers=args.nr_of_cpus,
                                    max_java_heapspace_in_mb=args.max_java_heapspace_in_mb))),
                ResultCollector.colors.RED, ResultCollector.markers.SQUARE),
        ]

    def __create_method_list_for_lstm_comparison(self, args, trainset: TargetSequenceDataset):
        return [
                    (
                        'PlainLstm',
                        DataToSequenceLstm(
                            trainset, epochs=500, verbose=False,
                            activation_candidates=['tanh'], min_delta=0.001,
                            optimizer_candidates=['Adam'],
                            layers_candidates=[(64,64), (256,256), (1024, 1024)],
                            dropout_candidates=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5], round_limit=5
                        ),
                        ResultCollector.colors.RED, ResultCollector.markers.X
                    ),
                    (
                        'AdversarialLstm1',
                        DataToSequenceAdversarialLstm(
                            trainset, epochs=500, verbose=False,
                            activation_candidates=['tanh'],
                            optimizer_candidates=['Adam'],
                            layers_candidates=[(64,64), (256,256), (1024, 1024)],
                            dropout_candidates=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5], round_limit=5,
                            generator_exclusive_epochs=1
                        ),
                        ResultCollector.colors.BLUE, ResultCollector.markers.PLUS
                    ),
                    (
                        'AdversarialLstm10',
                        DataToSequenceAdversarialLstm(
                            trainset, epochs=500, verbose=False,
                            activation_candidates=['tanh'],
                            optimizer_candidates=['Adam'],
                            layers_candidates=[(64,64), (256,256), (1024, 1024)],
                            dropout_candidates=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5], round_limit=5,
                            generator_exclusive_epochs=10
                        ),
                        ResultCollector.colors.GREEN, ResultCollector.markers.SQUARE
                    ),
                    (
                        'AdversarialLstm50',
                        DataToSequenceAdversarialLstm(
                            trainset, epochs=500, verbose=False,
                            activation_candidates=['tanh'],
                            optimizer_candidates=['Adam'],
                            layers_candidates=[(64,64), (256,256), (1024, 1024)],
                            dropout_candidates=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5], round_limit=5,
                            generator_exclusive_epochs=50
                        ),
                        ResultCollector.colors.ORANGE, ResultCollector.markers.DIAMOND
                    ),
            ]

    def __create_method_list_for_gerd_beta_comparison(self, args, true_model: Rule):
        return [
                    (
                        'ConSequence_Gerd_beta0',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1, beta=0.0))),
                            computation_engine=create_computation_engine(args)
                        ),
                        ResultCollector.colors.RED, ResultCollector.markers.PENTAGON
                    ),
                    (
                        'ConSequence_Gerd_beta1',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1, beta=1.0))),
                            computation_engine=create_computation_engine(args)
                        ),
                        ResultCollector.colors.BLUE, ResultCollector.markers.PENTAGON
                    ),
                    (
                        'ConSequence_Gerd_beta2',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1, beta=2.0))),
                            computation_engine=create_computation_engine(args)
                        ),
                        ResultCollector.colors.GREEN, ResultCollector.markers.PENTAGON
                    ),
                    (
                        'ConSequence_Gerd_beta3',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1, beta=3.0))),
                            computation_engine=create_computation_engine(args)
                        ),
                        ResultCollector.colors.PINK, ResultCollector.markers.PENTAGON
                    ),
                    (
                        'ConSequence_Gerd_beta4',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1, beta=4.0))),
                            computation_engine=create_computation_engine(args)
                        ),
                        ResultCollector.colors.GOLD, ResultCollector.markers.PENTAGON
                    ),
                    (
                        'ConSequence_Gerd_beta5',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1, beta=5.0))),
                            computation_engine=create_computation_engine(args)
                        ),
                        ResultCollector.colors.ORANGE, ResultCollector.markers.PENTAGON
                    ),
            ]

    def __create_method_list_for_beam_width_in_rule_miner(self, args, true_model: Rule):
        return [
                    (
                        'Oracle_ReliableRuleMiner_BeamWidth_1',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=OracleEventFlowGraphMiner(
                                rule_to_eventflow_graph(true_model).copy() if true_model else None
                            ),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1))
                            )
                        ),
                        ResultCollector.colors.ORANGE, ResultCollector.markers.SQUARE
                    ),
                    (
                        'Oracle_ReliableRuleMiner_BeamWidth_5',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=OracleEventFlowGraphMiner(
                                rule_to_eventflow_graph(true_model).copy() if true_model else None
                            ),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=5))
                            )
                        ),
                        ResultCollector.colors.BLUE, ResultCollector.markers.PLUS
                    ),
                    (
                        'Oracle_ReliableRuleMiner_BeamWidth_10',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=OracleEventFlowGraphMiner(
                                rule_to_eventflow_graph(true_model).copy() if true_model else None
                            ),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=10))
                            )
                        ),
                        ResultCollector.colors.GREEN, ResultCollector.markers.CIRCLE
                    ),
                    (
                        'Oracle_ReliableRuleMiner_BeamWidth_20',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=OracleEventFlowGraphMiner(
                                rule_to_eventflow_graph(true_model).copy() if true_model else None
                            ),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=20))
                            )
                        ),
                        ResultCollector.colors.RED, ResultCollector.markers.PENTAGON
                    ),
            ]

    def __create_method_list_for_beam_width_in_rule_miner_no_oracle(self, args, true_model: Rule):
        return [
                    (
                        'ConSequence_ReliableRuleMiner_BeamWidth_1',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=alignment_finder_factory_score
                            ),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=1))
                            )
                        ),
                        ResultCollector.colors.ORANGE, ResultCollector.markers.SQUARE
                    ),
                    (
                        'ConSequence_ReliableRuleMiner_BeamWidth_5',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=alignment_finder_factory_score
                            ),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=5))
                            )
                        ),
                        ResultCollector.colors.BLUE, ResultCollector.markers.PLUS
                    ),
                    (
                        'ConSequence_ReliableRuleMiner_BeamWidth_10',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=alignment_finder_factory_score
                            ),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=10))
                            )
                        ),
                        ResultCollector.colors.GREEN, ResultCollector.markers.CIRCLE
                    ),
                    (
                        'ConSequence_ReliableRuleMiner_BeamWidth_20',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=alignment_finder_factory_score
                            ),
                            router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                logger=do_nothing,
                                search_strategy=ConditionBeamSearch(beam_width=20))
                            )
                        ),
                        ResultCollector.colors.RED, ResultCollector.markers.PENTAGON
                    ),
            ]

    def __create_method_list_for_alignment_comparison(self, args, true_model: Rule):
        return [
                    (
                        'ConSequence_AStar_Oracle',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: AStar(g, NullHeuristic())
                            ),
                            router_learner=OracleRouterLearner()
                        ),
                        ResultCollector.colors.ORANGE, ResultCollector.markers.SQUARE
                    ),
                    (
                        'ConSequence_Petrinet_Oracle',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=PetrinetAligner
                            ),
                            router_learner=OracleRouterLearner()
                        ),
                        ResultCollector.colors.CYAN, ResultCollector.markers.TRIANGLE_RIGHT
                    ),
                    (
                        'ConSequence_GreedyBeamWidth5_Oracle',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: GreedyBeamSearch(g, ReachabilityHeuristic(g), 5)
                            ),
                            router_learner=OracleRouterLearner()
                        ),
                        ResultCollector.colors.RED, ResultCollector.markers.TRIANGLE_RIGHT
                    ),
                    (
                        'ConSequence_BeamWidth1_Oracle',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 1)
                            ),
                            router_learner=OracleRouterLearner()
                        ),
                        ResultCollector.colors.BLUE, ResultCollector.markers.PLUS
                    ),
                    (
                        'ConSequence_BeamWidth5_Oracle',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 5)
                            ),
                            router_learner=OracleRouterLearner()
                        ),
                        ResultCollector.colors.GREEN, ResultCollector.markers.CIRCLE
                    ),
                    (
                        'ConSequence_BeamWidth10_Oracle',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)
                            ),
                            router_learner=OracleRouterLearner()
                        ),
                        ResultCollector.colors.RED, ResultCollector.markers.PENTAGON
                    ),
                    (
                        'ConSequence_BeamWidth20_Oracle',
                        RuDi(
                            logger=do_nothing,
                            event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                do_nothing, patience=10,
                                alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 20)
                            ),
                            router_learner=OracleRouterLearner()
                        ),
                        ResultCollector.colors.LIME, ResultCollector.markers.PIXEL
                    ),
            ]

    def __create_method_list_bottom_up_discard_parameter(self, args, true_model: Rule):
        return [
                    (
                        'ConSequence_BottomUp_Discard_True', RuDi(logger=do_nothing,
                                   event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                       do_nothing, patience=10,
                                       discard_failed_candidates=True,
                                       alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                       alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                                   router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                       logger=do_nothing,
                                       search_strategy=ConditionBeamSearch(beam_width=1)))),
                        ResultCollector.colors.BLUE, ResultCollector.markers.PLUS
                    ),
                    (
                        'ConSequence_BottomUp_Discard_False', RuDi(logger=do_nothing,
                                   event_flow_graph_miner=SequenceBottomUpEventFlowGraphMiner(
                                       do_nothing, patience=10,
                                       discard_failed_candidates=False,
                                       alignment_finder_factory_model_extension=alignment_finder_factory_model_extension,
                                       alignment_finder_factory_score=lambda g: BeamSearch(g, ReachabilityHeuristic(g), 10)),
                                   router_learner=RuleClassifierRouterLearner(ReliableRuleMiner(
                                       logger=do_nothing,
                                       search_strategy=ConditionBeamSearch(beam_width=1)))),
                        ResultCollector.colors.RED, ResultCollector.markers.PLUS
                    ),
            ]

    def __create_method_list_window_classifier_varying_classifier(self, args, true_model: Rule):
        return [
            (
                'WindowClassifier_ReliableRuleMiner',
                WindowClassifier(
                    classification_rule_miner=ReliableRuleMiner(logger=do_nothing)
                ),
                ResultCollector.colors.BLUE, ResultCollector.markers.PLUS
            ),
            (
                'WindowClassifier_CAntMinerPB',
                WindowClassifier(
                    classification_rule_miner=CAntMinerPB(
                        random_seed=args.random_seed,
                        nr_of_workers=args.nr_of_cpus)
                ),
                ResultCollector.colors.RED, ResultCollector.markers.TRIANGLE_RIGHT
            ),
            (
                'WindowClassifier_Classy',
                WindowClassifier(
                    classification_rule_miner=Classy()),
                ResultCollector.colors.GREEN, ResultCollector.markers.X
            ),
        ]

    def __create_method_list_window_classifier_window_size(self, args, true_model: Rule):
        return [
            (
                'WindowClassifier_w3',
                WindowClassifier(
                    nr_of_past_events_as_features=3,
                    classification_rule_miner=ReliableRuleMiner(logger=do_nothing)
                ),
                ResultCollector.colors.BLUE, ResultCollector.markers.PLUS
            ),
            (
                'WindowClassifier_w5',
                WindowClassifier(
                    nr_of_past_events_as_features=5,
                    classification_rule_miner=ReliableRuleMiner(logger=do_nothing)
                ),
                ResultCollector.colors.RED, ResultCollector.markers.X
            ),
            (
                'WindowClassifier_w10',
                WindowClassifier(
                    nr_of_past_events_as_features=10,
                    classification_rule_miner=ReliableRuleMiner(logger=do_nothing)
                ),
                ResultCollector.colors.GREEN, ResultCollector.markers.PIXEL
            ),
        ]

    def __create_method_list_for_boomer_hyperparameter_comparison(
            self, args, trainset: TargetSequenceDataset):
        try:
            return [
                ('BoomerWithPruning', MultiLabelSequencePredictor(
                    DatasetSplittingMiner(
                        lambda x_train, x_val: Boomer(
                            pruning=NaiveF1ScorePruning(x_val),
                            nr_of_threads=args.nr_of_cpus).mine_rules(x_train),
                        random_seed=args.random_seed),
                    LpOrdering()),
                ResultCollector.colors.BLUE, ResultCollector.markers.PENTAGON),
                ('BoomerNoPruning', MultiLabelSequencePredictor(
                    Boomer(nr_of_threads=args.nr_of_cpus), LpOrdering()),
                ResultCollector.colors.RED, ResultCollector.markers.X),
            ]
        except NameError as e:
            return []

    def __create_method_list_for_dats_abstraction_comparison(self, args):
        return [
            ('DATS_List_MixedNB', Dats(state_representation_function=list_abstraction),
             ResultCollector.colors.BLUE, ResultCollector.markers.PENTAGON),
            ('DATS_Set_MixedNB', Dats(state_representation_function=set_abstraction),
             ResultCollector.colors.RED, ResultCollector.markers.X),
            ('DATS_Last_MixedNB', Dats(state_representation_function=last_abstraction),
             ResultCollector.colors.ORANGE, ResultCollector.markers.CIRCLE),
            ('DATS_Last2_MixedNB', Dats(state_representation_function=last_n_abstraction(2)),
             ResultCollector.colors.GREEN, ResultCollector.markers.PIXEL),
            ('DATS_Last5_MixedNB', Dats(state_representation_function=last_n_abstraction(5)),
             ResultCollector.colors.PINK, ResultCollector.markers.PLUS),
        ]

    def __create_method_list_for_dats_alpha_comparison(self, args):
        return [
            ('DATS_alpha001', Dats(state_representation_function=last_n_abstraction(5), classifier_parameters={'alpha': 0.01}),
             ResultCollector.colors.BLUE, ResultCollector.markers.PENTAGON),
            ('DATS_alpha05', Dats(state_representation_function=last_n_abstraction(5), classifier_parameters={'alpha': 0.5}),
             ResultCollector.colors.PINK, ResultCollector.markers.PLUS),
            ('DATS_alpha1', Dats(state_representation_function=last_n_abstraction(5), classifier_parameters={'alpha': 1.0}),
             ResultCollector.colors.GREEN, ResultCollector.markers.PIXEL),
        ]

    def __create_method_list_for_dats_min_frequency_comparison(self, args):
        return [
            ('DATS_min_freq_0', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.0),
             ResultCollector.colors.BLUE, ResultCollector.markers.PENTAGON),
            ('DATS_min_freq_01', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.1),
             ResultCollector.colors.RED, ResultCollector.markers.PIXEL),
            ('DATS_min_freq_02', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.2),
             ResultCollector.colors.GREEN, ResultCollector.markers.POINT),
            ('DATS_min_freq_03', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.3),
             ResultCollector.colors.PINK, ResultCollector.markers.HEXAGON),
            ('DATS_min_freq_04', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.4),
             ResultCollector.colors.ORANGE, ResultCollector.markers.SQUARE),
            ('DATS_min_freq_05', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.5),
             ResultCollector.colors.BLUE, ResultCollector.markers.X),
            ('DATS_min_freq_06', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.6),
             ResultCollector.colors.RED, ResultCollector.markers.X),
            ('DATS_min_freq_07', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.7),
             ResultCollector.colors.GREEN, ResultCollector.markers.X),
            ('DATS_min_freq_08', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.8),
             ResultCollector.colors.PINK, ResultCollector.markers.X),
            ('DATS_min_freq_09', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.9),
             ResultCollector.colors.LIGHT_BROWN, ResultCollector.markers.X),
        ]

    def __create_method_list_for_dats_min_frequency_comparison_catnb(self, args):
        return [
            ('DATS_catnb_min_freq_0', Dats(
                state_representation_function=list_abstraction,
                classifier_type=CategoricalNaiveBayesClassifier,
                min_relative_sequence_frequency=0.0),
             ResultCollector.colors.BLUE, ResultCollector.markers.PENTAGON),
            ('DATS_catnb_min_freq_01', Dats(
                state_representation_function=list_abstraction,
                classifier_type=CategoricalNaiveBayesClassifier,
                min_relative_sequence_frequency=0.1),
             ResultCollector.colors.RED, ResultCollector.markers.PIXEL),
            ('DATS_catnb_min_freq_02', Dats(
                state_representation_function=list_abstraction,
                classifier_type=CategoricalNaiveBayesClassifier,
                min_relative_sequence_frequency=0.2),
             ResultCollector.colors.GREEN, ResultCollector.markers.POINT),
            ('DATS_catnb_min_freq_03', Dats(
                state_representation_function=list_abstraction,
                classifier_type=CategoricalNaiveBayesClassifier,
                min_relative_sequence_frequency=0.3),
             ResultCollector.colors.PINK, ResultCollector.markers.HEXAGON),
            ('DATS_catnb_min_freq_04', Dats(
                state_representation_function=list_abstraction,
                classifier_type=CategoricalNaiveBayesClassifier,
                min_relative_sequence_frequency=0.4),
             ResultCollector.colors.ORANGE, ResultCollector.markers.SQUARE),
            ('DATS_catnb_min_freq_05', Dats(
                state_representation_function=list_abstraction,
                classifier_type=CategoricalNaiveBayesClassifier,
                min_relative_sequence_frequency=0.5),
             ResultCollector.colors.BLUE, ResultCollector.markers.X),
            ('DATS_catnb_min_freq_06', Dats(
                state_representation_function=list_abstraction,
                classifier_type=CategoricalNaiveBayesClassifier,
                min_relative_sequence_frequency=0.6),
             ResultCollector.colors.RED, ResultCollector.markers.X),
            ('DATS_catnb_min_freq_07', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.7),
             ResultCollector.colors.GREEN, ResultCollector.markers.X),
            ('DATS_catnb_min_freq_08', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.8),
             ResultCollector.colors.PINK, ResultCollector.markers.X),
            ('DATS_catnb_min_freq_09', Dats(
                state_representation_function=list_abstraction,
                min_relative_sequence_frequency=0.9),
             ResultCollector.colors.LIGHT_BROWN, ResultCollector.markers.X),
        ]

    def __create_method_list_for_mlrbc_hyperparameter_comparison(self, args):
        return [
            (
                'MLRBC_LP_Default',
                MultiLabelSequencePredictor(
                    Mlrbc(logger=None),
                    LpOrdering()
                ),
                ResultCollector.colors.BROWN, ResultCollector.markers.CIRCLE
            ),
            (
                'MLRBC_LP_Roulette',
                MultiLabelSequencePredictor(
                    Mlrbc(logger=None, selection_method='roulette'),
                    LpOrdering()
                ),
                ResultCollector.colors.RED, ResultCollector.markers.X
            ),
            (
                'MLRBC_LP_5000iters',
                MultiLabelSequencePredictor(
                    Mlrbc(logger=None, max_learning_iterations=5000),
                    LpOrdering()
                ),
                ResultCollector.colors.BLUE, ResultCollector.markers.POINT
            ),
            (
                'MLRBC_LP_nu1',
                MultiLabelSequencePredictor(
                    Mlrbc(logger=None, nu=1),
                    LpOrdering()
                ),
                ResultCollector.colors.GREEN, ResultCollector.markers.DIAMOND
            ),
            (
                'MLRBC_LP_nu10',
                MultiLabelSequencePredictor(
                    Mlrbc(logger=None, nu=10),
                    LpOrdering()
                ),
                ResultCollector.colors.ORANGE, ResultCollector.markers.PENTAGON
            ),
        ]

    def __create_method_list_for_event_classifiers_comparison(self, args):
        return [
            (
                'PosCl_CAntMinerPB',
                EventClassifier(
                    CAntMinerPB(nr_of_workers=1),
                    computation_engine=create_computation_engine(args)
                ),
                ResultCollector.colors.RED, ResultCollector.markers.X
            ),
            (
                'PosCl_Classy',
                EventClassifier(
                    Classy(),
                    computation_engine=create_computation_engine(args)
                ),
                ResultCollector.colors.BLUE, ResultCollector.markers.POINT
            ),
            (
                'PosCl_ReliableRuleMiner',
                EventClassifier(
                    ReliableRuleMiner(logger=None),
                    computation_engine=create_computation_engine(args)
                ),
                ResultCollector.colors.ORANGE, ResultCollector.markers.DIAMOND
            )
        ]
