# -*- coding: utf-8 -*-

import os
import argparse

import dill as pickle
from tqdm import tqdm

from experimentlogger.experiment_logger import ExperimentLogger
from prolothar_common.experiments.stopwatch import Stopwatch
from prolothar_common.models.dataset import TargetSequenceDataset

from prolothar_rule_mining.rule_miner.data_to_sequence.data_to_sequence_lstm import BaseLstm
from prolothar_rule_mining.rule_miner.data_to_sequence.data_to_sequence_lstm import DataToSequenceLstm
from prolothar_rule_mining.rule_miner.data_to_sequence.data_to_sequence_lstm import DataToSequenceAdversarialLstm

from experiments.rule_mining.dataset_loader import DatasetLoader
from experiments.rule_mining.experiment_utils import plot_and_save_true_model

class LstmRunner:

    def __init__(self):
        self.__dataset_loader = DatasetLoader()
        self.__logger: ExperimentLogger = None

    def run_from_cli(self):
        args = self.__parse_cli_arguments()

        self.__logger = ExperimentLogger(
            os.path.join('experiments', 'rule_mining', 'data_to_sequence_lstm', 'results'),
            args.dataset)
        self.__logger.log(str(args))

        dataset,true_model = self.__dataset_loader.read_dataset(args, 0)
        trainset,testset = dataset.split(
            args.testset_proportion, random_seed=args.random_seed)
        if args.trainset_size > 0:
            trainset = trainset.random_subset(args.trainset_size,
                                              random_seed=args.random_seed)
        self.__logger.log('dataset contains %d instances' % len(dataset))
        self.__logger.log('trainset contains %d instances' % len(trainset))
        self.__logger.log('testset contains %d instances' % len(testset))

        if true_model is not None:
            plot_and_save_true_model(true_model, trainset, self.__logger)

        stopwatch = Stopwatch()
        stopwatch.start()
        model = self.__create_model(args, trainset)
        self.__logger.log('')
        self.__logger.log('elapsed time: %r' % stopwatch.get_elapsed_time())
        model = model.mine_rules(trainset)
        with self.__logger.open_binary_file('mined_model.pickle') as f:
            pickle.dump(model, f)

        self.__explain_model(model, testset)

    def __explain_model(self, model: BaseLstm, testset: TargetSequenceDataset):
        self.__logger.log('explain model predictions on testset')
        explanation_directory = os.path.join(
            self.__logger.get_path_to_experiment_directory(), 'explanations')
        os.mkdir(explanation_directory)
        for instance in tqdm(testset.random_subset(min(50, len(dataset)))):
            predicted_sequence, explanation_list = model.execute_with_explanation(instance)
            instance_explanation_directory = os.path.join(
                explanation_directory, str(instance.get_id()))
            os.mkdir(instance_explanation_directory)
            with open(os.path.join(instance_explanation_directory, 'prediction.txt'), 'w') as f:
                f.write(str(predicted_sequence))
            for i, explanation in enumerate(explanation_list):
                with open(
                        os.path.join(
                            instance_explanation_directory,
                            f'explanation_{i}_{predicted_sequence[i]}.html'),
                        'w', encoding='utf-8') as f:
                    f.write(explanation.as_html())
                with open(
                        os.path.join(
                            instance_explanation_directory,
                            f'explanation_{i}_{predicted_sequence[i]}.txt'),
                        'w', encoding='utf-8') as f:
                    f.write(str(explanation.as_map()))

    def __create_model(self, args, trainset: TargetSequenceDataset) -> BaseLstm:
        if args.architecture == 'plain':
            return DataToSequenceLstm(
                trainset, epochs=500, verbose=True,
                activation_candidates=['tanh'], min_delta=0.001,
                optimizer_candidates=['Adam'],
                layers_candidates=[(64,64), (256,256), (1024, 1024)],
                dropout_candidates=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5],
                round_limit=args.hyperparameter_round_limit
            )
        elif args.architecture == 'adversarial':
            return DataToSequenceAdversarialLstm(
                trainset, epochs=500, verbose=True,
                activation_candidates=['tanh'],
                optimizer_candidates=['Adam'],
                layers_candidates=[(64,64), (256,256), (1024, 1024)],
                dropout_candidates=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5],
                generator_exclusive_epochs=10,
                round_limit=args.hyperparameter_round_limit
            )
        else:
            raise NotImplementedError(args.model)

    def __parse_cli_arguments(self):
        parser = argparse.ArgumentParser('Data2Seq rule mining on a single dataset')

        parser.add_argument(
                '--trainset_size', type=int,
                help=('nr of instances in the trainset. default means maximal size'),
                required=False, default=-1)

        parser.add_argument(
                '--testset_proportion', type=float,
                help='proportion in (0,1) of the testset',
                required=False, default=0.2)

        parser.add_argument(
                '--architecture', type=str,
                help=('model architecture to use'),
                required=False, default='plain',
                choices=['plain', 'adversarial'])

        parser.add_argument(
                '--hyperparameter_round_limit', type=int,
                help=('number of training runs with different hyperparameters'),
                required=False, default=5)

        self.__dataset_loader.add_cli_arguments(parser)

        return parser.parse_args()

if __name__ == '__main__':
    LstmRunner().run_from_cli()
