from typing import Dict, Tuple

from random import Random
import argparse
from distutils.util import strtobool

from prolothar_common.models.dataset import TargetSequenceDataset
try:
    import prolothar_data_private.kostenfertigungsplan.blechdaten_fertigungsplan as walzwerk_fpl
except ImportError:
    #private data is not available in public experiments
    pass
try:
    import prolothar_data_public.bpi_challenge.bpi_2019.bpi_challenge_2019 as bpi2019
    import prolothar_data_public.bpi_challenge.bpi_2017.bpi_challenge_2017 as bpi2017
    import prolothar_data_public.bpi_challenge.bpi_2015.bpi_challenge_2015 as bpi2015
    import prolothar_data_public.proteomics.spectrometry as proteomics
except ImportError:
    #some datasets are not available in CMT uploaded version due to file size restrictions
    pass
import prolothar_data_public.e2e as e2e
import prolothar_data_public.rolling_mill.rolling_mill as rolling_mill
import prolothar_data_public.dconvert as dconvert
import prolothar_data_public.sepsis_cases.sepsis as sepsis
import prolothar_data_public.production_analysis.production as production_analysis
import prolothar_data_public.helpdesk.helpdesk as helpdesk
import prolothar_data_public.covid.germany as covid_germany
import prolothar_data_public.traffic_fines.traffic_fines as traffic_fines
from prolothar_rule_mining.models.dataset_generator import TargetSequenceDatasetGenerator as DatasetGenerator
from prolothar_rule_mining.rule_miner.data_to_sequence.rules import Rule

def parse_rule_list_length(arg: list):
    if len(arg) == 1:
        return arg[0]
    else:
        return { i: x for i,x in enumerate(arg) }

class DatasetLoader:
    def __init__(self):
        self.__cached_dataset = None
        self.__cached_models: Dict[int, Rule] = {}

    def add_cli_arguments(self, parser: argparse.ArgumentParser,
                          include_nr_of_repetitions: bool = True):
        parser.add_argument(
                '--dataset', type=str,
                help='the dataset to use in the experiment',
                required=True)

        parser.add_argument(
                '--cache_artificial_models', type=lambda x: bool(strtobool(x)),
                help=('if True then artificial models are cached, otherwise '
                       'at every sample point a new model is generated.'),
                required=False, default=True)

        parser.add_argument(
                '--random_seed', type=int,
                help='seed for random generators to create non-stochastic results',
                required=False, default=None)

        if include_nr_of_repetitions:
            parser.add_argument(
                    '--nr_of_repetitions', type=int,
                    help='nr of runs per nr of instances',
                    required=False, default=10)

        parser.add_argument(
                '--nr_of_categorical_features', type=int,
                help='nr of categorical features in the artificial dataset',
                required=False, default=2)

        parser.add_argument(
                '--nr_of_numerical_features', type=int,
                help='nr of numerical features in the artificial dataset',
                required=False, default=3)

        parser.add_argument(
                '--nr_of_categories', type=int,
                help='nr of categories for categorical features in the artificial dataset',
                required=False, default=3)

        parser.add_argument(
                '--max_condition_depth', type=int,
                help='maximum depth of artificial generated rule conditions',
                required=False, default=0)

        parser.add_argument(
                '--max_nr_of_nodes_in_model', type=int,
                help='maximal number of nodes in artificial eventflow graphs',
                required=False, default=25)

        parser.add_argument(
                '--min_rule_length', type=int,
                help='min length of rule list (if model = "list")',
                required=False, default=[10,1,1], nargs='+')
        parser.add_argument(
                '--max_rule_length', type=int,
                help='min length of rule list (if model = "list")',
                required=False, default=[20,3,1], nargs='+')
        parser.add_argument(
                '--max_rule_depth', type=int,
                help='max depth of if else rules (if model = "list")',
                required=False, default=1)

        parser.add_argument(
                '--edge_probability', type=float,
                help=('controls how densily connected an artificial model is. '
                      'only applies if model == eventflowgraph'),
                required=False, default=0.1)

        parser.add_argument(
                '--size_of_artificial_dataset', type=int,
                help='number of instances in the artificial dataset',
                required=False, default=10000)

        parser.add_argument(
                '--nr_of_sequence_symbols', type=int,
                help='nr of sequence symbols in the alphabet for the artificial dataset',
                required=False, default=20)

        parser.add_argument(
                '--model', type=str,
                help='the model type for artificial data',
                choices=['eventflowgraph', 'list', 'eventflowgraph2'],
                default='eventflowgraph',
                required=False)

    def read_dataset(self, args, repetition_index: int) -> Tuple[TargetSequenceDataset, Rule]:
        true_model = None
        if args.dataset == 'DH_WW_FPL':
            if self.__cached_dataset is None:
                self.__cached_dataset = \
                    walzwerk_fpl.load_blechdaten_fertigungsplan_dataset(
                        remove_identifier_attributes=True
                    )
            dataset = self.__cached_dataset
        elif args.dataset == 'Purchases':
            dataset = bpi2019.load_dataset()
        elif args.dataset == 'Loans':
            dataset = bpi2017.load_dataset()
        elif args.dataset == 'Permits':
            dataset = bpi2015.load_dataset()
        elif args.dataset == 'Production':
            dataset = production_analysis.load_dataset()
        elif args.dataset == 'Proteomics':
            dataset = proteomics.read_dataset(100)
        elif args.dataset == 'dconvert_class':
            dataset = dconvert.load_class_level_dataset()
        elif args.dataset == 'dconvert_method':
            dataset = dconvert.load_method_level_dataset()
        elif args.dataset == 'E2E':
            dataset = e2e.load_dataset()
        elif args.dataset == 'Sepsis':
            dataset = sepsis.load_dataset()
        elif args.dataset == 'Helpdesk':
            dataset = helpdesk.load_dataset()
        elif args.dataset == 'Covid19Germany':
            dataset = covid_germany.load_seven_day_incidence_county_dataset()
        elif args.dataset == 'RoadFines':
            dataset = traffic_fines.load_dataset()
        elif args.dataset == 'RollingMill':
            dataset = rolling_mill.load_dataset()
        elif args.dataset == 'artificial':
            dataset, true_model = self.__generate_artificial_dataset(
                args, repetition_index)
        elif args.dataset == 'sanity_check':
            generator = DatasetGenerator(
                nr_of_categorical_features=args.nr_of_categorical_features,
                nr_of_numerical_features=args.nr_of_numerical_features,
                nr_of_categories=args.nr_of_categories,
                nr_of_instances=args.size_of_artificial_dataset,
                random=Random(args.random_seed),
                nr_of_sequence_symbols=args.nr_of_sequence_symbols)
            dataset = generator.generate_with_completely_random_target_sequences()
        elif args.dataset == 'sanity_check2':
            generator = DatasetGenerator(
                nr_of_categorical_features=args.nr_of_categorical_features,
                nr_of_numerical_features=args.nr_of_numerical_features,
                nr_of_categories=args.nr_of_categories,
                nr_of_instances=args.size_of_artificial_dataset,
                random=Random(args.random_seed),
                nr_of_sequence_symbols=args.nr_of_sequence_symbols)
            dataset = generator.generate_with_shuffled_sequences()
        elif args.dataset.endswith('.arff'):
            with open(args.dataset, encoding='utf-8') as f:
                dataset = TargetSequenceDataset.create_from_arff(f.read(), 'sequence')
        else:
            raise NotImplementedError('unknown dataset: %s' % args.dataset)
        return dataset,true_model

    def __generate_artificial_dataset(
            self, args, repetition_index: int) -> Tuple[TargetSequenceDataset, Rule]:
        dataset_generator = DatasetGenerator(
            min_rule_length=parse_rule_list_length(args.min_rule_length),
            max_rule_length=parse_rule_list_length(args.max_rule_length),
            max_rule_depth=args.max_rule_depth,
            nr_of_categorical_features=args.nr_of_categorical_features,
            nr_of_numerical_features=args.nr_of_numerical_features,
            nr_of_categories=args.nr_of_categories,
            nr_of_instances=args.size_of_artificial_dataset,
            random=Random(args.random_seed),
            nr_of_sequence_symbols=args.nr_of_sequence_symbols,
            model=args.model,
            max_condition_depth=args.max_condition_depth,
            edge_probability=args.edge_probability,
            max_nr_of_nodes_in_model=args.max_nr_of_nodes_in_model
        )
        try:
            true_model = self.__cached_models[repetition_index]
            dataset = dataset_generator.generate_features_only_dataset()
            dataset = dataset_generator.apply_rule_to_dataset(dataset, true_model)
        except KeyError:
            dataset,true_model = dataset_generator.generate()
            if args.cache_artificial_models:
                self.__cached_models[repetition_index] = true_model
        return dataset, true_model
