# -*- coding: utf-8 -*-

import os
import argparse

from experimentlogger.experiment_logger import ExperimentLogger
from prolothar_common.experiments.stopwatch import Stopwatch

from prolothar_common.models.dataset import TargetSequenceDataset
from prolothar_common.models.dataset.instance import TargetSequenceInstance

from prolothar_common.models.labeled_transition_system.state_representation import set_abstraction
from prolothar_common.models.labeled_transition_system.state_representation import list_abstraction
from prolothar_common.models.labeled_transition_system.state_representation import last_abstraction
from prolothar_common.models.labeled_transition_system.state_representation import last_n_abstraction

from prolothar_rule_mining.rule_miner.data_to_sequence.dats import Dats

from experiments.rule_mining.dataset_loader import DatasetLoader
from experiments.rule_mining.noise_generator import NoiseGenerator
from experiments.rule_mining.experiment_utils import plot_and_save_true_model
from experiments.rule_mining.experiment_utils import create_computation_engine

class DatsRunner:

    def __init__(self):
        self.__logger: ExperimentLogger = None
        self.__dataset_loader = DatasetLoader()
        self.__noise_generator = NoiseGenerator()

    def run_from_cli(self):
        args = self.__parse_cli_arguments()

        self.__logger = ExperimentLogger(
            os.path.join('experiments', 'rule_mining', 'dats', 'results'),
            args.dataset)
        self.__logger.log(str(args))

        dataset, true_model = self.__dataset_loader.read_dataset(args, 0)
        if args.dataset == 'Sepsis':
            truncated_dataset = TargetSequenceDataset(
                dataset.get_categorical_attribute_names(),
                dataset.get_numerical_attribute_names())
            for instance in dataset:
                truncated_dataset.add_instance(TargetSequenceInstance(
                    instance.get_id(),
                    instance.get_features_dict(),
                    [s.replace('IV ', '') for s in instance.get_target_sequence()[1:-1]]
                ))
            dataset = truncated_dataset
        self.__noise_generator.apply(dataset, args)
        self.__logger.log('dataset contains %d instances' % len(dataset))

        if true_model is not None:
            plot_and_save_true_model(true_model, dataset, self.__logger)

        rule_miner = Dats(
            self.__create_state_representation_function(args),
            computation_engine=create_computation_engine(args),
            min_relative_sequence_frequency=args.min_relative_sequence_frequency
        )

        stopwatch = Stopwatch()
        stopwatch.start()

        rules = rule_miner.mine_rules(dataset)

        self.__logger.log('')
        self.__logger.log('elapsed time: %r' % stopwatch.get_elapsed_time())

        rules.get_transition_system().plot(
            view=False,
            filepath=self.__logger.get_absolute_path('transition_system'))
        rules.get_transition_system().plot(
            view=False, use_state_representation_for_labels=True,
            filepath=self.__logger.get_absolute_path('transition_system_with_labels'))

    def __create_state_representation_function(self, args):
        if args.state_representation == 'set':
            return set_abstraction
        elif args.state_representation == 'list':
            return list_abstraction
        elif args.state_representation == 'last':
            return last_abstraction
        elif args.state_representation == 'lastn':
            return last_n_abstraction(args.n)
        else:
            raise NotImplementedError(f'unknown state representation: {args.pruning}')

    def __parse_cli_arguments(self):
        parser = argparse.ArgumentParser('Boomer rule mining on a single dataset')

        parser.add_argument(
                '--trainset_size', type=int,
                help=('nr of instances in the trainset. default means maximal size'),
                required=False, default=-1)

        parser.add_argument(
                '--state_representation', type=str,
                help='state representation function',
                required=False, default='list', choices=['last', 'lastn', 'set', 'list'])

        parser.add_argument(
                '--n', type=int,
                help='n in the state representation function if applicable',
                required=False, default=5)

        parser.add_argument(
                '--min_relative_sequence_frequency', type=float,
                help='filter to reduce size of the transition system',
                required=False, default=0.0)

        parser.add_argument(
                '--nr_of_cpus', type=int,
                help=('number of cpus to use for parallel computations if '
                      'applicable. Default is to use all available cpus'),
                required=False, default=1)

        parser.add_argument(
                '--parallel_engine', type=str,
                help='name of the computation engine for parallel computations',
                required=False, default='ray',
                choices=['ray'])

        self.__dataset_loader.add_cli_arguments(parser)
        self.__noise_generator.add_cli_arguments(parser)

        return parser.parse_args()

if __name__ == '__main__':
    DatsRunner().run_from_cli()
