import os
import pickle

import psutil

from dhparallel.engine.abstract.computation_engine import ComputationEngine
from dhparallel.engine.multiprocess.multiprocess import MultiprocessComputationEngine
from dhparallel.engine.single_thread.single_thread import SingleThreadComputationEngine
from dhparallel.engine.ray.ray import RayComputationEngine

from prolothar_rule_mining.models.event_flow_graph import EventFlowGraph
from prolothar_rule_mining.models.event_flow_graph.cover import compute_cover

def plot_and_save_true_model(true_model, trainset, logger):
    if hasattr(true_model, 'to_string'):
        logger.log('save true_model to text file')
        with logger.open_text_file('generated_model.txt') as f:
            f.write(true_model.to_string())
        logger.log('saved true_model to text file')

        event_flow_graph = true_model.to_eventflow_graph()
    else:
        event_flow_graph = true_model.get_event_flow_graph()

        for node, router in true_model.get_node_router_table().items():
            with logger.open_text_file('generated_model_%d.txt' % node.node_id) as f:
                f.write(router.get_rule().to_string())

    plot_and_save_eventflow_graph(event_flow_graph, trainset, 'generated_graph', logger)

def plot_and_save_eventflow_graph(
        graph: EventFlowGraph, trainset, graph_name: str, logger):
    logger.log('plot %s' % graph_name)
    graph.plot(show=False, filepath=os.path.join(
            logger.get_path_to_experiment_directory(), graph_name))

    logger.log('plot %s with weighted edges' % graph_name)
    compute_cover(trainset, graph, assign_instances_to_edges=True)
    for edge in graph.edges():
        edge.attributes['nr_of_instances'] = len(edge.attributes['instances'])
    graph.plot(show=False, filepath=os.path.join(
            logger.get_path_to_experiment_directory(),
            graph_name + '_with_weighted_edges'),
            edge_weight_attribute='nr_of_instances',
            edge_label_attribute='nr_of_instances')

    logger.log('plot %s with weighted edges and ids' % graph_name)
    graph.plot(show=False, filepath=os.path.join(
            logger.get_path_to_experiment_directory(),
            graph_name + '_with_weighted_edges_and_ids'),
            edge_weight_attribute='nr_of_instances',
            edge_label_attribute='nr_of_instances',
            with_node_ids=True)

    logger.log('plot %s with counts on edges and ids' % graph_name)
    graph.plot(show=False, filepath=os.path.join(
            logger.get_path_to_experiment_directory(),
            graph_name + '_with_counts_edges_and_ids'),
            edge_label_attribute='nr_of_instances',
            with_node_ids=True)

    logger.log('save %s to json file' % graph_name)
    with logger.open_text_file('%s.json' % graph_name) as f:
        f.write(graph.to_json())

    logger.log('save %s to pickle file' % graph_name)
    with logger.open_binary_file('graph.pickle') as f:
        pickle.dump(graph, f)

    logger.log('plotted and saved %s' % graph_name)

def create_computation_engine(args) -> ComputationEngine:
    if args.nr_of_cpus == 1 or psutil.cpu_count() == 1:
        return SingleThreadComputationEngine()
    elif args.parallel_engine == 'multiprocessing':
        if args.nr_of_cpus < 1:
            return MultiprocessComputationEngine()
        else:
            return MultiprocessComputationEngine(nr_of_workers=args.nr_of_cpus)
    else:
        return RayComputationEngine()
