# -*- coding: utf-8 -*-

import os
import argparse
import psutil

import pickle

from experimentlogger.experiment_logger import ExperimentLogger
from prolothar_common.experiments.stopwatch import Stopwatch

from prolothar_rule_mining.models.event_flow_graph.alignment.a_star import AStar
from prolothar_rule_mining.models.event_flow_graph.alignment.heuristics import NullHeuristic
from prolothar_rule_mining.models.event_flow_graph.router.learning import RuleClassifierRouterLearner
from prolothar_rule_mining.rule_miner.classification.myra import CAntMinerPB
from prolothar_rule_mining.rule_miner.classification.rce import ReliableRuleMiner
from prolothar_rule_mining.rule_miner.classification.rce.strategy.beam_search import BeamSearch as ConditionBeamSearch
from prolothar_rule_mining.rule_miner.classification.classy import Classy
from prolothar_rule_mining.rule_miner.classification.fdd_rule_miner import FddRuleMiner
from prolothar_rule_mining.models.event_flow_graph.alignment.beam_search import BeamSearch
from prolothar_rule_mining.models.event_flow_graph.alignment.petrinet import PetrinetAligner
from prolothar_rule_mining.models.event_flow_graph.alignment.heuristics import ReachabilityHeuristic
from prolothar_rule_mining.models.event_flow_graph.alignment.greedy_shortest_path import GreedyShortestPath
from prolothar_rule_mining.models.event_flow_graph.alignment.path_enumeration import PathEnumeration
from prolothar_rule_mining.rule_miner.data_to_sequence.rudi.event_flow_graph_miner import SequenceBottomUpEventFlowGraphMiner
from prolothar_rule_mining.rule_miner.data_to_sequence.rudi.event_flow_graph_miner import SequenceTopDownEventFlowGraphMiner
from prolothar_rule_mining.rule_miner.data_to_sequence.rudi.event_flow_graph_miner import EventFlowGraphMiner
from prolothar_rule_mining.rule_miner.data_to_sequence.rudi.rudi import RuDi

from experiments.rule_mining.dataset_loader import DatasetLoader
from experiments.rule_mining.noise_generator import NoiseGenerator
from experiments.rule_mining.experiment_utils import plot_and_save_true_model
from experiments.rule_mining.experiment_utils import plot_and_save_eventflow_graph
from experiments.rule_mining.experiment_utils import create_computation_engine

class RudiRunner:

    def __init__(self):
        self.__logger: ExperimentLogger = None
        self.__dataset_loader = DatasetLoader()
        self.__noise_generator = NoiseGenerator()

    def run_from_cli(self):
        args = self.__parse_cli_arguments()

        self.__logger = ExperimentLogger(
            os.path.join('experiments', 'rule_mining', 'rudi', 'results'),
            args.dataset)
        self.__logger.log(str(args))

        dataset, true_model = self.__dataset_loader.read_dataset(args, 0)
        if args.dataset.endswith('.arff'):
            trainset = dataset
        else:
            trainset,testset = dataset.split(
                args.testset_proportion, random_seed=args.random_seed)
        if args.trainset_size > 0:
            trainset = trainset.random_subset(args.trainset_size,
                                              random_seed=args.random_seed)
        self.__noise_generator.apply(trainset, args)
        self.__logger.log('dataset contains %d instances' % len(dataset))
        self.__logger.log('trainset contains %d instances' % len(trainset))
        if not args.dataset.endswith('.arff'):
            self.__logger.log('testset contains %d instances' % len(testset))

        if true_model is not None:
            plot_and_save_true_model(true_model, trainset, self.__logger)

        event_flow_graph_miner = self.__create_event_flow_graph_miner(args)

        rudi = RuDi(
            logger=self.__logger.log,
            event_flow_graph_miner=event_flow_graph_miner,
            router_learner=self.__create_router_learner(args),
            computation_engine=create_computation_engine(args))
        self.__logger.log('run %r' % rudi)

        stopwatch = Stopwatch()
        stopwatch.start()

        mined_graph = event_flow_graph_miner.mine_event_flow_graph(trainset)

        self.__logger.log('')
        self.__logger.log('elapsed time: %r' % stopwatch.get_elapsed_time())


        self.__logger.log('infer rules for decision points')
        rules = rudi.infer_rules_from_event_flow_graph(mined_graph, trainset)
        self.__logger.log('save inferred rules')
        for node, router in rules.get_node_router_table().items():
            with self.__logger.open_text_file('mined_model_%d.txt' % node.node_id) as f:
                f.write(router.get_rule().to_string())
        with self.__logger.open_binary_file('mined_model.pickle') as f:
            pickle.dump(rules, f)

        plot_and_save_eventflow_graph(mined_graph, trainset, 'mined_graph', self.__logger)

    def __create_event_flow_graph_miner(self, args) -> EventFlowGraphMiner:
        if args.event_flow_graph_miner == 'bottomup':
            return SequenceBottomUpEventFlowGraphMiner(
                logger=self.__logger.log,
                alignment_finder_factory_model_extension=self.__create_alignment_finder_factory(
                    args.alignment_finder_extension, args),
                alignment_finder_factory_score=self.__create_alignment_finder_factory(
                    args.alignment_finder_score, args),
                patience=args.event_flow_graph_miner_patience)
        elif args.event_flow_graph_miner == 'topdown':
            return SequenceTopDownEventFlowGraphMiner(
                logger=self.__logger.log,
                alignment_finder_factory_score=self.__create_alignment_finder_factory(
                    args.alignment_finder_score, args),
                patience=args.event_flow_graph_miner_patience)
        elif args.event_flow_graph_miner == 'topdown_fast':
            return SequenceTopDownEventFlowGraphMiner(
                logger=self.__logger.log, exact_mdl=False,
                alignment_finder_factory_score=self.__create_alignment_finder_factory(
                    args.alignment_finder_score, args),
                patience=args.event_flow_graph_miner_patience)
        else:
            raise NotImplementedError(args.event_flow_graph_miner)


    def __create_alignment_finder_factory(self, alignment_finder_name: str, args):
        if alignment_finder_name == 'greedy':
            return lambda g: GreedyShortestPath(g, ReachabilityHeuristic(g))
        elif alignment_finder_name == 'beamsearch':
            return lambda g: BeamSearch(
                g, ReachabilityHeuristic(g), args.alignment_finder_beam_width)
        elif alignment_finder_name == 'astar':
            return lambda g: AStar(g, NullHeuristic())
        elif alignment_finder_name == 'petrinet':
            return PetrinetAligner
        elif alignment_finder_name == 'pathenum':
            return PathEnumeration
        else:
            raise NotImplementedError(alignment_finder_name)

    def __create_router_learner(self, args):
        classifier = {
            'Classy': Classy(),
            'CAntMinerPB': CAntMinerPB(nr_of_workers=1),
            'ReliableRuleMiner': ReliableRuleMiner(
                logger=None if args.nr_of_cpus > 1 else self.__logger.log,
                search_strategy=ConditionBeamSearch(
                    max_nr_of_base_candidates=args.max_nr_of_base_conditions,
                    beam_width=args.beam_width_condition_search))
        }[args.classifier]

        if args.feature_selector == 'FDD':
            classifier = FddRuleMiner(classifier)
        elif args.feature_selector is not None:
            raise NotImplementedError(args.feature_selector)

        return RuleClassifierRouterLearner(classifier)

    def __parse_cli_arguments(self):
        parser = argparse.ArgumentParser('Data2Seq rule mining on a single dataset')

        parser.add_argument(
                '--trainset_size', type=int,
                help=('nr of instances in the trainset. default means maximal size'),
                required=False, default=-1)

        parser.add_argument(
                '--testset_proportion', type=float,
                help='proportion in (0,1) of the testset',
                required=False, default=0.2)

        parser.add_argument(
                '--alignment_finder_beam_width', type=int,
                help='beam width if alignment_finder_score == beamsearch',
                required=False, default=10)

        parser.add_argument(
                '--alignment_finder_score', type=str,
                help=('alignment finder algorithm to compute the MDL score for '
                      'eventflow graph extension candidates'),
                required=False, default='beamsearch',
                choices=['greedy', 'beamsearch', 'astar', 'petrinet'])

        parser.add_argument(
                '--alignment_finder_extension', type=str,
                help=('alignment finder algorithm to compute extension points for '
                      'eventflow graph extension candidates'),
                required=False, default='petrinet',
                choices=['greedy', 'beamsearch', 'astar', 'petrinet', 'pathenum'])

        parser.add_argument(
                '--classifier', type=str,
                help=('classifier algorithm used at split points'),
                required=False, default='ReliableRuleMiner',
                choices=['ReliableRuleMiner', 'Classy', 'CAntMinerPB'])

        parser.add_argument(
                '--max_nr_of_base_conditions', type=int,
                help=('if classifier is ReliableRuleMiner, the number of base '
                      'candidates to keep for extension during search. '
                      'default means no limitation. the limitation is useful '
                      'if some categorical attributes have many different values '
                      'such as an ID.'),
                required=False, default=-1)

        parser.add_argument(
                '--beam_width_condition_search', type=int,
                help=('if classifier is ReliableRuleMiner, the width parameter '
                      'for the beam search.'),
                required=False, default=5)

        parser.add_argument(
                '--feature_selector', type=str,
                help=('feature selector run before the classifier at split points'),
                required=False, default=None,
                choices=['FDD'])

        parser.add_argument(
                '--event_flow_graph_miner', type=str,
                help=('event flow graph miner'),
                required=False, default='bottomup',
                choices=['bottomup', 'topdown', 'topdown_fast'])
        parser.add_argument(
                '--event_flow_graph_miner_patience', type=int,
                help=('patience in event flow graph miner algorithm'),
                required=False, default=10)

        parser.add_argument(
                '--nr_of_cpus', type=int,
                help=('number of cpus to use for parallel computations if '
                      'applicable. Default is to use all available cpus'),
                required=False, default=len(psutil.Process().cpu_affinity()))

        parser.add_argument(
                '--parallel_engine', type=str,
                help='name of the computation engine for parallel computations',
                required=False, default='multiprocessing',
                choices=['ray', 'multiprocessing'])

        self.__dataset_loader.add_cli_arguments(parser)
        self.__noise_generator.add_cli_arguments(parser)

        return parser.parse_args()

if __name__ == '__main__':
    RudiRunner().run_from_cli()
