"""
ffd = functional dependency discovery
"""
from typing import List, Set, Dict

from abc import ABC, abstractmethod
import os
import sys
import tempfile
import pathlib
import subprocess
import re

from prolothar_common.models.dataset import Dataset

HERE = pathlib.Path(__file__).parent

REGEX_DISCRETIZATION_SUFFIX_ATTRIBUTE_NAME = re.compile(r'_in_.*_(supEq|eq)FreqBins')

class FunctionalPattern:
    """
    one result of a FoDiscovery, i.e. one functional dependency pattern
    """
    def __init__(
            self, attributes: Set[str], target: str, metrics: Dict[str, float]):
        self.attributes = attributes
        self.target = target
        self.metrics = metrics

    def __repr__(self) -> str:
        return 'FunctionalPattern(%r,%r,%r)' %  (
            self.attributes, self.target, self.metrics)

class FoDiscovery(ABC):
    """
    python interface to
    https://github.com/pmandros/fodiscovery/
    """

    def __init__(
            self, nr_of_bins: int, bounding_function: str, nr_of_results: int,
            verbose: bool):
        self.__nr_of_bins = nr_of_bins
        self.__bounding_function = bounding_function
        self.__nr_of_results = nr_of_results
        self.__verbose = verbose

    def run(self, dataset: Dataset, target_attribute: str) -> List[FunctionalPattern]:
        with tempfile.TemporaryDirectory() as temporary_directory:
            self.__run_java_program(
                dataset, target_attribute, temporary_directory)

            output_files = os.listdir(temporary_directory)
            #no functional dependencies found
            if len(output_files) != 1:
                return []

            attribute_names = (dataset.get_categorical_attribute_names() +
                               dataset.get_numerical_attribute_names())
            with open(os.path.join(temporary_directory, output_files[0])) as f:
                return [
                    self.__parse_functional_pattern(line, attribute_names)
                    for line in f.read().splitlines()
                ]

    def __parse_functional_pattern(
            self, line: str, attribute_names: List[str]) -> FunctionalPattern:
        #FunctionalPatternImplementation(([XVORMACO, XQUETSOL],[sequence]), [Co-domain entropy: 0.35081592989560484,mutual inf.: 0.20474446818509318,expected mutual information: 0.031232765084017622,fraction of information: 0.5836236348959999,reliable fraction of information: 0.49459470997428434])

        line = line.lstrip('FunctionalPatternImplementation((').rstrip(')')
        #[XVORMACO, XQUETSOL],[sequence]), [Co-domain entropy: 0.35081592989560484,mutual inf.: 0.20474446818509318,expected mutual information: 0.031232765084017622,fraction of information: 0.5836236348959999,reliable fraction of information: 0.49459470997428434]
        pattern_string, metrics_string = line.split('), ', maxsplit=1)

        #[XVORMACO, XQUETSOL],[sequence]
        attributes, target = pattern_string.lstrip('[').rstrip(']').split('],[')
        target = self.__remove_discretization_suffix_from_attribute_name(target)
        attributes = attributes.split(', ')
        attributes = set(self.__remove_discretization_suffix_from_attribute_name(a) for a in attributes)
        attributes = set(self.__map_attribute_name(attribute, attribute_names)
                         for attribute in attributes)

        #[Co-domain entropy: 0.35081592989560484,mutual inf.: 0.20474446818509318,expected mutual information: 0.031232765084017622,fraction of information: 0.5836236348959999,reliable fraction of information: 0.49459470997428434]
        metrics = {}
        for metric_string in metrics_string.lstrip('[').rstrip(']').split(','):
            metric_name, metric_value = metric_string.split(': ')
            metric_value = float(metric_value)
            metrics[metric_name] = metric_value

        return FunctionalPattern(attributes, target, metrics)

    def __remove_discretization_suffix_from_attribute_name(self, attribute_name: str) -> str:
        return REGEX_DISCRETIZATION_SUFFIX_ATTRIBUTE_NAME.sub('', attribute_name)

    def __map_attribute_name(self, attribute: str, attribute_names: List[str]) -> str:
        """
        sometimes, it happens that FoDiscovery outputs Attribute1 or Attribute%d
        as an attribute name. in these cases, we have to remap the index to a
        proper attribute name
        """
        if re.match(r'Attribute_\d+', attribute):
            return attribute_names[int(attribute.lstrip('Attribute_'))]
        else:
            return attribute

    def __run_java_program(self,
            dataset: Dataset, target_attribute: str, temporary_directory: str):
        path_to_dataset = os.path.join(temporary_directory, 'dataset.arff')

        with open(path_to_dataset, 'w') as f:
            arff = dataset.export_to_arff(relation_name='Dataset')
            f.write(arff)

        arguments = [
            'java', '-cp',
            os.path.join(HERE, 'FoDiscovery.jar'),
            self._get_java_class(),
            '-DATASET', path_to_dataset,
            '-OUTPUTFOLDER', temporary_directory,
            '-L', str(self.__nr_of_bins),
            '-OPT', self.__bounding_function,
            '-TARGET', str(self.get_arff_feature_index(target_attribute, dataset)),
            '-K', str(self.__nr_of_results)
        ] + self._get_additional_cli_parameters()
        if self.__verbose:
            print('cli arguments: %r' % arguments)
        with subprocess.Popen(
                arguments, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                universal_newlines=True) as java_process:
            while java_process.stdout.readable():
                line = java_process.stdout.readline()
                if not line:
                    break
                if self.__verbose:
                    sys.stdout.write(line)

        os.remove(path_to_dataset)

    def get_arff_feature_index(self, target_attribute: str, dataset: Dataset) -> int:
        if target_attribute in ['sequence', 'class']:
            return dataset.get_nr_of_attributes() + 1
        else:
            return (dataset.get_categorical_attribute_names() +
                    dataset.get_numerical_attribute_names()).index(target_attribute) + 1

    @abstractmethod
    def _get_additional_cli_parameters(self) -> list:
        pass

    @abstractmethod
    def _get_java_class(self) -> str:
        pass

class GreedyFoDiscovery(FoDiscovery):
    """
    greedy implementation of FoDiscovery
    de.mpiinf.fodiscovery.singleexp.FoBeam
    """

    def __init__(
            self, nr_of_bins: int = 5, beam_width: int = 1, verbose: bool = False,
            bounding_function: str = 'CHAIN', nr_of_results: int = 1):
        super().__init__(nr_of_bins, bounding_function, nr_of_results, verbose)
        self.__beam_width = beam_width

    def _get_additional_cli_parameters(self) -> list:
        return ['-BEAMWIDTH', str(self.__beam_width)]

    def _get_java_class(self) -> str:
        return 'de.mpiinf.fodiscovery.singleexp.FoBeam'

class BranchAndBoundFoDiscovery(FoDiscovery):
    """
    Branch and Bound implementation of FoDiscovery that yields an optimal
    solution but can have exponential runtime
    de.mpiinf.fodiscovery.singleexp.FoBeam
    """

    def __init__(
            self, nr_of_bins: int = 5, alpha: float = 1, verbose: bool = False,
            bounding_function: str = 'CHAIN', nr_of_results: int = 1):
        super().__init__(nr_of_bins, bounding_function, nr_of_results, verbose)
        self.__alpha = alpha

    def _get_additional_cli_parameters(self) -> list:
        return ['-ALPHA', str(self.__alpha)]

    def _get_java_class(self) -> str:
        return 'de.mpiinf.fodiscovery.singleexp.FoOPUS'