# -*- coding: utf-8 -*-

"""
interface to the Classy Rule Mining Algorithm
See https://github.com/HMProenca/MDLRuleLists
"""

from typing import FrozenSet, Dict, Tuple, List
import re
import numpy as np
import pandas as pd
from rulelist import RuleList

from prolothar_common.models.dataset import ClassificationDataset
from prolothar_common.models.dataset.attributes import Attribute

from prolothar_common.models.dataset.transformer.one_hot_encoding import OneHotEncoding
from prolothar_common.models.dataset.transformer.quantile_based_discretization import QuantileBasedDiscretization
from prolothar_common.models.dataset.transformer.remove_attributes_with_one_unique_value import RemoveAttributesWithOneUniqueValue

from prolothar_rule_mining.rule_miner.classification.rules import Rule
from prolothar_rule_mining.rule_miner.classification.rules import FirstFiringRuleModel
from prolothar_rule_mining.rule_miner.classification.rules import ListOfRules
from prolothar_rule_mining.rule_miner.classification.rules import IfThenElseRule
from prolothar_rule_mining.rule_miner.classification.rules import ReturnClassRule

from prolothar_rule_mining.models.conditions import Condition
from prolothar_rule_mining.models.conditions import AndCondition
from prolothar_rule_mining.models.conditions import OrCondition
from prolothar_rule_mining.models.conditions import EqualsCondition
from prolothar_rule_mining.models.conditions import LessThanCondition
from prolothar_rule_mining.models.conditions import GreaterOrEqualCondition

condition_types = {
    '=': EqualsCondition,
    '>=': GreaterOrEqualCondition,
    '<': LessThanCondition
}
operator_types = {
    'AND': AndCondition,
    'OR': OrCondition
}

class Classy():
    """interface to the Classy Rule Mining Algorithm"""

    def __init__(self, min_support: int = 5,
                 verbose: bool = False):
        """
        creates and configures a new instance of the Classy algorithm

        Parameters
        ----------
        nr_of_bins_for_c_attributes : int, optional
            nr of bins for the discretization of continuous attributes.
            The default is 5.
        min_support : int, optional
            minimum support a rule must have. The default is 5.

        Returns
        -------
        None.

        """
        self.__min_support = min_support
        self.__verbose = verbose

    def mine_rules(self, dataset: ClassificationDataset) -> Rule:
        """create a rule to predict classes in the data set """
        if self.__verbose:
            print('prepare dataset')
        transformed_dataset = RemoveAttributesWithOneUniqueValue().transform(dataset)

        #there were either no attributes right from the beginning or there were
        #only attributes with one unique value which were filtered out. anyway,
        #a non-empty model does not make sense. there is no information in the
        #the attributes on the target
        if transformed_dataset.get_nr_of_attributes() == 0 or len(dataset.get_set_of_classes()) < 2:
            return FirstFiringRuleModel(ListOfRules())

        dataset_df = transformed_dataset.to_dataframe()
        for categorical_attribute in transformed_dataset.get_categorical_attribute_names():
            dataset_df[categorical_attribute] = dataset_df[categorical_attribute].astype('object')
        y = pd.Series([instance.get_class() for instance in transformed_dataset]).astype('object')

        if self.__verbose:
            print('run classy')

        rule_list = RuleList(
            target_model='categorical', task='prediction',
            min_support=self.__min_support)
        rule_list.fit(dataset_df, y)

        if self.__verbose:
            print('map classy model to rule model')
        return self.__map_classy_model(str(rule_list), dataset)

    def __map_classy_model(
            self, model_str: str,
            dataset: ClassificationDataset) -> Rule:
        model = ListOfRules()
        for line in model_str.splitlines():
            if 'THEN' in line:
                model.append_rule(self.__parse_classy_rule(line.strip(), dataset))
        #make sure that there is at least one class (majority class) predicted
        if len(model) == 0 or isinstance(model.get_list()[-1], IfThenElseRule):
            model.append_rule(ReturnClassRule(max(
                dataset.get_set_of_classes(),
                key=dataset.get_class_count)))
        return FirstFiringRuleModel(model)

    def __parse_classy_rule(self, line: str, dataset: ClassificationDataset) -> Rule:
        #the default rule does not have a condition
        if line.startswith('ELSE  usage'):
            return ReturnClassRule(self.__parse_class_label(line))

        condition_str, consequence_str = line.replace('ELSE IF ', '').replace('If ', '').split('  THEN  ')
        return IfThenElseRule(
            self.__parse_classy_condition(condition_str, dataset),
            if_branch=ListOfRules([
                ReturnClassRule(self.__parse_class_label(consequence_str))
            ])
        )

    def __parse_classy_condition(self, condition_str: str, dataset: ClassificationDataset) -> Condition:
        subcondition_strings = re.split(r'  (AND|OR)  ', condition_str)
        if len(subcondition_strings) > 1:
            return self.__parse_complex_classy_condition(subcondition_strings, dataset)

        operator_split = re.split(r' (=|>=|<=|<) ', condition_str)
        #e.g. "size > 20"
        if len(operator_split) == 3:
            attribute_name, operator, value = operator_split
            if operator != '=':
                value = float(value)
            return condition_types[operator](dataset.get_attribute_by_name(attribute_name), value)
        #e.g. "20 <= size < 30"
        elif len(operator_split) == 5:
            lower_value, operator_1, attribute_name, operator_2, upper_value = operator_split
            if operator_1 != '<=' or operator_2 != '<':
                raise NotImplementedError(condition_str)
            return AndCondition([
                GreaterOrEqualCondition(
                    dataset.get_attribute_by_name(attribute_name), float(lower_value)),
                LessThanCondition(
                    dataset.get_attribute_by_name(attribute_name), float(upper_value))
            ])
        else:
            raise NotImplementedError(operator_split)

    def __parse_complex_classy_condition(
            self, subcondition_strings: List[str],
            dataset: ClassificationDataset) -> Condition:
        operator = subcondition_strings[1]
        for i in range(3, len(subcondition_strings), 2):
            if subcondition_strings[i] != operator:
                raise NotImplementedError(subcondition_strings)
        return operator_types[operator]([
            self.__parse_classy_condition(subcondition_strings[i], dataset)
            for i in range(0, len(subcondition_strings), 2)
        ])

    def __parse_class_label(self, consequence_str: str):
        #'usage = 23 : target = 0Pr(0) = 0.0;Pr(1) = 0....;'
        probability_per_label = {}
        for label_term in consequence_str.split(' : ')[1][len('target = 0'):-1].split(';'):
            #Pr(label) = 0.54
            label, probability = label_term.split(' = ')
            label = label[3:-1]
            probability = float(probability)
            probability_per_label[label] = probability
        return max(probability_per_label, key=probability_per_label.get)

    def is_output_order_invariant(self) -> bool:
        return False