# -*- coding: utf-8 -*-

"""
interface to the MYRA Ant Colony Optimization algorithms.
See https://github.com/febo/myra
"""

from abc import ABC, abstractmethod

import tempfile
import os
import subprocess
from subprocess import TimeoutExpired
import psutil
import pathlib

from prolothar_common.models.dataset import Dataset

from prolothar_rule_mining.rule_miner.classification.rules import Rule
from prolothar_rule_mining.rule_miner.classification.rules import FirstFiringRuleModel
from prolothar_rule_mining.rule_miner.classification.rules import ListOfRules
from prolothar_rule_mining.rule_miner.classification.rules import IfThenElseRule
from prolothar_rule_mining.rule_miner.classification.rules import ReturnClassRule
from prolothar_rule_mining.models.conditions import Condition
from prolothar_rule_mining.models.conditions import AndCondition
from prolothar_rule_mining.models.conditions import EqualsCondition
from prolothar_rule_mining.models.conditions import LessOrEqualCondition
from prolothar_rule_mining.models.conditions import GreaterThanCondition

HERE = pathlib.Path(__file__).parent

class _AbstractCAntminer():
    """template for implementations of AntMiner"""

    def __init__(
            self, nr_of_workers: int = 1, random_seed: int = None,
            timeout_in_seconds: int = 3600,
            max_java_heapspace_in_mb: int = -1):
        self.__nr_of_workers = nr_of_workers
        self.__random_seed = random_seed
        self.__timeout_in_seconds = timeout_in_seconds
        if max_java_heapspace_in_mb <= 0:
            self.__max_java_heapspace_in_mb = int(
                psutil.virtual_memory().available / 1024 / 1024)
        else:
            self.__max_java_heapspace_in_mb = max_java_heapspace_in_mb

    def mine_rules(
            self, dataset: Dataset, verbose: bool = False) -> IfThenElseRule:
        for attribute_name in dataset.get_categorical_attribute_names():
            for value in dataset.get_attribute_by_name(
                    attribute_name).get_unique_values():
                if isinstance(value, str) and (',' in value or ' ' in value):
                    raise NotImplementedError(
                        'attribute %r contains values with "," or " "' %
                        attribute_name)

        with tempfile.TemporaryDirectory() as temporary_directory:
            arff_file = os.path.join(temporary_directory, 'dataset.arff')
            with open(arff_file, 'w') as f:
                f.write(dataset.export_to_arff(
                    target_attribute_name='class',
                    sequence_element_separator=';'))

            arguments = [
                'java', '-cp',
                os.path.join(HERE, 'target', 'myra-4.6.jar'),
                '-Xmx%dm' % self.__max_java_heapspace_in_mb,
                self._get_java_class(),
                '-f', arff_file,
                '--parallel', str(self.__nr_of_workers)
            ]
            if verbose:
                print('myra cli arguments: %r' % arguments)
            if self.__random_seed is not None:
                arguments.extend(['-s', self.__random_seed])
            try:
                myra_process = subprocess.run(
                    arguments, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                    universal_newlines=True, timeout=self.__timeout_in_seconds)
            except TimeoutExpired as e:
                print(e.stdout)
                print(e.stderr)
                raise e
        if verbose:
            print(myra_process.stdout)
            print(myra_process.stderr)
        if myra_process.returncode != 0:
            print(myra_process.stdout)
            print(myra_process.stderr)
            print(temporary_directory)
            raise Exception('myra_process returned exit code %d' %
                            myra_process.returncode)

        return self.__parse_rules_from_stddout(myra_process.stdout, dataset)

    @abstractmethod
    def _get_java_class(self) -> str:
        pass

    def __parse_rules_from_stddout(
            self, stddout: str, dataset: Dataset) -> IfThenElseRule:
       #"2:" because we want remove "===" and empty line at the beginning
       lines = stddout[
            stddout.index('=== Discovered Model ==='):
            stddout.index('Number of rules:')].strip().split('\n')[2:]

       parsed_rule = ListOfRules()
       for line in lines:
            parsed_rule.append_rule(self.__parse_rule_from_line(line, dataset))

       return FirstFiringRuleModel(parsed_rule)

    def is_output_order_invariant(self) -> bool:
        return False

    def __parse_rule_from_line(
            self, line: str, dataset: Dataset) -> Rule:
        #remove "IF " at the beginning
        if not line.startswith('IF'):
            raise ValueError('no if: ' + line)
        line = line[3:]
        condition_string, sequence_string = line.split(' THEN ')
        if condition_string == '<empty>':
            return self.__parse_sequence_rule(sequence_string)
        else:
            return IfThenElseRule(
                self.__parse_condition(condition_string, dataset),
                ListOfRules([self.__parse_sequence_rule(sequence_string)]))

    def __parse_condition(
            self, condition_string: str, dataset: Dataset) -> Condition:
        condition_string = condition_string.replace('" "', '""')
        split = condition_string.split(' AND ')
        if len(split) > 1:
            return AndCondition([
                self.__parse_condition(c, dataset) for c in split])
        split = condition_string.split()
        if len(split) == 3:
            #<attribute> <operator> <value> - example: size > 42
            return self.__parse_simple_operator_condition(
                dataset, split[0], split[1], split[2])
        elif len(split) == 5:
            #<value> <operator> <attribute> <operator> <value>
            #example: 40 < size <= 52
            return self.__parse_range_condition(
                dataset, split[0], split[1], split[2], split[3], split[4])
        else:
            raise NotImplementedError(condition_string)

    def __parse_range_condition(
            self, dataset: Dataset, lower_value: str, lower_operator: str,
            attribute_name: str, upper_operator: str,
            upper_value: str) -> Condition:
        attribute = dataset.get_attribute_by_name(attribute_name.strip('"'))
        return AndCondition([
            self.__condition_constructor(lower_operator)(
                attribute, self.__parse_value(lower_value)),
            self.__condition_constructor(upper_operator)(
                attribute, self.__parse_value(upper_value))
        ])

    def __parse_simple_operator_condition(
            self, dataset: Dataset, attribute: str, operator: str,
            value: str) -> Condition:
        attribute = dataset.get_attribute_by_name(attribute.strip('"'))
        value = self.__parse_value(value)
        return self.__condition_constructor(operator)(attribute, value)

    def __condition_constructor(self, operator: str):
        if operator == '=':
            condition = EqualsCondition
        elif operator == '<':
            condition = LessOrEqualCondition
        elif operator == '<=':
            condition = LessOrEqualCondition
        elif operator == '>':
            condition = GreaterThanCondition
        else:
            raise NotImplementedError('unknown operator: %s' % operator)
        return condition

    def __parse_value(self, value):
        if value.startswith('"'):
            value = value[1:-1]
            if value == 'False':
                value = False
            if value == 'True':
                value = True
        else:
            value = value.replace(',', '.')
            if '.' in value:
                value = float(value)
            else:
                value = int(value)
        return value

    def __parse_sequence_rule(self, sequence_string: str) -> ReturnClassRule:
        return ReturnClassRule(sequence_string.strip('"'))

class CAntMiner(_AbstractCAntminer):
    """
    An extension of Ant-Miner to cope with continuous attributes.
    It works essentially the same as Ant-Miner, but continuous attributes
    undergo a dynamic discretisation process when selected to be included
    in a rule.
    """

    def __init__(
            self, random_seed: int = None,
            nr_of_workers: int = len(psutil.Process().cpu_affinity())):
        """
        creates and configures a new CAntMiner intance

        Parameters
        ----------
        random_seed : int, optional
            can be used to fix random generators in the stochastic algorithm.
            Stable results are only achieved if nr_of_workers = 1.
            See https://github.com/febo/myra for details.
            The default is None (current system time).
        nr_of_workers : int, optional
            nr of CPU cores the miner is allowed to use. Use 1 for single
            thread execution. The default is len(os.sched_getaffinity(0)).
        """
        super().__init__(random_seed=random_seed, nr_of_workers=nr_of_workers)

    def _get_java_class(self) -> str:
        return 'myra.classification.rule.impl.cAntMiner'

class CAntMinerPB(_AbstractCAntminer):
    """
    cAnt-MinerPB incorporates a new strategy to discover a list of
    classification rules, which guides the search performed by the ACO
    algorithm using the quality of a candidate list of rules, instead of a
    single rule. The main motivation is to avoid the problem of rule
    interaction derived from the order in which the rules are discovered —
    i.e., the outcome of a rule affects the rules that can be discovered
    subsequently since the search space is modified due to the removal of
    examples covered by previous rules.
    """

    def __init__(
            self, random_seed: int = None,
            nr_of_workers: int = len(psutil.Process().cpu_affinity()),
            max_java_heapspace_in_mb: int = -1):
        """
        creates and configures a new CAntMinerPB intance

        Parameters
        ----------
        random_seed : int, optional
            can be used to fix random generators in the stochastic algorithm.
            Stable results are only achieved if nr_of_workers = 1.
            See https://github.com/febo/myra for details.
            The default is None (current system time).
        nr_of_workers : int, optional
            nr of CPU cores the miner is allowed to use. Use 1 for single
            thread execution. The default is len(os.sched_getaffinity(0)).
        max_java_heapspace_in_mb: int, optional
            used for the Xmx parameter of the java executable to run myra.jar.
            Default is the available RAM.
        """
        super().__init__(random_seed=random_seed, nr_of_workers=nr_of_workers,
                         max_java_heapspace_in_mb=max_java_heapspace_in_mb)

    def _get_java_class(self) -> str:
        return 'myra.classification.rule.impl.cAntMinerPB'
