from typing import Tuple, List
from collections import Counter, defaultdict
from dhcollections.list_utils import iterate_self_product

from frozendict import frozendict

from ortools.linear_solver.pywraplp import Solver

from prolothar_rule_mining.rule_miner.data_to_sequence.multiset_orderer.abstract_ordering import AbstractOrdering

from prolothar_common.models.dataset import MultisetDataset
from prolothar_common.models.dataset import TargetSequenceDataset

class LpOrdering(AbstractOrdering):

    def __init__(self, solver_name: str = 'GLOP'):
        self.__label_position_count = defaultdict(Counter)
        self.__order_cache = {}
        self.__solver_name = solver_name

    def train(self, multiset_dataset: MultisetDataset, sequence_dataset: TargetSequenceDataset):
        for instance in sequence_dataset:
            for i, label in enumerate(instance.get_target_sequence()):
                self.__label_position_count[label][i] += 1

    def order(self, multiset: Counter) -> Tuple:
        solver = Solver.CreateSolver(self.__solver_name)

        ordered_labels = []
        for label, frequency in multiset.items():
            for _ in range(frequency):
                ordered_labels.append(label)

        cache_key = frozendict(multiset)
        try:
            return self.__order_cache[cache_key]
        except KeyError:
            x,c = self.__create_lp_variables(solver, ordered_labels)
            self.__create_lp_constraints(solver, x)
            self.__create_lp_objective(solver, x, c)

            solver_status = solver.Solve()
            if solver_status == Solver.OPTIMAL:
                sequence = self.__parse_sequence_from_lp_result(x, ordered_labels)
                self.__order_cache[cache_key] = sequence
                return sequence
            else:
                raise NotImplementedError(solver_status)

    def __parse_sequence_from_lp_result(self, x, ordered_labels) -> Tuple:
        sequence = [None] * len(ordered_labels)
        for i, x_row in enumerate(x):
            for j, x in enumerate(x_row):
                if x.solution_value() == 1:
                    sequence[j] = ordered_labels[i]
        return tuple(sequence)

    def __create_lp_objective(self, solver: Solver, x, c):
        #we maximize over seen label positions in the training set
        solver.Maximize(sum(
            x[i][j] * c[i][j] for i,j in iterate_self_product(list(range(len(x))))))

    def __create_lp_constraints(self, solver: Solver, x):
        #contraint 1: there must be exactly one label per position
        for j in range(len(x)):
            solver.Add(sum(x[i][j] for i in range(len(x))) == 1)
        #contraint 2: a label cannot be at multiple positions
        for i in range(len(x)):
            solver.Add(sum(x[i][j] for j in range(len(x))) == 1)

    def __create_lp_variables(self, solver: Solver, ordered_labels: List):
        #x_i_j := 1 if label i is at position j, 0 otherwise
        x = [
            [
                solver.IntVar(0, 1, f'x_{i}_{j}') for j in range(len(ordered_labels))
            ]
            for i in range(len(ordered_labels))
        ]
        #c_i_j := number of times label i was seen at position j (in the training set)
        c = [
            [
                self.__label_position_count[label][i] for i in range(len(ordered_labels))
            ]
            for label in ordered_labels
        ]
        return x,c