from typing import Union
from collections import Counter

from prolothar_common.models.dataset import TargetSequenceDataset
from prolothar_common.models.dataset import MultiLabelDataset
from prolothar_common.models.dataset import MultisetDataset
from prolothar_common.models.dataset.instance import Instance
from prolothar_common.models.dataset.instance import MultiLabelInstance
from prolothar_common.models.dataset.instance import MultisetInstance

from prolothar_rule_mining.rule_miner.multilabel.rules import Rule as MultiLabelClassificationRule

class MultiLabelSequencePredictor:
    """
    a sequence predictor that first predicts occurence and frequency
    of events and then applies some ordering
    """
    def __init__(self, multilabel_rule_miner, ordering):
        self.__multilabel_rule_miner = multilabel_rule_miner
        self.__ordering = ordering
        self.__multilabel_rule: Union[MultiLabelClassificationRule, None] = None

    def mine_rules(self, dataset: TargetSequenceDataset) -> 'MultiLabelSequencePredictor':
        self.__train_multilabel_rule_miner(dataset)
        self.__train_ordering(dataset)

        return self

    def __train_multilabel_rule_miner(self, dataset: TargetSequenceDataset):
        multilabel_dataset = MultiLabelDataset(
            dataset.get_categorical_attribute_names(),
            dataset.get_numerical_attribute_names())
        for instance in dataset:
            labels = set()
            for label, count in Counter(instance.get_target_sequence()).items():
                labels.add(f'{label}_{count}')
            multilabel_dataset.add_instance(MultiLabelInstance(
                instance.get_id(), instance.get_features_dict(),
                labels))
        self.__multilabel_rule = self.__multilabel_rule_miner.mine_rules(multilabel_dataset)

    def __train_ordering(self, dataset: TargetSequenceDataset):
        multiset_dataset = MultisetDataset(
            dataset.get_categorical_attribute_names(),
            dataset.get_numerical_attribute_names())
        for instance in dataset:
            multiset_dataset.add_instance(MultisetInstance(
                instance.get_id(), instance.get_features_dict(),
                Counter(instance.get_target_sequence())))
        self.__ordering.train(multiset_dataset, dataset)

    def execute(self, instance: Instance):
        prediction = self.__multilabel_rule.predict(instance)
        multiset = {
            label: int(frequency) for label, frequency in (
                class_with_frequency.rsplit('_', 1) for
                class_with_frequency in prediction
            )
        }
        return self.__ordering.order(multiset)

    def get_multilabel_rule(self) -> Union[MultiLabelClassificationRule, None]:
        return self.__multilabel_rule

    def count_nr_of_terms(self):
        return self.__multilabel_rule.count_nr_of_terms()