# -*- coding: utf-8 -*-

"""
interface to the Boomer Rule Mining Algorithm
See https://arxiv.org/pdf/2006.13346.pdf
"""

from typing import Dict
import os
import tempfile
import re

from xml.sax.saxutils import escape as xml_escape

from boomer.boosting.boosting_learners import Boomer as BoomerAlgorithm
from boomer.boosting.boosting_learners import LOSS_EXAMPLE_WISE_LOGISTIC
from boomer.boosting.boosting_learners import LOSS_LABEL_WISE_LOGISTIC
from boomer.boosting.boosting_learners import LOSS_LABEL_WISE_SQUARED_ERROR
from boomer.data import load_data_set_and_meta_data
from boomer.printing import format_model

from prolothar_common.models.dataset import MultiLabelDataset
from prolothar_common.models.dataset.instance import Instance
from prolothar_common.models.dataset.transformer.one_hot_encoding import OneHotEncoding

from prolothar_rule_mining.rule_miner.multilabel.rules import Rule
from prolothar_rule_mining.models.conditions import Condition
from prolothar_rule_mining.models.conditions import EqualsCondition
from prolothar_rule_mining.models.conditions import NotCondition
from prolothar_rule_mining.models.conditions import AndCondition
from prolothar_rule_mining.models.conditions import LessOrEqualCondition
from prolothar_rule_mining.models.conditions import GreaterThanCondition

from prolothar_rule_mining.rule_miner.multilabel.boomer.pruning import Pruning
from prolothar_rule_mining.rule_miner.multilabel.boomer.pruning import NoPruning
from prolothar_rule_mining.rule_miner.multilabel.boomer.boomer_rule import BoomerRule
from prolothar_rule_mining.rule_miner.multilabel.boomer.boomer_rule_list import BoomerRuleList

class Boomer():
    """
    interface to the Boomer Multilabel Classifiation Rule Mining Algorithm
    https://github.com/mrapp-ke/Boomer
    https://arxiv.org/pdf/2006.13346.pdf
    """

    def __init__(self, loss: str = LOSS_LABEL_WISE_LOGISTIC, nr_of_threads: int = -1,
                 max_nr_of_rules: int = 1000, pruning: Pruning = NoPruning()):
        if loss not in (LOSS_EXAMPLE_WISE_LOGISTIC, LOSS_LABEL_WISE_LOGISTIC,
                        LOSS_LABEL_WISE_SQUARED_ERROR):
            raise ValueError('invalid loss: "%s"' % loss)
        self.__loss = loss
        self.__max_nr_of_rules = max_nr_of_rules
        self.__pruning = pruning
        self.__nr_of_threads = nr_of_threads

    def mine_rules(self, dataset: MultiLabelDataset) -> Rule:
        transformed_dataset = OneHotEncoding(
            attribute_value_join_character='___').transform(dataset)
        boomer = BoomerAlgorithm(
            loss=self.__loss, max_rules=self.__max_nr_of_rules,
            pruning='irep', num_threads=self.__nr_of_threads)
        with tempfile.TemporaryDirectory() as temporary_directory:
            self.__write_dataset_to_tempdir(transformed_dataset,
                                            temporary_directory)
            x,y,meta_data = load_data_set_and_meta_data(
                temporary_directory, 'dataset.arff', 'labels.xml')
            boomer.fit(x,y)
            mined_rules = self.__parse_boomer_rule_list(
                format_model(meta_data, boomer.model_),
                transformed_dataset, dataset)

        self.__pruning.prune(mined_rules)
        return mined_rules

    def __write_dataset_to_tempdir(self, dataset: MultiLabelDataset,
                                   temporary_directory: str):
        #http://mulan.sourceforge.net/format.html
        arff_file = os.path.join(temporary_directory, 'dataset.arff')
        labels_file = os.path.join(temporary_directory, 'labels.xml')
        with open(arff_file, 'w') as f:
            f.write(dataset.export_to_arff(label_prefix='label_'))
        with open(labels_file, 'w') as f:
            f.write('<labels xmlns="http://mulan.sourceforge.net/labels">')
            for symbol in sorted(dataset.get_set_of_labels()):
                f.write('<label name="label_%s"></label>' % xml_escape(symbol))
            f.write('</labels>')

    def __parse_boomer_rule_list(
            self, rule_list: str, transformed_dataset: MultiLabelDataset,
            original_dataset: MultiLabelDataset) -> BoomerRuleList:
        #example rule: "{num_feature_2 > 0.016135069} -> (label 1 = 0.32)"
        model = BoomerRuleList(original_dataset)
        for rule in rule_list.splitlines(keepends=False):
            condition, literal = rule.split(' -> ')
            parsed_rule = BoomerRule(
                self.__parse_condition(condition, transformed_dataset,
                                       original_dataset),
                self.__parse_event_confidence_map(literal)
            )
            model.append_rule(parsed_rule)
        return model

    def __parse_condition(self, condition_str, transformed_dataset: MultiLabelDataset,
                          original_dataset: MultiLabelDataset) -> Condition:
        if '&' in condition_str:
            return AndCondition([
                self.__parse_condition(
                    subcondition, transformed_dataset, original_dataset)
                for subcondition in condition_str.split(' & ')
            ])
        if condition_str == '{}':
            return None
        attribute, operator, value = re.split(
            r' (\<=|\>) ', condition_str.lstrip('{').rstrip('}'))
        attribute = transformed_dataset.get_attribute_by_name(attribute)
        if attribute.is_numerical():
            condition_types = {
                '>': GreaterThanCondition,
                '<=': LessOrEqualCondition
            }
            value = float(value)
        else:
            if '___' in attribute.get_name():
                attribute, value = attribute.get_name().split('___')
                attribute = original_dataset.get_attribute_by_name(attribute)
            else:
                #already one hot encoded feature with comparision value 0.5
                attribute = original_dataset.get_attribute_by_name(attribute.get_name())
                value = 0
            condition_types = {
                '<=': lambda a,v: NotCondition(EqualsCondition(a,v)),
                '>': EqualsCondition
            }
        return condition_types[operator](attribute, value)

    def __parse_event_confidence_map(self, literal_str: str) -> Dict[str, float]:
        event_confidence_map = {}
        literal_str = literal_str.lstrip('(').rstrip(')')
        for event_confidence_pair in literal_str.split(', '):
            event, confidence_str = event_confidence_pair.split(' = ')
            if event.startswith('label_'):
                event = event[6:]
            event_confidence_map[event] = float(confidence_str)
        return event_confidence_map

    def is_output_order_invariant(self) -> bool:
        return True
