from typing import List, Set

from math import log
import numpy as np

from prolothar_common.models.dataset import MultiLabelDataset
from prolothar_common.models.dataset.instance import Instance

from prolothar_rule_mining.rule_miner.multilabel.boomer.boomer_rule import BoomerRule

class BoomerRuleList():
    """
    the rule model of boomer. consists of a list of subrules that for a given
    instance add up confidence values for all labels. the prediction is given
    by the risk minimized confidence vector.
    """
    def __init__(self, trainset: MultiLabelDataset):
        self.__rules: List[BoomerRule] = []
        self.__ordered_list_of_events = list(sorted(
            trainset.get_set_of_labels()))
        self.__unique_label_combinations = set()
        for instance in trainset:
            self.__unique_label_combinations.add(
                frozenset(instance.get_labels()))
        self.__unique_label_combinations = list(self.__unique_label_combinations)
        self.__train_label_vectors = [
            np.array([
                1 if event in instance else -1
                for event in self.__ordered_list_of_events
            ]) for instance in self.__unique_label_combinations
        ]

    def count_nr_of_terms(self) -> int:
        """
        returns the number of condition terms in this rule list
        """
        return sum(rule.count_nr_of_terms() for rule in self.__rules)

    def append_rule(self, rule: BoomerRule):
        self.__rules.append(rule)

    def predict(self, instance: Instance) -> Set[str]:
        confidence_per_event = {}
        for rule in self.__rules:
            if rule.applies_to_instance(instance):
                rule.add_confidence(confidence_per_event)

        confidence_vector = np.array([
            confidence_per_event.get(event, 0)
            for event in self.__ordered_list_of_events
        ])

        arg_min = None
        min_loss = float('inf')

        for i, train_label_vector in enumerate(self.__train_label_vectors):
            loss = log(1 + np.sum(np.exp(-np.multiply(
                confidence_vector, train_label_vector
            ))))
            if loss < min_loss:
                min_loss = loss
                arg_min = i

        return set(self.__unique_label_combinations[arg_min])

    def __len__(self) -> int:
        """returns the number of rules in this list, including the default rule"""
        return len(self.__rules)

    def __repr__(self) -> str:
        return repr(self.__rules)

    def __iter__(self):
        return iter(self.__rules)

    def pop_last_rule(self) -> BoomerRule:
        """
        removes and returns the last rule in this list, which is the rule
        with the lowest effect
        """
        return self.__rules.pop()

    def pop_last_n_rules(self, n: int) -> List[BoomerRule]:
        """
        removes and returns the last n rules in this list

        Parameters
        ----------
        n : int
            determines how many rules should be popped

        Returns
        -------
        List[BoomerRule]
            the first element in the list is the last rule at the beginning
        """
        result = []
        for _ in range(n):
            result.append(self.pop_last_rule())
        return result
