"""
pruning algorithms to postprocess the boomer model
"""

from abc import ABC, abstractmethod

from prolothar_common.models.dataset import MultiLabelDataset

from prolothar_rule_mining.rule_miner.multilabel.evaluation import evaluate
from prolothar_rule_mining.rule_miner.multilabel.boomer.boomer_rule_list import BoomerRuleList

class Pruning(ABC):
    """
    template for all pruning strategies
    """
    @abstractmethod
    def prune(self, rule_list: BoomerRuleList):
        """
        prunes the given rule_list

        Parameters
        ----------
        rule_list : BoomerRuleList
            will be pruned inplace
        """

class NoPruning(Pruning):
    """
    NullObjectPattern for Pruning, i.e. a pseudo-pruning strategy that does no
    pruning at all
    """
    def prune(self, rule_list: BoomerRuleList):
        #no pruning
        pass

class NaiveF1ScorePruning(Pruning):
    """
    Removes rules from the end of the list (which are supposed to have the smallest
    effect) until F1 score on a given dataset (e.g. the trainset) drops too much.
    """
    def __init__(self, dataset: MultiLabelDataset, accepted_drop: float = 0):
        """
        configures this pruning strategy

        Parameters
        ----------
        dataset : MultiLabelDataset
            dataset on which the loss of f1 score is computed while pruning the rule list
        accepted_drop : float, optional
            the accepted drop in f1 score until pruning is stopped, by default 0
        """
        self.__dataset = dataset
        self.__accepted_drop = accepted_drop

    def prune(self, rule_list: BoomerRuleList):
        initial_f1_score = evaluate(rule_list, self.__dataset).micro_f1
        #evaluation is an expensive operation that we want to avoid
        #this is why we try to remove a lot of rules first instead of just 1
        pruning_size = len(rule_list) // 2
        while len(rule_list) > 1 and pruning_size > 0:
            removed_rules = rule_list.pop_last_n_rules(pruning_size)
            pruned_f1_score = evaluate(rule_list, self.__dataset).micro_f1
            if pruned_f1_score < initial_f1_score - self.__accepted_drop:
                for removed_rule in removed_rules[::-1]:
                    rule_list.append_rule(removed_rule)
                pruning_size = pruning_size // 2
            else:
                pruning_size = min(pruning_size, len(rule_list) // 2)
