from random import Random
import copy

from prolothar_rule_mining.rule_miner.multilabel.mlrbc.classifier import Classifier

class ClassifierSet:
    def __init__(self, max_population_size: int = 10000, random: Random = None):
        """ Overloaded initialization: Handles creation of a new population or a rebooted population (i.e. a previously saved population). """
        # Major Parameters
        self.popSet = []  # List of classifiers/rules
        self.matchSet = []  # List of references to rules in population that match
        self.correctSet = []  # List of references to rules in population that both match and specify correct phenotype
        self.microPopSize = 0  # Tracks the current micro population size
        self.labelPowerSetList = []

        # Evaluation Parameters-------------------------------
        self.aveGenerality = 0.0
        self.expRules = 0.0
        self.attributeSpecList = []
        self.attributeAccList = []
        self.avePhenotypeRange = 0.0
        self.__max_population_size = max_population_size
        self.__random = random if random is not None else Random()

    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    # CLASSIFIER SET CONSTRUCTOR METHODS
    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    def makeMatchSet(self, attribute_info, beta, p_spec, nu, theta_sub, acc_sub, state_phenotype, exploreIter):
        """ Constructs a match set from the population. Covering is initiated if the match set is empty or a rule with the current correct phenotype is absent. """
        # Initial values
        state = state_phenotype[0]
        phenotype = state_phenotype[1]
        doCovering = True  # Covering check: Twofold (1)checks that a match is present, and (2) that at least one match dictates the correct phenotype.
        setNumerositySum = 0  # new

        # -------------------------------------------------------
        # MATCHING
        # -------------------------------------------------------
        for i in range(len(self.popSet)):  # Go through the population
            cl = self.popSet[i]  # One classifier at a time
            if cl.match(state):  # Check for match
                # print("Condition: "+ str(cl.reportClassifier()) + "  Phenotype: "+ str(cl.phenotype)) #debug all [M]
                self.matchSet.append(i)  # If match - add classifier to match set
                setNumerositySum += cl.numerosity  # New Increment the set numerosity sum

                # Covering Check--------------------------------------------------------
                if cl.phenotype == phenotype:  # Check for phenotype coverage
                    doCovering = False

        # -------------------------------------------------------
        # COVERING
        # -------------------------------------------------------
        while doCovering:
            newCl = Classifier(
                attribute_info, beta, p_spec, nu, theta_sub, acc_sub,
                a=setNumerositySum + 1, b=exploreIter, c=state, d=phenotype)
            self.addClassifierToPopulation(newCl, True)
            self.matchSet.append(len(self.popSet) - 1)  # Add covered classifier to matchset
            doCovering = False

    def makeCorrectSet(self, phenotype):
        for i in range(len(self.matchSet)):
            ref = self.matchSet[i]
            if self.popSet[ref].phenotype == phenotype:
                self.correctSet.append(ref)

    def isPhenotypeSubset(self, ref, phenotype):
        isSusbet = True
        cl_phenotype = self.popSet[ref].phenotype
        for (L, L_cl) in zip(phenotype, cl_phenotype):
            if L == '0' and L_cl != "0":
                isSusbet = False
                break
        return isSusbet

    def makeEvalMatchSet(self, state):
        """ Constructs a match set for evaluation purposes which does not activate either covering or deletion. """
        for i in range(len(self.popSet)):  # Go through the population
            cl = self.popSet[i]  # A single classifier
            if cl.match(state):  # Check for match
                self.matchSet.append(i)  # Add classifier to match set

    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    # CLASSIFIER DELETION METHODS
    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    def deletion(self, exploreIter):
        """ Returns the population size back to the maximum set by the user by deleting rules. """
        while self.microPopSize > self.__max_population_size:
            self.deleteFromPopulation()

    def deleteFromPopulation(self):
        """ Deletes one classifier in the population.  The classifier that will be deleted is chosen by roulette wheel selection
        considering the deletion vote. Returns the macro-classifier which got decreased by one micro-classifier. """
        meanFitness = self.getPopFitnessSum() / float(self.microPopSize)

        # Calculate total wheel size------------------------------
        sumCl = 0.0
        voteList = []
        for cl in self.popSet:
            vote = cl.getDelProp(meanFitness)
            sumCl += vote
            voteList.append(vote)
        # --------------------------------------------------------
        choicePoint = sumCl * self.__random.random()  # Determine the choice point

        newSum = 0.0
        for i in range(len(voteList)):
            cl = self.popSet[i]
            newSum = newSum + voteList[i]
            if newSum > choicePoint:  # Select classifier for deletion
                # Delete classifier----------------------------------
                cl.updateNumerosity(-1)
                self.microPopSize -= 1
                if cl.numerosity < 1:  # When all micro-classifiers for a given classifier have been depleted.
                    self.removeMacroClassifier(i)
                    self.deleteFromMatchSet(i)
                    self.deleteFromCorrectSet(i)
                return

        print("ClassifierSet: No eligible rules found for deletion in deleteFromPopulation.")
        return

    def removeMacroClassifier(self, ref):
        """ Removes the specified (macro-) classifier from the population. """
        self.popSet.pop(ref)

    def deleteFromMatchSet(self, deleteRef):
        """ Delete reference to classifier in population, contained in self.matchSet."""
        if deleteRef in self.matchSet:
            self.matchSet.remove(deleteRef)

        # Update match set reference list--------
        for j in range(len(self.matchSet)):
            ref = self.matchSet[j]
            if ref > deleteRef:
                self.matchSet[j] -= 1

    def deleteFromCorrectSet(self, deleteRef):
        """ Delete reference to classifier in population, contained in self.corectSet."""
        if deleteRef in self.correctSet:
            self.correctSet.remove(deleteRef)

        # Update match set reference list--------
        for j in range(len(self.correctSet)):
            ref = self.correctSet[j]
            if ref > deleteRef:
                self.correctSet[j] -= 1

    # ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    # GENETIC ALGORITHM
    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    def runGA(
        self, exploreIter, state, phenotype, theta_GA, selection_method, theta_sel,
        attribute_info, beta, p_spec, nu, theta_sub, acc_sub, fitness_reduction, chi,
        upsilon, do_subsumption):
        """ The genetic discovery mechanism in eLCS is controlled here. """
        # -------------------------------------------------------
        # GA RUN REQUIREMENT
        # -------------------------------------------------------
        if (exploreIter - self.getIterStampAverage()) < theta_GA:  # Does the correct set meet the requirements for activating the GA?
            return

        self.setIterStamps(exploreIter)  # Updates the iteration time stamp for all rules in the correct set (which the GA opperates in).
        changed = False

        # -------------------------------------------------------
        # SELECT PARENTS - Niche GA - selects parents from the correct class
        # -------------------------------------------------------
        if selection_method == "roulette":
            selectList = self.selectClassifierRW()
            clP1 = selectList[0]
            clP2 = selectList[1]
        elif selection_method == "tournament":
            selectList = self.selectClassifierT(theta_sel)
            clP1 = selectList[0]
            clP2 = selectList[1]
        else:
            raise NotImplementedError("ClassifierSet: Error - requested GA selection method not available.")

        # -------------------------------------------------------
        # INITIALIZE OFFSPRING
        # -------------------------------------------------------
        cl1 = Classifier(attribute_info, beta, p_spec, nu,
            theta_sub, acc_sub, a=clP1, b=exploreIter)
        if clP2 == None:
            cl2 = Classifier(attribute_info, beta, p_spec, nu,
            theta_sub, acc_sub, a=clP1, b=exploreIter)
        else:
            cl2 = Classifier(attribute_info, beta, p_spec, nu,
            theta_sub, acc_sub, a=clP2, b=exploreIter)
        # -------------------------------------------------------
        # CROSSOVER OPERATOR - Uniform Crossover Implemented (i.e. all attributes have equal probability of crossing over between two parents)
        # -------------------------------------------------------
        if not cl1.equals(cl2) and self.__random.random() < chi:
            changed = cl1.uniformCrossover(cl2)

            # -------------------------------------------------------
        # INITIALIZE KEY OFFSPRING PARAMETERS
        # -------------------------------------------------------
        if changed:
            cl1.setAccuracy((cl1.accuracy + cl2.accuracy) / 2.0)
            cl1.setFitness(fitness_reduction * (cl1.fitness + cl2.fitness) / 2.0)
            cl2.setAccuracy(cl1.accuracy)
            cl2.setFitness(cl1.fitness)
        else:
            cl1.setFitness(fitness_reduction * cl1.fitness)
            cl2.setFitness(fitness_reduction * cl2.fitness)
        # -------------------------------------------------------
        # MUTATION OPERATOR
        # -------------------------------------------------------
        nowchanged = cl1.Mutation(state, phenotype, upsilon)
        howaboutnow = cl2.Mutation(state, phenotype, upsilon)

        # -------------------------------------------------------
        # ADD OFFSPRING TO POPULATION
        # -------------------------------------------------------
        if changed or nowchanged or howaboutnow:
            self.insertDiscoveredClassifiers(cl1, cl2, clP1, clP2, exploreIter, do_subsumption)  # Subsumption

    # -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    # SELECTION METHODS
    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    def selectClassifierRW(self):
        """ Selects parents using roulette wheel selection according to the fitness of the classifiers. """
        selectList = [None, None]
        setList = copy.deepcopy(self.correctSet)
        if len(setList) < 2:
            currentCount = 1  # Pick only one parent
        else:
            currentCount = 0  # Pick two parents
        # -----------------------------------------------
        # -----------------------------------------------
        while currentCount < 2:
            fitSum = self.getFitnessSum(setList)

            choiceP = self.__random.random() * fitSum
            i = 0
            ref = setList[i]
            sumCl = self.popSet[ref].fitness
            while choiceP > sumCl:
                i = i + 1
                ref = setList[i]  # WnB must be added to increment fitness
                sumCl += self.popSet[ref].fitness

            selectList[currentCount] = self.popSet[ref]
            setList.remove(ref)
            currentCount += 1

        # -----------------------------------------------
        if selectList[0] == None:
            selectList[0] = selectList[1]

        return selectList

    def selectClassifierT(self, theta_sel):
        """  Selects parents using tournament selection according to the fitness of the classifiers. """
        selectList = [None, None]
        currentCount = 0
        setList = self.correctSet  # correct set is a list of reference IDs

        while currentCount < 2:
            tSize = int(len(setList) * theta_sel)
            posList = self.__random.sample(setList, tSize)

            bestF = 0
            bestC = self.correctSet[0]
            for j in posList:
                if self.popSet[j].fitness > bestF:
                    bestF = self.popSet[j].fitness
                    bestC = j

            selectList[currentCount] = self.popSet[bestC]
            currentCount += 1

        return selectList


        # -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

    # SUBSUMPTION METHODS
    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    def subsumeClassifier(self, cl=None, cl1P=None, cl2P=None):
        """ Tries to subsume a classifier in the parents. If no subsumption is possible it tries to subsume it in the current set. """
        if cl1P != None and cl1P.subsumes(cl):
            self.microPopSize += 1
            cl1P.updateNumerosity(1)
        elif cl2P != None and cl2P.subsumes(cl):
            self.microPopSize += 1
            cl2P.updateNumerosity(1)
        else:
            self.subsumeClassifier2(cl);  # Try to subsume in the correct set.

    def subsumeClassifier2(self, cl):
        """ Tries to subsume a classifier in the correct set. If no subsumption is possible the classifier is simply added to the population considering
        the possibility that there exists an identical classifier. """
        choices = []
        for ref in self.correctSet:
            if self.popSet[ref].subsumes(cl):
                choices.append(ref)

        if len(choices) > 0:  # Randomly pick one classifier to be subsumer
            choice = int(self.__random.random() * len(choices))
            self.popSet[choices[choice]].updateNumerosity(1)
            self.microPopSize += 1
            return

        self.addClassifierToPopulation(cl,
                                        False)  # If no subsumer was found, check for identical classifier, if not then add the classifier to the population

    def doCorrectSetSubsumption(self):
        """ Executes correct set subsumption.  The correct set subsumption looks for the most general subsumer classifier in the correct set
        and subsumes all classifiers that are more specific than the selected one. """
        subsumer = None
        for ref in self.correctSet:
            cl = self.popSet[ref]
            if cl.isSubsumer():
                if subsumer == None or cl.isMoreGeneral(subsumer):
                    subsumer = cl

        if subsumer != None:  # If a subsumer was found, subsume all more specific classifiers in the correct set
            i = 0
            while i < len(self.correctSet):
                ref = self.correctSet[i]
                if subsumer.isMoreGeneral(self.popSet[ref]):
                    subsumer.updateNumerosity(self.popSet[ref].numerosity)
                    self.removeMacroClassifier(ref)
                    self.deleteFromMatchSet(ref)
                    self.deleteFromCorrectSet(ref)
                    i = i - 1
                i = i + 1

    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    # OTHER KEY METHODS
    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    def addClassifierToPopulation(self, cl, covering):
        """ Note new change here: Adds a classifier to the set and increases the microPopSize value accordingly."""
        oldCl = None
        if not covering:
            oldCl = self.getIdenticalClassifier(cl)
        if oldCl != None:  # found identical classifier
            oldCl.updateNumerosity(1)
            self.microPopSize += 1
        else:
            self.popSet.append(cl)
            self.microPopSize += 1

    def insertDiscoveredClassifiers(self, cl1, cl2, clP1, clP2, exploreIter, do_subsumption):
        """ Inserts both discovered classifiers and activates GA subsumption if turned on. Also checks for default rule (i.e. rule with completely general condition) and
        prevents such rules from being added to the population, as it offers no predictive value within eLCS. """
        # -------------------------------------------------------
        # SUBSUMPTION
        # -------------------------------------------------------
        if do_subsumption:
            if len(cl1.specifiedAttList) > 0:
                self.subsumeClassifier(cl1, clP1, clP2)
            if len(cl2.specifiedAttList) > 0:
                self.subsumeClassifier(cl2, clP1, clP2)
        # -------------------------------------------------------
        # ADD OFFSPRING TO POPULATION
        # -------------------------------------------------------
        else:  # Just add the new classifiers to the population.
            if len(cl1.specifiedAttList) > 0:
                self.addClassifierToPopulation(cl1,
                                                False)  # False passed because this is not called for a covered rule.
            if len(cl2.specifiedAttList) > 0:
                self.addClassifierToPopulation(cl2,
                                                False)  # False passed because this is not called for a covered rule.

    # --Note New changes here-------------
    def updateSets(self, exploreIter, state_phenotype_conf):
        """ Updates all relevant parameters in the current match and correct sets. """

        matchSetNumerosity = 0
        for ref in self.matchSet:
            matchSetNumerosity += self.popSet[ref].numerosity

        for ref in self.matchSet:
            self.popSet[ref].updateExperience()
            self.popSet[ref].updateMatchSetSize(matchSetNumerosity)
            if ref in self.correctSet:
                self.popSet[ref].updateCorrect()
            self.popSet[ref].updateMLperformance(state_phenotype_conf[1], None)  # New
            self.popSet[ref].updateAccuracy()
            self.popSet[ref].updateFitness()


    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    # OTHER METHODS
    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    def getIterStampAverage(self):
        """ Returns the average of the time stamps in the correct set. """
        sumCl = 0.0
        numSum = 0.0
        for i in range(len(self.correctSet)):
            ref = self.correctSet[i]
            sumCl += self.popSet[ref].timeStampGA * self.popSet[ref].numerosity
            numSum += self.popSet[ref].numerosity  # numerosity sum of correct set
            if numSum < 1:
                print('Pause')
        return sumCl / float(numSum)

    def setIterStamps(self, exploreIter):
        """ Sets the time stamp of all classifiers in the set to the current time. The current time
        is the number of exploration steps executed so far.  """
        for i in range(len(self.correctSet)):
            ref = self.correctSet[i]
            self.popSet[ref].updateTimeStamp(exploreIter)

    def getFitnessSum(self, setList):
        """ Returns the sum of the fitnesses of all classifiers in the set. """
        sumCl = 0.0
        for i in range(len(setList)):
            ref = setList[i]
            sumCl += self.popSet[ref].fitness
        return sumCl

    def getPopFitnessSum(self):
        """ Returns the sum of the fitnesses of all classifiers in the set. """
        sumCl = 0.0
        for cl in self.popSet:
            sumCl += cl.fitness * cl.numerosity
        return sumCl

    def getIdenticalClassifier(self, newCl):
        """ Looks for an identical classifier in the population. """
        for cl in self.popSet:
            if newCl.equals(cl):
                return cl
        return None

    def clearSets(self):
        """ Clears out references in the match and correct sets for the next learning iteration. """
        self.matchSet = []
        self.correctSet = []

    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    # EVALUTATION METHODS
    # --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    def runPopAveEval(self, exploreIter, number_of_attributes):
        """ Calculates some summary evaluations across the rule population including average generality. """
        genSum = 0
        agedCount = 0
        for cl in self.popSet:
            genSum += ((number_of_attributes - len(cl.condition)) / number_of_attributes)
        if self.microPopSize == 0:
            self.aveGenerality = 'NA'
        else:
            self.aveGenerality = genSum / float(self.microPopSize)

    # New method
    def runAttGeneralitySum(self, isEvaluationSummary):
        """ Determine the population-wide frequency of attribute specification, and accuracy weighted specification.  Used in complete rule population evaluations. """
        if isEvaluationSummary:
            self.attributeSpecList = []
            self.attributeAccList = []
            for i in range(cons.env.formatData.number_of_attributes):
                self.attributeSpecList.append(0)
                self.attributeAccList.append(0.0)
            for cl in self.popSet:
                for ref in cl.specifiedAttList:  # for each attRef
                    self.attributeSpecList[ref] += cl.numerosity
                    self.attributeAccList[ref] += cl.numerosity * cl.accuracy

    def getPopTrack(self, Hloss, accuracy, exploreIter, trackingFrequency):
        """ Returns a formated output string to be printed to the Learn Track output file. """
        return 'iteration %d: |population|=%d, |micro_population|=%d, hloss=%r, ave_generality=%.2f' % (
            exploreIter, len(self.popSet), self.microPopSize, Hloss, self.aveGenerality
        )