from typing import FrozenSet
from random import Random
import itertools

from prolothar_common.models.dataset import MultiLabelDataset

class DataManagement:
    def __init__(self, dataset: MultiLabelDataset, random: Random = None):
        # Set random seed if specified.-----------------------------------------------
        self.__random = random if random is not None else Random()

        self.number_of_attributes = dataset.get_nr_of_attributes()
        # Stores Discrete (0) or Continuous (1) for each attribute
        #for numerical attributes the range is also stored
        self.attribute_info = []
        self.original_attributes = []
        for attribute in dataset.get_categorical_attribute_names():
            self.attribute_info.append([0, []])
            self.original_attributes.append(dataset.get_attribute_by_name(attribute))
        for attribute in dataset.get_numerical_attribute_names():
            values = dataset.get_attribute_by_name(attribute).get_unique_values()
            self.attribute_info.append([1, [min(values), max(values)]])
            self.original_attributes.append(dataset.get_attribute_by_name(attribute))
        # Stores all possible discrete phenotype states/classes
        self.__original_labels = sorted(dataset.get_set_of_labels())
        self.phenotype_list = [self.__encode_phenotype(instance.get_labels()) for instance in dataset]

        self.nr_of_instances = len(dataset)

        self.data = []
        for instance in dataset:
            attribute_states = [
                self.__transform_empty_string(instance[attribute]) for attribute in itertools.chain(
                    dataset.get_categorical_attribute_names(),
                    dataset.get_numerical_attribute_names())
            ]
            self.data.append([attribute_states, self.__encode_phenotype(instance.get_labels()), instance.get_id()])
          # One time randomization of the order the of the instances in the data,
          # so that if the data was ordered by phenotype, this potential learning bias
          # (based on instance ordering) is eliminated.
        self.__random.shuffle(self.data)

    def __encode_phenotype(self, label_set):
        return ''.join('1' if label in label_set else '0' for label in self.__original_labels)

    def decode_phenotype(self, phenotype) -> FrozenSet:
        class_labels = set()
        for i,bit in enumerate(phenotype):
            if bit == '1':
                class_labels.add(self.__original_labels[i])
        return frozenset(class_labels)

    def __transform_empty_string(self, attribute_value):
        if attribute_value == '':
            return '_EMPTY_'
        return attribute_value
