from typing import Callable

from prolothar_common.func_tools import do_nothing

from prolothar_rule_mining.rule_miner.multilabel.mlrbc.classifier_set import ClassifierSet
from prolothar_rule_mining.rule_miner.multilabel.mlrbc.class_accuracy import ClassAccuracy
from prolothar_rule_mining.rule_miner.multilabel.mlrbc.prediction import Prediction
from prolothar_rule_mining.rule_miner.multilabel.mlrbc.data_management import DataManagement

class ELcs:
    def __init__(
        self, max_learning_iterations: int = 1000, logger: Callable[[str],None] = None,
        tracking_frequency: int = 10, do_subsumption: bool = True,
        beta: float = 0.1, p_spec: float = 0.4, nu: int = 5,
        theta_sub: int = 20, acc_sub: float = 0.99, theta_GA: int = 25,
        selection_method: str = 'tournament', theta_sel: float = 0.2,
        fitness_reduction: float = 0.1, chi: float = 0.8, upsilon: float = 0.04,
        max_population_size: int = 10000):
        """ Initializes the eLCS algorithm """
        self.__max_learning_iterations = max_learning_iterations
        self.__tracking_frequency = tracking_frequency
        self.__do_subsumption = do_subsumption
        self.__beta = beta
        self.__p_spec = p_spec
        self.__nu = nu
        self.__theta_sub = theta_sub
        self.__acc_sub = acc_sub
        self.__theta_GA = theta_GA
        self.__selection_method = selection_method
        self.__theta_sel = theta_sel
        self.__fitness_reduction = fitness_reduction
        self.__chi = chi
        self.__upsilon = upsilon

        self.__logger = logger if logger is not None else do_nothing

        # Instantiate Population---------
        self.population = ClassifierSet(max_population_size=max_population_size)
        self.correct = [0.0] * tracking_frequency
        self.hloss = [1] * tracking_frequency
        self.Pr = [0.0] * tracking_frequency

    def run_eLCS(self, data_manager: DataManagement):
        """ Runs the initialized eLCS algorithm. """
        nr_of_iterations = 0
        instance_index = 0
        while nr_of_iterations < self.__max_learning_iterations:

            state_phenotype_conf = data_manager.data[instance_index]
            self.runIteration(state_phenotype_conf, nr_of_iterations, data_manager.attribute_info)

            if (nr_of_iterations % self.__tracking_frequency) == (self.__tracking_frequency - 1) and nr_of_iterations > 0:
                self.population.runPopAveEval(nr_of_iterations, data_manager.number_of_attributes)
                trackedAccuracy = sum(self.correct) / self.__tracking_frequency
                trackedHloss = sum(self.hloss) / self.__tracking_frequency
                self.__logger(self.population.getPopTrack(round(trackedHloss,3), trackedAccuracy, nr_of_iterations + 1,
                                                                        self.__tracking_frequency))  # Report learning progress to standard out and tracking file.

            nr_of_iterations += 1  # Increment current learning iteration
            instance_index = (instance_index + 1) % data_manager.nr_of_instances

    def runIteration(self, state_phenotype_conf, nr_of_iterations: int, attribute_info):
        # -----------------------------------------------------------------------------------------------------------------------------------------
        # FORM A MATCH SET - includes covering
        # -----------------------------------------------------------------------------------------------------------------------------------------
        self.population.makeMatchSet(
            attribute_info, self.__beta, self.__p_spec,
            self.__nu, self.__theta_sub, self.__acc_sub,
            state_phenotype_conf, nr_of_iterations)
        # -----------------------------------------------------------------------------------------------------------------------------------------
        # MAKE A PREDICTION - utilized here for tracking estimated learning progress.  Typically used in the explore phase of many LCS algorithms.
        # -----------------------------------------------------------------------------------------------------------------------------------------
        Acc = ClassAccuracy()
        prediction = Prediction(self.population)
        phenotypePrediction = prediction.getDecision()
        # Prediction Successful
        if phenotypePrediction is not None and phenotypePrediction != 'Tie':
            self.hloss[nr_of_iterations % self.__tracking_frequency] = Acc.hammingLoss(phenotypePrediction, state_phenotype_conf[1])

        # -----------------------------------------------------------------------------------------------------------------------------------------
        # FORM A CORRECT SET
        # -----------------------------------------------------------------------------------------------------------------------------------------
        self.population.makeCorrectSet(state_phenotype_conf[1])
        # -----------------------------------------------------------------------------------------------------------------------------------------
        # UPDATE PARAMETERS
        # -----------------------------------------------------------------------------------------------------------------------------------------
        self.population.updateSets(nr_of_iterations, state_phenotype_conf)
        # -----------------------------------------------------------------------------------------------------------------------------------------
        # SUBSUMPTION - APPLIED TO CORRECT SET - A heuristic for addition additional generalization pressure to eLCS
        # -----------------------------------------------------------------------------------------------------------------------------------------
        if self.__do_subsumption:
            self.population.doCorrectSetSubsumption()
        # -----------------------------------------------------------------------------------------------------------------------------------------
        # RUN THE GENETIC ALGORITHM - Discover new offspring rules from a selected pair of parents
        # -----------------------------------------------------------------------------------------------------------------------------------------
        self.population.runGA(
            nr_of_iterations, state_phenotype_conf[0], state_phenotype_conf[1],
            self.__theta_GA, self.__selection_method, self.__theta_sel,
            attribute_info, self.__beta, self.__p_spec, self.__nu,
            self.__theta_sub, self.__acc_sub, self.__fitness_reduction, self.__chi,
            self.__upsilon, self.__do_subsumption)
        # -----------------------------------------------------------------------------------------------------------------------------------------
        # SELECT RULES FOR DELETION - This is done whenever there are more rules in the population than 'N', the maximum population size.
        # -----------------------------------------------------------------------------------------------------------------------------------------
        self.population.deletion(nr_of_iterations)
        self.population.clearSets()  # Clears the match and correct sets for the next learning iteration
