from typing import Callable, List

from prolothar_common.models.dataset import MultiLabelDataset
from prolothar_common.func_tools import do_nothing

from prolothar_rule_mining.models.conditions import Condition
from prolothar_rule_mining.models.conditions import AndCondition
from prolothar_rule_mining.models.conditions import EqualsCondition
from prolothar_rule_mining.models.conditions import InCondition
from prolothar_rule_mining.models.conditions import InRangeCondition

from prolothar_rule_mining.rule_miner.classification.rules import ListOfRules
from prolothar_rule_mining.rule_miner.classification.rules import IfThenElseRule
from prolothar_rule_mining.rule_miner.classification.rules import VoteForClassRule

from prolothar_rule_mining.rule_miner.multilabel.mlrbc.data_management import DataManagement
from prolothar_rule_mining.rule_miner.multilabel.mlrbc.elcs import ELcs
from prolothar_rule_mining.rule_miner.multilabel.mlrbc.classifier import Classifier

class Mlrbc:
    """
    Multi-label Rule-based Classifier (MLRBC) algorithm
    https://github.com/ConflictedPhilosopher/MLRBC/blob/master/MLRBC.py
    """

    def __init__(
            self, max_learning_iterations: int = 1000,
            logger: Callable[[str],None] = print,
            do_subsumption: bool = True,
            tracking: bool = True,
            beta: float = 0.1, p_spec: float = 0.4, nu: int = 5,
            theta_sub: int = 20, acc_sub: float = 0.99, theta_GA: int = 25,
            selection_method: str = 'tournament', theta_sel: float = 0.2,
            fitness_reduction: float = 0.1, chi: float = 0.8, upsilon: float = 0.04,
            max_population_size: int = 10000):
        """
        creates and configures a new instance of the MLRBC algorithm

        Parameters
        ----------
        max_learning_iterations : int, optional
            number of learning iterations, by default 1000
        logger : Callable[[str],None], optional
            for debugging purpose, by default print
        do_subsumption : bool, optional
            Subsumption is a heuristic that actively seeks to increase
            generalization in the rule population, by default True
        tracking : bool, optional
            whether to activate loss tracking once per epoch for debugging
            purposes, by default True
        beta : float, optional
            Learning parameter; Used in calculating average correct set size,
            by default 0.1
        p_spec : float, optional
            [description], by default 0.4
        nu : int, optional
            Power parameter used to determine the importance of high accuracy
            when calculating fitness. (typically set to 5, recommended setting
            of 1 in noisy data), by default 5
        theta_sub : int, optional
            The subsumption experience threshold, by default 20
        acc_sub : float, optional
            Subsumption accuracy requirement, by default 0.99
        theta_GA : int, optional
            The GA threshold; The GA is applied in a set when the average time
            since the last GA in the set is greater than theta_GA, by default 25
        selection_method : str, optional
            Select GA parent selection strategy ('tournament' or 'roulette'),
            by default 'tournament'
        theta_sel : float, optional
            The fraction of the correct set to be included in tournament
            selection, by default 0.2
        fitness_reduction : float, optional
            Initial fitness reduction in GA offspring rules, by default 0.1
        chi : float, optional
            (X) The probability of applying crossover in the GA.
            (typically set to 0.5-1.0), by default 0.8
        upsilon : float, optional
            (u) The probability of mutating an allele within an offspring.
            (typically set to 0.01-0.05), by default 0.04
        max_population_size : int, optional
            Maximum size of the rule population (a.k.a. Micro-classifier
            population size, where max_population_size is the sum of the classifier
            numerosities in the population), by default 0.04
        """
        self.__max_learning_iterations = max_learning_iterations
        self.__logger = logger if logger is not None else do_nothing
        self.__do_subsumption = do_subsumption
        self.__tracking = tracking
        self.__beta = beta
        self.__p_spec = p_spec
        self.__nu = nu
        self.__theta_sub = theta_sub
        self.__acc_sub = acc_sub
        self.__theta_GA = theta_GA
        self.__selection_method = selection_method
        self.__theta_sel = theta_sel
        self.__fitness_reduction = fitness_reduction
        self.__chi = chi
        self.__upsilon = upsilon
        self.__max_population_size = max_population_size

    def mine_rules(self, dataset: MultiLabelDataset) -> ListOfRules:
        data_management = DataManagement(dataset)
        elcs_algorithm = ELcs(
            logger=self.__logger,
            do_subsumption=self.__do_subsumption,
            max_learning_iterations=self.__max_learning_iterations,
            tracking_frequency=len(dataset) if self.__tracking else 0,
            beta=self.__beta, p_spec=self.__p_spec, nu=self.__nu,
            theta_sub=self.__theta_sub, acc_sub=self.__acc_sub,
            theta_GA=self.__theta_GA, selection_method=self.__selection_method,
            theta_sel=self.__theta_sel, fitness_reduction=self.__fitness_reduction,
            chi=self.__chi, upsilon=self.__upsilon,
            max_population_size=self.__max_population_size
        )
        elcs_algorithm.run_eLCS(data_management)
        return self.__parse_elcs_result(elcs_algorithm.population.popSet, data_management)

    def __parse_elcs_result(self, classifier_list: List[Classifier], data_management: DataManagement):
        rule_list = ListOfRules()
        for classifier in classifier_list:
            rule_list.append_rule(IfThenElseRule(
                self.__parse_condition(classifier, data_management),
                ListOfRules([
                    VoteForClassRule(data_management.decode_phenotype(classifier.phenotype))
                ])
            ))
        return rule_list

    def __parse_condition(self, classifier: Classifier, data_management: DataManagement) -> Condition:
        if not classifier.specifiedAttList:
            raise NotImplementedError()
        elif len(classifier.specifiedAttList) == 1:
            return self.__parse_attribute_condition(
                classifier.specifiedAttList[0], classifier.condition[0], data_management)
        else:
            return AndCondition([
                self.__parse_attribute_condition(attribute_index, condition, data_management)
                for attribute_index, condition in zip(
                    classifier.specifiedAttList, classifier.condition)
            ])

    def __parse_attribute_condition(
            self, attribute_index, condition, data_management: DataManagement) -> Condition:
        attribute = data_management.original_attributes[attribute_index]
        if attribute.is_categorical():
            if condition == '_EMPTY_':
                condition = ''
            return EqualsCondition(attribute, condition)
        elif attribute.is_numerical():
            if len(condition) != 2:
                raise NotImplementedError(condition)
            return InRangeCondition(attribute, condition[0], condition[1])
        else:
            raise NotImplementedError(attribute)