# -*- coding: utf-8 -*-

"""provides a class for convenient logging of experiments"""

import os
from datetime import datetime

from typing.io import IO

class ExperimentLogger():
    """a class for convenient logging of experiments, e.g. ML pipelines"""

    def __init__(self, base_directory: str, experiment_name: str):
        """creates a new logger for an experiment

        Args:
            base_directory:
                path to the base directory of the logger. will be
                created if it does not exist. under this directory a new folder
                will be created with the experiment_name and a current timestamp
            experiment_name:
                Name of the experiment
        """
        self._base_directory = base_directory
        self._experiment_name = experiment_name

        if not os.path.exists(self._base_directory):
            os.makedirs(self._base_directory)

        self._path_to_experiment_directory = os.path.join(
                self._base_directory,
                '%s %s' % (self._experiment_name,
                           datetime.now().strftime('%d-%m-%Y (%H-%M-%S-%f)')))
        os.makedirs(self._path_to_experiment_directory)

        self._log_file = self.open_text_file('log.txt')

    def open_text_file(self, relative_path: str, encoding = 'utf-8') -> IO:
        """opens a new writable textfile in the directory of this experiment"""
        return self.open_file(relative_path, 'w', encoding = encoding)

    def open_binary_file(self, relative_path: str) -> IO:
        """opens a new writable binary file in the directory of this experiment"""
        return self.open_file(relative_path, 'wb')

    def open_file(self, relative_path: str, file_mode: str, encoding = None) -> IO:
        """opens a new file file in the directory of this experiment

        Args:
            relative_path: relative path or name of the file
            file_mode: e.g. 'wb'
        """
        return open(self.get_absolute_path(relative_path), file_mode,
                    encoding = encoding)

    def get_absolute_path(self, relative_path: str) -> str:
        """
        transforms the given relative_path to an absolute path in the logger
        directory.

        Example:
        get_absolute_path('myfile.png')
        returns something like
        "E:\\Projekte\\Python\\dh-experiment-logger\\testtemp\\Testexperiment 18-05-2021 (11-29-49-446524)\\myfile.png"
        """
        return os.path.abspath(os.path.join(self._path_to_experiment_directory, relative_path))

    def log(self, msg: str):
        """writes a new message to the log file for this experiment and prints
        the message to the console"""
        formatted_msg = '%s : %s' % (
                datetime.now().strftime('%d-%m-%Y (%H:%M:%S.%f)'), msg)
        print(formatted_msg)
        self._log_file.write(formatted_msg + '\n')

    def close(self):
        """closes the underlying logfile"""
        self._log_file.close()

    def get_path_to_experiment_directory(self) -> str:
        """returns the path to the directory where this experiments stores its
        log files"""
        return self._path_to_experiment_directory

    def __del__(self):
        self.close()

