# -*- coding: utf-8 -*-

import os
import math
import itertools
import pandas as pd
from runstats import Statistics
import seaborn as sns
from shsplots.matplotlib.plot_context import PlotContext

class _Colors:
    """constants for colors in the result collector plots"""
    GOLD = sns.xkcd_rgb["gold"]
    BLACK = sns.xkcd_rgb["black"]
    GREEN = sns.xkcd_rgb["green"]
    BLUE = sns.xkcd_rgb["blue"]
    RED = sns.xkcd_rgb["red"]
    ORANGE = sns.xkcd_rgb["orange"]
    GREY = sns.xkcd_rgb["grey"]
    OLIVE = sns.xkcd_rgb["olive"]
    CYAN = sns.xkcd_rgb["cyan"]
    BROWN = sns.xkcd_rgb["brown"]
    LIGHT_BROWN = sns.xkcd_rgb["light brown"]
    LIME = sns.xkcd_rgb["lime"]
    PINK = sns.xkcd_rgb["pink"]
    LIGHT_BLUE = sns.xkcd_rgb["light blue"]
    BLUISH_PURPLE = sns.xkcd_rgb["bluish purple"]

class _Markers:
    """constants for markers in the result collector plots"""
    POINT = "."
    PIXEL = ","
    CIRCLE = "o"
    STAR = "*"
    X = "X"
    PLUS = "P"
    TRIANGLE_DOWN = "v"
    TRIANGLE_LEFT = "<"
    TRIANGLE_RIGHT = ">"
    TRIANGLE_UP = "^"
    SQUARE = "s"
    HEXAGON = "H"
    PENTAGON = "p"
    DIAMOND = "D"

class ResultCollector():

    colors = _Colors()
    markers = _Markers()

    def __init__(self, x_name: str):
        self.__result_type_dict = {}
        self.__color_dict = {}
        self.__marker_dict = {}
        self.__x_name = x_name

    def add_result(self, miner_name: str, result_type: str,
                   x: int, value: float):
        if result_type not in self.__result_type_dict:
            self.__result_type_dict[result_type] = {}
        miner_dict = self.__result_type_dict[result_type]
        if miner_name not in miner_dict:
            miner_dict[miner_name] = {}
        x_dict = miner_dict[miner_name]
        if x not in x_dict:
            x_dict[x] = Statistics()
        x_dict[x].push(value)

    def save_plots(self, directory: str):
        for result_type, miner_dict in self.__result_type_dict.items():
            for show_legend, show_stddev, show_stderr in itertools.product([True, False], repeat=3):
                if not (show_stderr and show_stddev):
                    self.__save_plots_for_result_type(
                            result_type, miner_dict, directory,
                            show_legend, show_stddev, show_stderr)

    def __save_plots_for_result_type(
            self, result_type: str, miner_dict, directory: str,
            show_legend: bool, show_stddev: bool, show_stderr: bool):
        plot_filepath = self.__determine_plot_filepath(
            directory, result_type, show_legend, show_stddev, show_stderr)

        plot_data, markers, colors = self.__get_plot_data(miner_dict)
        dashes = [
            "",
            (4, 1.5),
            (1, 1),
            (3, 1, 1.5, 1),
            (5, 1, 1, 1),
            (5, 1, 2, 1, 2, 1),
            (2, 2, 3, 1.5),
            (1, 2.5, 3, 1.2),
            (2, 0.3, 0.7, 0.5),
            "",
            (4, 1.5),
            (1, 1),
            (3, 1, 1.5, 1),
            (5, 1, 1, 1),
            (5, 1, 2, 1, 2, 1),
            (2, 2, 3, 1.5),
            (1, 2.5, 3, 1.2),
            (2, 0.3, 0.7, 0.5)
        ][:len(markers)]

        with PlotContext(show=False, use_tight_layout=True,
                         filepath=plot_filepath) as plot_context:
            plot = sns.lineplot(data=plot_data, markers=markers, dashes=dashes,
                                palette=colors, ax=plot_context.get_axes())

            min_value, max_value = self.__plot_values_for_miners(
                miner_dict, show_stddev, show_stderr, show_legend, plot,
                directory, result_type)

            #make sure that plots with values between 0 and 1 always have the
            #same scale for better comparison between different plots
            if min_value >= 0 and max_value <= 1:
                plot.set_ylim(0, 1)

            plot.set(xlabel=self.__x_name, ylabel=result_type)

            if show_legend:
                plot.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
            else:
                plot.legend().set_visible(False)

    def __plot_values_for_miners(
            self, miner_dict, show_stddev: bool, show_stderr: bool,
            show_legend: bool, plot, directory: str, result_type: str):
        i = 0
        min_value = float('inf')
        max_value = float('-inf')
        for miner_name, x_dict in sorted(miner_dict.items()):
            stddevs_for_miner = []
            stderrors_for_miner = []
            x_list = []
            values_of_miner = []
            for x, statistics in sorted(x_dict.items()):
                if show_stddev and len(statistics) > 1:
                    self.__plot_stddev_bar(
                        plot, miner_name, i, len(miner_dict), x, statistics)
                if show_stderr and len(statistics) > 1:
                    self.__plot_stderr_bar(
                        plot, miner_name, i, len(miner_dict), x, statistics)
                if len(statistics) > 1:
                    stddevs_for_miner.append(statistics.stddev())
                    stderrors_for_miner.append(statistics.stddev() /
                                            math.sqrt(len(statistics)))
                x_list.append(x)
                values_of_miner.append(statistics.mean())
            min_value = min(min_value, min(values_of_miner))
            max_value = max(max_value, max(values_of_miner))
            if show_stddev and show_legend and not show_stderr:
                self.__save_values_to_text_file(
                    miner_name, directory, result_type, x_list, values_of_miner,
                    stddevs_for_miner, stderrors_for_miner)
            i += 1
        return min_value, max_value

    def __plot_stddev_bar(
            self, plot, miner_name: str, miner_index: int, nr_of_miners: int,
            x, statistics_at_x: Statistics):
        plot.vlines(
                x=self.__get_x_of_bar(plot, x, miner_index, nr_of_miners),
                ymin=(statistics_at_x.mean() - abs(statistics_at_x.stddev())),
                ymax=(statistics_at_x.mean() + abs(statistics_at_x.stddev())),
                color=self.__color_dict.get(miner_name, _Colors.BLACK))

    def __plot_stderr_bar(
            self, plot, miner_name: str, miner_index: int, nr_of_miners: int,
            x, statistics_at_x: Statistics):
        stderr = statistics_at_x.stddev() / math.sqrt(len(statistics_at_x))
        plot.vlines(
                x=self.__get_x_of_bar(plot, x, miner_index, nr_of_miners),
                ymin=(statistics_at_x.mean() - abs(stderr)),
                ymax=(statistics_at_x.mean() + abs(stderr)),
                color=self.__color_dict.get(miner_name, _Colors.BLACK))

    def __get_x_of_bar(self, plot, x, miner_index: int, nr_of_miners: int):
        dodge_index = (miner_index - nr_of_miners // 2)
        #if the number of miners is even, then there is no center index
        #=> shift right half (including center 0) of indices to the right
        if dodge_index >= 0 and nr_of_miners % 2 == 0:
            dodge_index += 1
        pixel_to_coordinate = plot.transData.inverted().transform
        bar_distance = pixel_to_coordinate((4,0))[0] - pixel_to_coordinate((0,0))[0]
        return x + dodge_index * bar_distance

    def __save_values_to_text_file(
            self, miner_name: str, directory: str, result_type: str, x_list,
            values_of_miner, stddevs_for_miner, stderrors_for_miner):
        with open(os.path.join(directory, result_type + '.txt'), 'a') as f:
            f.write('%s: %r\n' % (self.__x_name, x_list))
            f.write('values for %s: %s\n' % (
                    miner_name, ', '.join(
                            '%.2f' % x for x in values_of_miner)))
            f.write('stddevs for %s: %s\n' % (
                    miner_name, ', '.join(
                            '%.2f' % x for x in stddevs_for_miner)))
            f.write('stderrs for %s: %s\n' % (
                    miner_name, ', '.join(
                            '%.2f' % x for x in stderrors_for_miner)))

    def __determine_plot_filepath(
            self, directory, result_type, show_legend, show_stddev, show_stderr):
        plot_name = result_type
        if show_legend:
            plot_name += '_with_legend'
        if show_stddev:
            plot_name += '_with_stddev'
        if show_stderr:
            plot_name += '_with_stderr'
        plot_name += '.pdf'
        return os.path.join(directory, plot_name)

    def __get_plot_data(self, miner_dict):
        plot_data = {}
        colors = []
        markers = []
        for miner_name, x_dict in miner_dict.items():
            x_list, values = zip(*sorted([
                (x, statistics.mean())
                for x, statistics in x_dict.items()
            ]))
            plot_data[miner_name] = values
            colors.append(self.__color_dict.get(miner_name, _Colors.BLACK))
            markers.append(self.__marker_dict.get(miner_name, _Markers.X))
        try:
            plot_data = pd.DataFrame.from_dict(plot_data)
        except ValueError:
            #ValueError is thrown if the miners have for this key a different
            #number of values. this usually happens if one of the miners
            #crashes during the experiments and delivers no outputs.
            #to prevent loss of the other values, we remove the deffective
            #miners for this plot.
            max_nr_of_values = max(len(v) for v in plot_data.values())
            plot_data = pd.DataFrame.from_dict({
                k: v for k,v in plot_data.items() if len(v) == max_nr_of_values
            })

        plot_data.index = x_list
        return plot_data, markers, colors

    def set_color(self, miner_name: str, color: str):
        self.__color_dict[miner_name] = color

    def set_marker(self, miner_name: str, marker: str):
        self.__marker_dict[miner_name] = marker