# -*- coding: utf-8 -*-

from typing import Tuple, Dict, Hashable

from math import log2
from math import log as ln
from scipy.special import betaln # type: ignore
from scipy.special import gammaln as lgamma # type: ignore

from lru import LRU

_PRECOMPUTED_LN: List[float] = []
_PRECOMPUTED_SUM_LOG_I_FROM_1_TO_N: List[float] = []
_LGAMMA_CACHE = LRU(1000)

def L_N(n: int) -> float:
    """returns the length of the code length of encoding a positive integer n
    with optimal prefix codes.
    https://www.researchgate.net/publication/38358970_A_Universal_Prior_for_Integers_and_Estimation_by_Minimum_Description_Length

    L_N(n) = log(n) + loglog(n) + logloglog(n) + ... + log(c_0)
    """
    if n < 1:
        raise ValueError('n must be a positive integer, but is %r' % n)
    try:
        return _PRECOMPUTED_LN[n-1]
    except IndexError:
        # intialize code length to c_0 of the paper
        # c_0 is a constant based on kraft inequalitiy and ensures that the probabilites
        # of all numbers sums up to 1
        code_length = log2(2.8565064)
        summand = log2(n)
        while summand > 0:
            code_length += summand
            summand = log2(summand)
        return code_length

def log2binom(n: int, k: int) -> float:
    """computes and returns log2(binom(n,k))
    """
    #https://stackoverflow.com/questions/21767690/python-log-n-choose-k
    #divided by ln(2) to transform to log_2
    return (-betaln(1 + n - k, 1 + k) - ln(n + 1)) / ln(2)

def log2multinom(n: int, ks) -> Tuple[int]:
    """computes the logarithm of a multinomial"""
    if n != sum(ks):
        raise ValueError('n (%d) must be the sum of all ks %r' % (n, ks))
    result = lgamma(n+1) #ln(n!)
    for k in ks:
        result -= lgamma(k+1)
    return result / ln(2)

def L_U(m: int, n: int) -> float:
    """computes log2binom(m-1,n-1), i.e.
    the encoded length for the number of combinations summing to m with n
    non-zero terms. L_U(0,0) := 0.
    """
    if m == 0 and n == 0:
        return 0
    return log2binom(m-1, n-1)

def prequential_coding_length(
        counts: Dict[Hashable,int], epsilon = 0.5) -> float:
    """computes the length in bits that is needed to encode a discrete sequence
    using prequential codes.

    Args:
        counts:
            a dictionary of the alphabet of the sequence. every symbol in the
            alphabet is assigned to its usage in the sequence.
        epsilon:
            the initial usage (epsilon in literature) of all symbols.
            default is 0.5
    """
    nr_of_symbols = len(counts)
    length_of_sequence = 0
    for i in counts.values():
        length_of_sequence += i

    if nr_of_symbols <= 1 or length_of_sequence == 0:
        return 0

    total_length = cached_lgamma(epsilon * nr_of_symbols + length_of_sequence)
    total_length -= cached_lgamma(epsilon * nr_of_symbols)
    for symbol_usage in counts.values():
        total_length -= (cached_lgamma(epsilon + symbol_usage) - cached_lgamma(epsilon))
    total_length = total_length / ln(2)

    return total_length

def cached_lgamma(x: float) -> float:
    """
    computes the lgamma function and caches the result in a LRU cache
    """
    try:
        return _LGAMMA_CACHE[x]
    except KeyError:
        y = lgamma(x)
        _LGAMMA_CACHE[x] = y
        return y

def sum_log_i_from_1_to_n(n: int) -> float:
    """
    computes log(1) + log(2) + log(3) + ... + log(n)

    Parameters
    ----------
    n : int
        must be > 0, otherwise a ValueError is raised

    Returns
    -------
    float
        log(1) + log(2) + log(3) + ... + log(n)
    """
    if n < 1:
        raise ValueError('n must not be < 1 but was %d' % n)
    try:
        return _PRECOMPUTED_SUM_LOG_I_FROM_1_TO_N[n-1]
    except IndexError:
        return lgamma(n+1) / ln(2)

_PRECOMPUTED_LN = [L_N(n) for n in range(1, 101)]
_PRECOMPUTED_SUM_LOG_I_FROM_1_TO_N = [
    sum_log_i_from_1_to_n(n) for n in range(1, 101)
]