from typing import Dict

import os
import subprocess
import tempfile
import pathlib
import json
from random import Random

from prolothar_common.models.eventlog import EventLog, Trace, Event

HERE = pathlib.Path(__file__).parent

NR_OF_SAMPLES = 500
RANDOM_SEED = 42

def sample_configuration(random: Random) -> Dict:
    return {
        'scale': random.choice([0.5, 0.75, 1, 1.25, 1.5, 2]),
        'v': random.choice([True] + [False] * 9),
        'h': random.choice([True] + [False] * 9),
        'dryRun': random.choice([True] + [False] * 9),
        'skipUpscaling': random.choice([True, False]),
        'clean': random.choice([True, False]),
        'androidMipmapInsteadOfDrawable': random.choice([True, False]),
        'androidIncludeLdpiTvdpi': random.choice([True, False]),
        'antiAliasing': random.choice([True, False]),
        'roundingMode': random.choice(['round', 'floor', 'ceil']),
        'outCompression': random.choice(['png', 'jpg', 'gif', 'bmp']),
        'compressionQuality': random.random()
    }

if __name__ == '__main__':
    random_generator = Random(RANDOM_SEED)
    with tempfile.TemporaryDirectory() as temporary_directory:
        class_level_log = EventLog()
        method_level_log = EventLog()

        for i in range(NR_OF_SAMPLES):
            print('run configuration %d of %d' % (i+1, NR_OF_SAMPLES))
            configuration = sample_configuration(random_generator)
            class_level_events = []
            method_level_events = []

            arguments = [
                'E:\\Entwicklung\\Java\\jdk-15.0.2\\bin\\java',
                '-javaagent:btrace-2.2.0\libs\\btrace-agent.jar=script=prolothar_data_public\\dconvert\\BTraceDensityConverter.class,stdout=true,debug=false',
                '-jar', 'prolothar_data_public\\dconvert\\dconvert-1.0.0-alpha7.jar',
                '-src', 'prolothar_data_public\\dconvert\\test_images\\svg_example_512.svg',
                '-dst', temporary_directory,
                '-threads', '1'
            ]
            for arg_name, arg_value in configuration.items():
                if isinstance(arg_value, bool):
                    if arg_value:
                        arguments.append('-' + arg_name)
                else:
                    arguments.append('-' + arg_name)
                    arguments.append(str(arg_value))
            btrace = subprocess.run(
                arguments, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                universal_newlines=True, check=True)

            for line in btrace.stdout.split('\n'):
                if line.startswith('at.favre'):
                    fully_qualified_method_name = line.replace('at.favre.tools.dconvert.', '')
                    method_level_events.append(fully_qualified_method_name)
                    class_name = fully_qualified_method_name.split('.')[-2:-1][0]
                    if not class_level_events or class_level_events[-1] != class_name:
                        class_level_events.append(class_name)

            class_level_log.add_trace(Trace(
                i, [Event(e) for e in class_level_events], configuration))
            method_level_log.add_trace(Trace(
                i, [Event(e) for e in method_level_events], configuration))

        with open(os.path.join(HERE, 'class_level_log.json'), 'w') as f:
            json.dump(class_level_log.to_dict(), f)
        with open(os.path.join(HERE, 'method_level_log.json'), 'w') as f:
            json.dump(method_level_log.to_dict(), f)
