# -*- coding: utf-8 -*-

from typing import Dict

import os
import pickle
import pathlib

from prolothar_common.models.dataset import TargetSequenceDataset
from prolothar_common.models.dataset.instance import TargetSequenceInstance
from prolothar_common.models.eventlog import EventLog, Trace, Event

HERE = pathlib.Path(__file__).parent

def read_all() -> EventLog:
    log = EventLog()
    log.add_traces(read_trainset_log().traces)
    log.add_traces(read_devset_log().traces)
    log.add_traces(read_testset_log().traces)
    return log

def read_trainset_log() -> EventLog:
    return _read_dataset(HERE / 'trainset.csv', 'trainset')

def read_devset_log() -> EventLog:
    return _read_dataset(HERE / 'devset.csv', 'devset')

def read_testset_log() -> EventLog:
    return _read_dataset(HERE / 'testset_w_refs.csv', 'testset')

def _read_dataset(path_to_dataset: str, dataset_name: str) -> EventLog:
    log = EventLog()
    with open(path_to_dataset, encoding='utf-8') as f:
        i = 1
        for line in f.readlines()[1:]:
            if line:
                log.add_trace(_parse_trace(line, '%s_%d' % (dataset_name, i)))
                i += 1
    return log

def _parse_trace(line: str, trace_id: str) -> Trace:
    #"name[Blue Spice], eatType[coffee shop], area[city centre]","A coffee shop in the city centre area called Blue Spice."
    attributes, utterance = line.split('","' if '","' in line else '",')
    #"name[Blue Spice], eatType[coffee shop], area[city centre]
    attributes = _parse_attributes(attributes[1:])
    #A coffee shop in the city centre area called Blue Spice."
    utterance = utterance.rstrip('"')
    utterance = utterance.replace(attributes.pop('name'), 'NAME')
    if 'near' in attributes:
        utterance = utterance.replace(attributes.pop('near'), 'NEAR')
    events = [Event(word.rstrip('"')) for word in utterance.split()]
    return Trace(trace_id, events, attributes=attributes)

def _parse_attributes(attribute_line: str) -> Dict[str, str]:
    #name[Blue Spice], eatType[coffee shop], area[city centre]
    attributes = {}
    for attribute_str in attribute_line.split(', '):
        #eatType[coffee shop]
        name, value = attribute_str[:-1].split('[')
        attributes[name] = value
    return attributes

def load_dataset() -> TargetSequenceDataset:
    path_to_dataset = os.path.join(HERE, 'e2e_dataset.pickle')

    if os.path.exists(path_to_dataset):
        with open(path_to_dataset, 'rb') as f:
            dataset = pickle.load(f)
    else:
        all_attributes = set()
        for trace in read_all():
            all_attributes.update(trace.attributes.keys())
        dataset = TargetSequenceDataset(all_attributes, [])

        for trace in read_all():
            instance_attributes = {}
            instance_attributes = {
                attribute: trace.attributes.get(attribute, 'None').replace(' ', '_')
                for attribute in all_attributes
            }
            dataset.add_instance(TargetSequenceInstance(
                trace.get_id(), instance_attributes, trace.to_activity_list()
            ))

        try:
            with open(path_to_dataset, 'wb') as f:
                pickle.dump(dataset, f)
        except PermissionError as e:
            print('No permission to write E2E dataset to disk for caching:')
            print(e)

    return dataset