# -*- coding: utf-8 -*-

import os
import pathlib
import pickle
import pandas as pd

from prolothar_common.models.eventlog import EventLog
from prolothar_common.models.dataset import TargetSequenceDataset
from prolothar_common.models.dataset.instance import TargetSequenceInstance

HERE = pathlib.Path(__file__).parent

def read_log():
    log = pd.read_csv(str(HERE / 'helpdesk.csv'), sep=',')
    log['Complete Timestamp'] = pd.to_datetime(log['Complete Timestamp'])
    log.drop(labels=['Variant', 'Variant index', 'seriousness'], inplace=True, axis=1)
    log.rename(columns={
            'Case ID': 'CaseID',
            'seriousness_2': 'seriousness',
            'Resource': 'resource',
            'Complete Timestamp': 'CompleteTimestamp'}, inplace=True)
    for column in log.columns:
        if pd.api.types.is_string_dtype(log[column].dtype):
            log[column] = log[column].str.replace(' ', '')

    log = EventLog.create_from_pandas_df(
        log, 'CaseID', 'Activity',
        event_attribute_columns=[
            'resource', 'CompleteTimestamp', 'seriousness', 'service_level'],
        trace_attribute_columns=[
            'customer', 'product', 'responsible_section', 'service_type',
            'support_section', 'workgroup'
        ]
    )

    return log

def load_dataset() -> TargetSequenceDataset:
    path_to_dataset = os.path.join(HERE, 'helpdesk_dataset.pickle')

    if os.path.exists(path_to_dataset):
        with open(path_to_dataset, 'rb') as f:
            dataset = pickle.load(f)
    else:
        dataset = TargetSequenceDataset(
            ['customer', 'product', 'responsible_section', 'service_type',
             'support_section', 'workgroup', 'initial_seriousness',
             'initial_service_level', 'initial_resource'],
            []
        )

        for trace in read_log():
            instance_attributes = {}
            for attribute in dataset.get_categorical_attribute_names():
                if attribute.startswith('initial_'):
                    instance_attributes[attribute] = trace.events[0].attributes[
                        attribute.replace('initial_', '')]
                else:
                    instance_attributes[attribute] = trace.attributes[
                        attribute]
            dataset.add_instance(TargetSequenceInstance(
                trace.get_id(), instance_attributes, trace.to_activity_list()
            ))

        try:
            with open(path_to_dataset, 'wb') as f:
                pickle.dump(dataset, f)
        except PermissionError as e:
            print('No permission to write helpdesk dataset to disk for caching:')
            print(e)

    return dataset
