# -*- coding: utf-8 -*-

import os
import pathlib
import pickle
from random import Random
import pandas as pd

from prolothar_common.models.eventlog import EventLog
from prolothar_common.models.dataset import TargetSequenceDataset
from prolothar_common.models.dataset.instance import TargetSequenceInstance

HERE = pathlib.Path(__file__).parent

def load_log_for_experiments(
        max_number_of_traces: int =-1, random: Random = Random()):
    log = pd.read_csv(str(HERE / 'Production_Data.csv'), sep=',')
    log['Start Timestamp'] = pd.to_datetime(log['Start Timestamp'])
    log['Complete Timestamp'] = pd.to_datetime(log['Complete Timestamp'])
    log.rename(columns={
            'Case ID': 'CaseID',
            'Part Desc.': 'PartDesc',
            'Work Order  Qty': 'WorkOrderQuantity',
            'Start Timestamp': 'StartTimestamp',
            'Complete Timestamp': 'CompleteTimestamp'}, inplace=True)

    log = EventLog.create_from_pandas_df(
            log, 'CaseID', 'Activity',
            event_attribute_columns=['Resource', 'StartTimestamp', 'CompleteTimestamp'],
            trace_attribute_columns=['PartDesc', 'WorkOrderQuantity'])

    if max_number_of_traces > 0 and max_number_of_traces < log.get_nr_of_traces():
        random.shuffle(log.traces)
        log.traces = log.traces[:max_number_of_traces]

    log.derive_event_duration_by_end_date(
            'StartTimestamp', 'CompleteTimestamp', 'Duration')

    log.join_sucessive_events_with_same_activity(
        startdate_attribute='StartTimestamp', enddate_attribute='CompleteTimestamp',
        duration_attribute='Duration', raise_error_for_attributes=False)

    log.add_start_activity_to_every_trace('START')
    log.add_end_activity_to_every_trace('END')

    return log

def load_dataset() -> TargetSequenceDataset:
    path_to_dataset = os.path.join(HERE, 'production_dataset.pickle')

    if os.path.exists(path_to_dataset):
        with open(path_to_dataset, 'rb') as f:
            dataset = pickle.load(f)
    else:
        dataset = TargetSequenceDataset(
            ['PartDesc'],
            ['WorkOrderQuantity']
        )

        for trace in load_log_for_experiments():
            instance_attributes = {
                'PartDesc': trace.attributes.get('PartDesc', '').replace(',', '').replace(' ', '_'),
                'WorkOrderQuantity': int(trace.attributes['WorkOrderQuantity'])
            }
            dataset.add_instance(TargetSequenceInstance(
                trace.get_id(), instance_attributes, trace.to_activity_list()
            ))

        try:
            with open(path_to_dataset, 'wb') as f:
                pickle.dump(dataset, f)
        except PermissionError as e:
            print('No permission to write production dataset to disk for caching:')
            print(e)

    return dataset
