# -*- coding: utf-8 -*-
import os
import pathlib
import pickle

from pm4py.objects.log.importer.xes import importer as xes_import_factory

import prolothar_common.pm4py_utils as pm4py_utils

from prolothar_common.models.eventlog import Event
from prolothar_common.models.dataset import TargetSequenceDataset
from prolothar_common.models.dataset.instance import TargetSequenceInstance

HERE = pathlib.Path(__file__).parent

def read_log():
    path_to_eventlog = os.path.join(HERE, 'sepsis_log.pickle')

    if os.path.exists(path_to_eventlog):
        with open(path_to_eventlog, 'rb') as f:
            log = pickle.load(f)
    else:
        xes = xes_import_factory.apply(str(HERE / 'Sepsis Cases - Event Log.xes'))

        log = pm4py_utils.convert_pm4py_log(xes)

        for trace in log.traces:
            trace.events.insert(0, Event(
                    'START', attributes=dict(trace.events[0].attributes)))
            trace.events.append(Event(
                    'END', attributes=dict(trace.events[-1].attributes)))

        log.traces = [
            trace for trace in log.traces
            if trace.contains_activity('ER Registration') and (
                trace.contains_activity('Release A') or
                trace.contains_activity('Release B') or
                trace.contains_activity('Release C') or
                trace.contains_activity('Release D') or
                trace.contains_activity('Release E')
            )
        ]

        try:
            with open(path_to_eventlog, 'wb') as f:
                pickle.dump(log, f)
        except PermissionError as e:
            print('No permission to write Sepsis event log to disk for caching:')
            print(e)

    return log

def read_log_without_self_loops():
    log = read_log()
    log.join_sucessive_events_with_same_activity(raise_error_for_attributes=False)
    return log

def load_dataset() -> TargetSequenceDataset:
    path_to_dataset = os.path.join(HERE, 'sepsis_dataset.pickle')

    if os.path.exists(path_to_dataset):
        with open(path_to_dataset, 'rb') as f:
            dataset = pickle.load(f)
    else:
        dataset = TargetSequenceDataset(
            ['InfectionSuspected', 'DiagnosticBlood', 'DisfuncOrg', 'SIRSCritTachypnea',
             'Hypotensie', 'SIRSCritHeartRate', 'Infusion', 'DiagnosticArtAstrup',
             'DiagnosticIC', 'DiagnosticSputum', 'DiagnosticLiquor',
             'DiagnosticOther', 'SIRSCriteria2OrMore', 'DiagnosticXthorax',
             'SIRSCritTemperature', 'DiagnosticUrinaryCulture', 'SIRSCritLeucos',
             'Oligurie', 'DiagnosticLacticAcid', 'Hypoxie', 'DiagnosticUrinarySediment',
             'DiagnosticECG'],
            ['Age']
        )

        for trace in read_log():
            instance_attributes = {}
            for attribute in dataset.get_categorical_attribute_names():
                for event in trace.events:
                    if event.activity_name == 'ER Registration':
                        instance_attributes[attribute] = event.attributes[attribute]
                        break
                else:
                    raise NotImplementedError('No "ER Registration" event found')
            instance_attributes['Age'] = int(trace.events[0].attributes['Age'])
            dataset.add_instance(TargetSequenceInstance(
                trace.get_id(), instance_attributes, trace.to_activity_list()
            ))

        try:
            with open(path_to_dataset, 'wb') as f:
                pickle.dump(dataset, f)
        except PermissionError as e:
            print('No permission to write Sepsis dataset to disk for caching:')
            print(e)

    return dataset
