# -*- coding: utf-8 -*-

from typing import Dict, List

import os
import tempfile
import subprocess
import pathlib

from prolothar_common.models.eventlog import EventLog
from prolothar_pattern_mining.thirdparty.sqs import write_sqs_input_file

HERE = pathlib.Path(__file__).parent

class InterestingSequenceMiner():
    """
    python interface to Interesting Sequence Miner (ISM).
    https://github.com/mast-group/sequence-mining
    """

    def find_patterns(self, log: EventLog) -> List[List[str]]:
        with tempfile.TemporaryDirectory() as temporary_directory:
            dict_activity_id = dict((activity,id) for id,activity
                                    in enumerate(log.compute_activity_set()))
            dict_id_activity = dict((str(v),k) for k, v in dict_activity_id.items())

            ism_input_file_path = self.__write_spmf_input_file(
                    log, temporary_directory, dict_activity_id)

            ism_output = self.__execute_ism(ism_input_file_path)

            return self.__parse_ism_output(ism_output, dict_id_activity)

    def __write_spmf_input_file(
            self, log: EventLog, temporary_directory: str,
            dict_activity_id: Dict[str,int]) -> str:
        spmf_input_file_path = os.path.join(temporary_directory,
                                           'log_spmf_data.txt')
        write_sqs_input_file(log, spmf_input_file_path, dict_activity_id,
                             symbol_separator=' -1 ', end_of_trace='-2\n')
        return spmf_input_file_path

    def __execute_ism(
            self, ism_input_file_path: str) -> str:
        ism_process = subprocess.run([
                'java', '-jar',
                str(HERE / 'sequence-mining-1.0.jar'),
                '-f', ism_input_file_path, '-v'
        ], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
           universal_newlines=True)
        if ism_process.returncode != 0:
            print('------------')
            print(ism_process.stdout)
            print(ism_process.stderr)
            print('------------')
            raise Exception('ISM returned exit code %d' %
                            ism_process.returncode)
        return ism_process.stdout

    def __parse_ism_output(
            self, stdout: str,
            dict_id_activity: Dict[str,str]) -> List[List[str]]:
        patterns = []
        pattern_section_started = False
        for line in stdout.split('\n'):
            print(line)
            if '[' in line and pattern_section_started:
                decoded_sequence = []
                #[2, 4, 5]       prob: 0,50000   int: 1,00000
                for encoded_activity in line[1:line.index(']')].split(', '):
                    decoded_sequence.append(dict_id_activity[encoded_activity])
                patterns.append(decoded_sequence)
            elif 'INTERESTING SEQUENCES' in line:
                pattern_section_started = True
        return patterns
