# -*- coding: utf-8 -*-

from typing import List

import tempfile
import os
import subprocess
import pathlib
import platform

HERE = pathlib.Path(__file__).parent

from prolothar_common.models.transactions import TransactionDatabase

class Pattern():
    def __init__(self, support: int = -1):
        self.support = support

class Singleton(Pattern):
    def __init__(self, item: str, support: int = -1):
        super().__init__(support = support)
        self.item = item

    def __str__(self) -> str:
        return '%s<%d>' % (self.item, self.support)

    def __repr__(self) -> str:
        return 'Singleton(%r)' % self.item

class PatternWithMultipleSubpatterns(Pattern):
    def __init__(self, operator: str, subpatterns: List[Pattern] = None,
                 support: int = -1):
        super().__init__(support = support)
        self.operator = operator
        self.subpatterns = subpatterns if subpatterns is not None else []

    def __str__(self) -> str:
        return '%s (%s)<%d>' % (self.operator, ', '.join(map(str, self.subpatterns)),
                                self.support)

    def __repr__(self) -> str:
        return self.__str__()

class Xor(PatternWithMultipleSubpatterns):
    def __init__(self, subpatterns: List[Pattern] = None, support: int = -1):
        super().__init__('XOR', subpatterns = subpatterns, support = support)

class And(PatternWithMultipleSubpatterns):
    def __init__(self, subpatterns: List[Pattern] = None, support: int = -1):
        super().__init__('AND', subpatterns = subpatterns, support = support)

class Mexican():
    """
    Jonas Fischer's algorithm for finding patterns in a transaction
    database.
    """

    def find_patterns(self, database: TransactionDatabase) -> List[Pattern]:
        with tempfile.TemporaryDirectory() as temporary_directory:
            input_file_path = os.path.join(temporary_directory, 'input.fic')
            output_file_path = os.path.join(temporary_directory, 'output.mex')
            id_to_item_dict = database.save_to_fic_file(input_file_path)
            self.__execute_mexican(input_file_path, output_file_path)
            return self.__parse_patterns(output_file_path, id_to_item_dict)

    def __execute_mexican(
            self, input_file_path: str, output_file_path: str):
        exe_name = 'MEX'
        if any(platform.win32_ver()):
            exe_name = 'MEX.exe'
        mexican_process = subprocess.run([
                os.path.join(HERE, 'mexican', exe_name),
                input_file_path, output_file_path
        ], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
           universal_newlines=True)
        if mexican_process.returncode != 0:
            print('------------')
            print(mexican_process.stdout)
            print(mexican_process.stderr)
            print('------------')
            raise Exception('mexican_process returned exit code %d' %
                            mexican_process.returncode)

    def __parse_patterns(self, output_file_path: str,
                         id_to_item_dict):
        patterns = []
        with open(output_file_path, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line:
                    #remove redundant support info at beginning of the line
                    patterns.append(self.__parse_pattern_line(
                        line.split('\t\t')[1], id_to_item_dict))
        return patterns

    def __parse_pattern_line(self, line: str, id_to_item_dict) -> Pattern:
        tokens = line.replace(',','').split()[::-1]
        node_stack = []
        while tokens:
            token = tokens.pop()
            if token == 'XOR':
                node_stack.append(Xor())
            elif token == 'AND':
                node_stack.append(And())
            elif token.startswith(']') or token.startswith(')'):
                #]<support>
                node_stack[-1].support = int(token[2:-1])
                if len(node_stack) > 1:
                    node_stack[-2].subpatterns.append(node_stack.pop())
            elif token.endswith('>'):
                node_stack[-1].subpatterns.append(
                    self.__parse_singleton(token, id_to_item_dict))
        return node_stack[-1]

    def __parse_singleton(self, token: str, id_to_item_dict) -> Singleton:
        #item<support>
        item_stop_index = token.index('<')
        return Singleton(id_to_item_dict[int(token[:item_stop_index])],
                         support = int(token[item_stop_index+1:-1]))


