#include <iostream>


#include "database.h"
#include "bitfactory.h"
#include "const.h"


Database::Database(const std::string& filePath, const FTYPE fileType) :
        isColumn(false)
{

    std::ifstream ifs (filePath);

    switch (fileType)
    {
        case (FICSIM) :
            readFicSimpleFile(ifs);
            break;

        default :
            std::cerr << "File type not supported. Aborting.\n\n";
            exit(1);
    }
}

Database::~Database()
{
    // Delete each element in the underlying data structure
    for (auto bcPointer : data)
    {
        delete bcPointer;
    }
}

void Database::determineDS()
{
    if (this->nones < CONST::SPARSETHRESH*this->nrow*this->ncol)
    {
        this->inUse = sparse;
    } else if (this->ncol <= 32)
    {
        this->inUse = u32;
    } else if (this->ncol <= 64)
    {
        this->inUse = u64;
    } else if (this->ncol <= 128)
    {
        this->inUse = a128;
    } else if (this->ncol <= 256)
    {
        this->inUse = a256;
    } else if (this->ncol <= 512)
    {
        this->inUse = a512;
    } else
    {
        this->inUse = gen;
    }
    // TODO: Remove this! Fixed DS type
    this->inUse = sparse;
}

void Database::readFicSimpleFile(std::ifstream& ifs)
{
    std::string line;
    size_t pos;
    size_t oldPos = 0;

    // Read fic version number - discard
    getline(ifs,line);
    // Read Meta information
    getline(ifs,line);
    // check if meta inf column
    pos = line.find_first_of(' ');
    // check if correct specifier
    std::string subCont(line, oldPos, pos - oldPos);
    if (subCont != "mi:")
    {
        std::cerr << "Wrong row specifier for fic format.\n"
            << "Expected Meta information column starting with \"mi:\"."
            << "Got \"" << subCont << "\"\n";
        exit(1);
    }
    oldPos = pos + 1;
    pos = line.find_first_of('=', oldPos);
    subCont = std::string(line, oldPos, pos - oldPos);
    if (subCont != "nR")
    {
        std::cerr << "Wrong row content for fic format.\n"
            << "Expected row count as meta information field, starting with \"nR=\"."
            << "Got \"" << subCont << "=\"\n";
        exit(1);
    }
    oldPos = pos + 1;
    pos = line.find_first_of(' ', oldPos);

    this->nrow = std::stoi(std::string(line, oldPos, pos - oldPos));
    oldPos = pos + 1;
    pos = line.find_first_of('=', oldPos);
    subCont = std::string(line, oldPos, pos - oldPos);
    if (subCont != "nT")
    {
        std::cerr << "Wrong row content for fic format.\n"
            << "Expected number of unique rows as meta information field, starting with \"nT=\"."
            << "Got \"" << subCont << "=\"\n";
        exit(1);
    }
    oldPos = pos + 1;
    pos = line.find_first_of(' ', oldPos);

    // set up database
    this->data.reserve(this->nrow);

    oldPos = pos + 1;
    pos = line.find_first_of('=', oldPos);
    subCont = std::string(line, oldPos, pos - oldPos);
    if (subCont != "nI")
    {
        std::cerr << "Wrong row content for fic format.\n"
            << "Expected count of ones in database as meta information field, starting with \"nI=\"."
            << "Got \"" << subCont << "=\"\n";
        exit(1);
    }
    oldPos = pos + 1;
    pos = line.find_first_of(' ', oldPos);
    this->nones = std::stoi(std::string(line, oldPos, pos - oldPos));
    oldPos = pos + 1;
    pos = line.find_first_of('=', oldPos);
    subCont = std::string(line, oldPos, pos - oldPos);
    if (subCont != "aS")
    {
        std::cerr << "Wrong row content for fic format.\n"
            << "Expected column counts meta information field, starting with \"aS=\"."
            << "Got \"" << subCont << "=\"\n";
        exit(1);
    }
    oldPos = pos + 1;
    pos = line.find_first_of(' ', oldPos);
    this->ncol = std::stoi(std::string(line, oldPos, pos - oldPos));
    oldPos = pos + 1;
    // Ignore sS
    pos = line.find_first_of(' ', oldPos);
    oldPos = pos + 1;
    pos = line.find_first_of('=', oldPos);
    subCont = std::string(line, oldPos, pos - oldPos);
    if (subCont != "mL")
    {
        std::cerr << "Wrong row content for fic format.\n"
            << "Expected maximum transaction length meta information field, starting with \"mL=\"."
            << "Got \"" << subCont << "=\"\n";
        exit(1);
    }
    oldPos = pos + 1;
    pos = line.find_first_of(' ', oldPos);
    this->mLen = std::stoi(std::string(line, oldPos, pos - oldPos));

    // Ignore ab line
    getline(ifs,line);

    // Get ac line
    getline(ifs,line);
    pos = line.find_first_of(' ');
    subCont = std::string(line, 0, pos);
    if (subCont != "ac:")
    {
        std::cerr << "Wrong row specifier for fic format.\n"
            << "Expected column margin counts starting with \"ac:\"."
            << "Got \"" << subCont << "\"\n";
        exit(1);
    }
    colMargins.reserve(ncol);
    for (uint32_t cIdx = 0; cIdx < ncol; ++cIdx)
    {
        oldPos = pos;
        pos = line.find_first_of(' ', oldPos+1);
        colMargins[cIdx] = std::stoi(std::string(line, oldPos, pos - oldPos));

    }
    // Get next line
    getline(ifs,line);
    pos = line.find_first_of(' ');
    subCont = std::string(line, 0, pos);
    // for fic 1.6 specification, watch out for ar
    if (subCont == "ar:")
    {
        // TODO: save ar info?

        // Get it line
        getline(ifs,line);
        pos = line.find_first_of(' ');
        subCont = std::string(line, 0, pos);
    }
    if (subCont != "it:")
    {
        std::cerr << "Wrong row specifier for fic format.\n"
            << "Expected database id mapping starting with \"it:\"."
            << "Got \"" << subCont << "\"\n";
        exit(1);
    }
    idMap.resize(ncol);
    for (uint32_t cIdx = 0; cIdx < ncol; ++cIdx)
    {
        oldPos = pos;
        pos = line.find_first_of(' ', oldPos+1);
        uint32_t extId = std::stoi(std::string(line, oldPos, pos - oldPos));
        idMap[cIdx] = extId;
        revIdMap[extId] = cIdx;

    }

    // check if cd and cl meta information row is present and skip
    getline(ifs,line);
    pos = line.find_first_of(' ');
    subCont = std::string(line, 0, pos);
    if (subCont == "cd:")
    {
        getline(ifs,line);
        pos = line.find_first_of(' ');
        subCont = std::string(line, 0, pos);
        if (subCont == "cl:")
        {
            getline(ifs,line);
        }
    }
    else if (subCont == "cl:")
    {
        getline(ifs,line);
    }

    // Set underlying data structure
    this->determineDS();

    // Read transactions
    for (uint32_t rIdx = 0; rIdx < nrow; ++rIdx)
    {
        pos = line.find_first_of(':');
        uint32_t nitems = std::stoi(std::string(line, 0, pos));
        pos += 1;
        std::vector<uint32_t> posOnes (nitems);
        for (uint32_t itemIdx = 0; itemIdx < nitems; ++itemIdx)
        {
            oldPos = pos;
            pos = line.find_first_of(' ', oldPos+1);
            posOnes[itemIdx] = std::stoi(std::string(line, oldPos, pos - oldPos));
        }
        this->data.push_back(static_cast<dbitvec*>(BitFactory::genBitrep(this->inUse, posOnes)));
        if (!getline(ifs,line) && rIdx < nrow-1)
        {
            std::cerr << "WARNING: #number of transactions (" << rIdx + 1 << ") differ from expected number of rows given by fic (" << nrow << ")\n\n";
            this->nrow = rIdx;
            break;
        }
    }

}



uint32_t Database::getTidCount(const dbitvec* h) const
{
    uint32_t tidCount = 0;
    for (BitCollection* t : data)
    {
        if (t->argIsSubset(*h))
            ++tidCount;
    }
    return tidCount;
}

uint32_t Database::getTidCountEstimate(const dbitvec* h) const
{
    uint32_t tidCount = this->nrow;
    const std::vector<uint32_t>* ones = h->getOnes();
    for (const uint32_t idx : *ones)
    {
        tidCount = std::min(tidCount, colMargins[idx]);
    }
    delete ones;
    return tidCount;
}


void Database::transposeToColumn()
{
    if (isColumn)
    {
        std::cerr << "Already in column representation!\n";
        return;
    }
    std::vector<dbitvec*> colRep (this->ncol);
    // init columns
    for (uint32_t j = 0; j < this->ncol; ++j)
    {
        colRep[j] = new dbitvec(this->nrow);
    }
    // transform
    for (uint32_t i = 0; i < this->nrow; ++i)
    {
        const auto ones = this->data[i]->getOnes();
        for (uint32_t j : *ones)
        {
            colRep[j]->setBit(i, true);
        }
        delete ones;
    }

    // delete old representation
    for (uint32_t i = 0; i < this->nrow; ++i)
    {
        delete this->data[i];
    }

    isColumn = true;
    this->data.swap(colRep);
}