#ifndef DATABASE_H
#define DATABASE_H

#include <vector>
#include <fstream>
#include <unordered_map>

#include "bitcollection.h"
#include "densebitvec.h"

// Enum for specifying the file type use to encode the database

enum FTYPE {FICSIM, FICCLA, FICBIN};

class Database
{

public:

    Database() = delete;

    Database(const std::string& filePath, const FTYPE fileType);

    ~Database();

    inline const DataType getDT() const
    {
        return this->inUse;
    }

    inline const uint32_t getColNum() const
    {
        return this->ncol;
    }

    inline const uint32_t getRowNum() const
    {
        return this->nrow;
    }

    // Read database in FIC format from input stream
    void readFicSimpleFile(std::ifstream& ifs);


    // get iterators for transactions in data base
    inline std::vector<dbitvec*>::const_iterator begin()
    {
        return data.cbegin();
    }
    inline std::vector<dbitvec*>::const_iterator end()
    {
        return data.cend();
    }

    // For given argument, return number of transactions containing this argument
    // ARGUMENTS:
    //          h       bit collection h which should be tested for subset relationship to transactions
    uint32_t getTidCount(const dbitvec* h) const;


    // For given argument, return optimistic estimate of number of transactions containing this argument
    // ARGUMENTS:
    //          h       bit collection h for which optimistic estimate for subset count should be computed
    uint32_t getTidCountEstimate(const dbitvec* h) const;

    // Access to id mapping
    uint32_t idMapping(const uint32_t i) const
    {
        return idMap[i];
    }

    // get the original feature id mapping
    inline const std::vector<uint32_t>& getIDMap() const
    {
        return idMap;
    }
    // get the original reverse id mapping
    inline const std::unordered_map<uint32_t, uint32_t>& getRevIDMap() const
    {
        return revIdMap;
    }

    // get database dimensions
    std::pair<uint32_t, uint32_t> getDimensions() const
    {
        return {nrow, ncol};
    }

    // transpose database to swap to column representation
    void transposeToColumn();


protected:

    // Determine type of the data structure and store it to inUse member variable.
    //  ARGUMENTS:
    //          fs      number of features/items/columns in database
    //          ts      number of samples/transactions/rows in database
    void determineDS();

    // each element corresponds to one transaction
    bool isColumn;
    std::vector<dbitvec*> data;


    // column margins of items
    std::vector<uint32_t> colMargins;

    // internal id mapping
    // i.e.: column one in this database representation corresponds to the column index idMap[0] in the original data
    std::vector<uint32_t> idMap;

    // reverse id mapping
    std::unordered_map<uint32_t, uint32_t> revIdMap;


    // number of rows and columns in database
    uint32_t nrow, ncol;
    // number of ones in data base
    uint32_t nones;
    // maximum number of ones in a transaction
    uint32_t mLen;

    // determines the data structure used to represent transactions and rules
    DataType inUse;

};

#endif // DATABASE_H