#include <limits.h>


#include "mdlalgo.h"


double MDLAlgo::dataCLPattern(const std::shared_ptr<Node>& n, const bool isRoot)
{
    double cl = 0;
    switch (n->op)
    {
        case XOR:
        {
            uint32_t rowCount = Pattern::nrow;
            for (const std::shared_ptr<Node> ch : n->children)
            {
                const uint32_t chCount = ch->tids->intersectionCount(*n->tids);
                cl += detail::logBinom(rowCount, chCount);
                rowCount -= chCount;
                // cl += dataCLPattern(ch, false);
            }
            break;
        }
        case AND:
        {
            // if (isRoot)
            // {
                const uint32_t holdCt = n->tids->getPopCount();
                cl += detail::logBinom(Pattern::nrow, holdCt);
            // }
            for (std::shared_ptr<Node> ch : n->children)
            {
                // cl += dataCLPattern(ch, false);
                if (ch->op == XOR)
                {
                    uint32_t rowCount = holdCt;
                    for (const std::shared_ptr<Node> chch : ch->children)
                    {
                        const uint32_t chCount = chch->tids->intersectionCount(*n->tids);
                        cl += detail::logBinom(rowCount, chCount);
                        rowCount -= chCount;
                    }
                }
            }
            break;
        }
        case NONE:
            cl += detail::logBinom(Pattern::nrow, n->tids->getPopCount());
            break;
        default:
            break;
    }
    return cl;
}


// TODO: merge singleton and normal pattern model costs!!
double MDLAlgo::modelCL(const std::vector<std::shared_ptr<Pattern> > patterns)
{
    double cl = detail::univCodeIntegerL(patterns.size());
    const uint32_t alphabetSize = Pattern::ncol;
    for (const auto p : patterns)
    {
        if (!p->getRoot()->tids->isEmpty())
        {
            cl += detail::logBinom(alphabetSize, p->getRoot()->items->getPopCount());
            // alphabetSize -= p->getRoot()->items->getPopCount();
            cl += modelCLPattern(p->getRoot(), true);
        }
    }
    return cl;
}
double MDLAlgo::modelCL(const std::list<std::shared_ptr<Pattern> > patterns)
{
    double cl = detail::univCodeIntegerL(patterns.size());
    const uint32_t alphabetSize = Pattern::ncol;
    for (const auto p : patterns)
    {
        if (!p->getRoot()->tids->isEmpty())
        {
            cl += detail::logBinom(alphabetSize, p->getRoot()->items->getPopCount());
            // alphabetSize -= p->getRoot()->items->getPopCount();
            cl += modelCLPattern(p->getRoot(), true);
        }
    }
    return cl;
}




double MDLAlgo::modelCLPattern(const std::shared_ptr<Node>& n, const bool isRoot)
{
    double cl = 0;
    // leaf case
    switch (n->op)
    {
        case NONE:
        {
            cl = 1;
            if (isRoot)
            {
                cl += detail::paramCompApprox(Pattern::nrow, 2);
            }
            return cl;
        }
        case AND:
        {
            std::vector<uint32_t> chSizes;
            for (const auto& ch : n->children)
            {
                chSizes.push_back(ch->items->getPopCount());
            }
            // 1 bit to say if leaf or inner node
            cl += 1;
            if (isRoot)
            {
                // transmit which operator this inner node is
                cl += 1;
                // transmit how many children the root has
                // and for each children how many children they have
                cl += detail::univCodeIntegerL(n->children.size());
                uint32_t remChs = n->children.size();
                for (const auto& ch : n->children)
                {
                    if (remChs == 0)
                    {
                        break;
                    }
                    if (ch->op == NONE)
                    {
                        continue;
                    }
                    cl += ceil(log2(remChs));
                    remChs -= ch->children.size();
                }
                // regret of binomial for where the pattern holds
                cl += detail::paramCompApprox(Pattern::nrow, 2);

                // recurse on xor children
                for (const auto& ch : n->children)
                {
                    if (ch->op == NONE)
                        continue;
                    uint32_t rowCount = n->tids->getPopCount();
                    for (const auto& chch : ch->children)
                    {
                        cl += detail::paramCompApprox(rowCount, 2);
                        rowCount -= chch->tids->intersectionCount(*n->tids);
                    }
                }
            }
            if (n->children.size() > 0)
            {
                // transmit how many items this node has
                cl += detail::paramCompApprox(n->items->getPopCount(), n->children.size()) +
                // transmit the items for each child
                detail::logMultinom(n->items->getPopCount(), chSizes);
            }
            break;
        }
        case XOR:
        {
            std::vector<uint32_t> chSizes;
            // if this is the root node, send the children accordingly
            // otherwise this already has been taking care of by AND root node
            uint32_t rows = Pattern::nrow;
            for (const auto& ch : n->children)
            {
                chSizes.push_back(ch->items->getPopCount());
                // costs for sending where child holds
                if (isRoot)
                {
                    cl += detail::paramCompApprox(rows, 2);
                    rows -= ch->tids->intersectionCount(*n->tids);
                }
            }
            // 1 bit to say if leaf or inner node
            cl += 1;
            if (isRoot)
            {
                // transmit which operator this inner node is
                cl += 1;
                // transmit how many children
                // and for each children how many children they have
                cl += detail::univCodeIntegerL(n->children.size());
                uint32_t remChs = n->children.size();
                for (const auto& ch : n->children)
                {
                    if (remChs == 0)
                    {
                        break;
                    }
                    if (ch->op == NONE)
                    {
                        continue;
                    }
                    cl += ceil(log2(remChs));
                    remChs -= ch->children.size();
                }
            }
            if (n->children.size() > 0)
            {
                // transmit how many items each child has
                cl += detail::paramCompApprox(n->items->getPopCount(), n->children.size()) +
                // transmit the items for each child
                detail::logMultinom(n->items->getPopCount(), chSizes);
            }
            break;
        }
    }


    // recurse on children
    for (const auto ch : n->children)
    {
        cl += modelCLPattern(ch, false);
    }
    return cl;
}