// #define NDEBUG
#include <cassert>
#include <algorithm>
#include <math.h>


#include "pattern.h"
#include "const.h"


uint64_t Pattern::nrow = 0;
uint64_t Pattern::ncol = 0;
std::shared_ptr<dbitvec> Pattern::fullData;

Pattern::Pattern() :
        root(std::make_shared<Node>(NONE, std::shared_ptr<SparseBitvec>(), std::vector<std::shared_ptr<Node> >(), std::shared_ptr<dbitvec>(new dbitvec(Pattern::nrow)), std::shared_ptr<dbitvec>(new dbitvec(Pattern::nrow)) ) )
    ,   depth(0)
{
}

Pattern::Pattern(unsigned int rootItem, std::shared_ptr<dbitvec> tidL) :
        root(std::make_shared<Node>(NONE, std::shared_ptr<SparseBitvec>(), std::vector<std::shared_ptr<Node> >(), tidL, tidL))
    ,   depth(0)
{
    std::vector<uint32_t> bits = {rootItem};
    root->items = std::shared_ptr<SparseBitvec>(new SparseBitvec(std::move(bits)));
}

Pattern::Pattern(std::shared_ptr<Node>&& r) :
        root(r)
    ,   depth(0)
{
}

std::shared_ptr<Pattern> Pattern::mergeNewOpRoot(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2, Operator op)
{
    assert(p1->depth <= 1);
    assert(p2->depth <= 1);
    assert(p1->root->op != op);
    assert(p2->root->op != op);
    // merge tid lists
    std::shared_ptr<dbitvec> tids;
    std::shared_ptr<dbitvec> tidsFiltered = NULL;
    switch (op)
    {
        case XOR:
        {
            tids = std::shared_ptr<dbitvec>(p1->root->tids->setsymdiff(*p2->root->tids));
            // std::cout << "\n\n";
            // p1->print(std::cout);
            // std::cout << "\n";
            // p2->print(std::cout);
            // std::cout << "\n";

            if (tids->getPopCount() < CONST::MINSUP*Pattern::nrow ||
                tids->getPopCount() < (1-CONST::MAXOVERLAP)*(p1->root->tids->getPopCount() + p2->root->tids->getPopCount()))
                return NULL;

            // test of how likely is gain assuming independence of features
            const uint64_t n12 = (p1->root->tids->getPopCount() + p2->root->tids->getPopCount() - tids->getPopCount())/2;
            // std::cout << "Overlap: " << n12 << "\n";

            const bool testPassed = MDLAlgo::detail::xorTest(Pattern::nrow, p1->root->tids->getPopCount(), p2->root->tids->getPopCount(), n12);
            if (!testPassed)
            {
                return NULL;
            }

            dbitvec* res = p1->root->tids->setunion(*p2->root->tids);
            res->thisAndNot(*tids);
            tidsFiltered = std::shared_ptr<dbitvec>(res);
            break;
        }
        case AND:
            if (p1->root->tids->getPopCount() < p2->root->tids->getPopCount() * CONST::MINOVERLAP ||
                p2->root->tids->getPopCount() < p1->root->tids->getPopCount() * CONST::MINOVERLAP)
                return NULL;
            tids = std::shared_ptr<dbitvec>(p1->root->tids->setinter(*p2->root->tids));
            if (tids->getPopCount() < CONST::MINSUP*Pattern::nrow ||
                tids->getPopCount() < CONST::MINOVERLAP * std::max(p1->root->tids->getPopCount(), p2->root->tids->getPopCount()))
                return NULL;
            break;

        default:
            break;
    }
    // generate new pattern
    std::shared_ptr<Node> root (
        new Node(op,
            std::shared_ptr<SparseBitvec>(p1->root->items->setunion(*p2->root->items)),
            {p1->root, p2->root},
            tids,
            tidsFiltered
        )
    );
    Pattern* p = new Pattern(std::move(root));
    p->depth = std::max(p1->depth, p2->depth) + 1;
    return std::shared_ptr<Pattern>(p);

}


std::shared_ptr<Pattern> Pattern::mergeHorizontal(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2, const int pos)
{
    // retrieve node of pattern 1 into which we want to merge
    std::shared_ptr<Pattern> p = p1->copy();
    std::shared_ptr<Node> n = p->root;
    if (pos != -1)
    {
        assert(n->children.size() > pos);
        n = n->children[pos];
    }
    std::shared_ptr<dbitvec> tids;
    std::shared_ptr<dbitvec> tidsFiltered;
    switch (n->op)
    {
        case XOR:
        {
            // TODO: This merge should not be useful if not root!
            if (pos != -1)
            {
                return NULL;
            }

            if (p1->root->tids->getPopCount() <= CONST::MINCOVER * Pattern::nrow ||
                p2->root->tids->getPopCount() <= CONST::MINCOVER * Pattern::nrow)
            {
                return NULL;
            }
            if (p2->root->op == XOR)
            {

                const uint64_t nOverlap = p1->root->tids->intersectionCount(*p2->root->tids) +
                p1->root->tidsFiltered->intersectionCount(*p2->root->tids) +
                p1->root->tids->intersectionCount(*p2->root->tidsFiltered) +
                p1->root->tidsFiltered->intersectionCount(*p2->root->tidsFiltered);
                // std::cout << "\n\n";
                // p1->print(std::cout);
                // std::cout << "\n";
                // p2->print(std::cout);
                // std::cout << "\n";
                const bool testPassed = MDLAlgo::detail::xorTest(Pattern::nrow, p1->root->tids->getPopCount() + p1->root->tidsFiltered->getPopCount(), p2->root->tids->getPopCount() + p2->root->tidsFiltered->getPopCount(), nOverlap);
                if (!testPassed)
                {
                    return NULL;
                }
                // retrieve tid list for joined n-ary xor
                tids = std::shared_ptr<dbitvec>(n->tids->xorJoin(*n->tidsFiltered, *p2->root->tids, *p2->root->tidsFiltered));
                // retrieve new filtered tid list
                std::vector<BitCollection*> tidL = {n->tids.get(), n->tidsFiltered.get(), p2->root->tids.get(), p2->root->tidsFiltered.get()};
                dbitvec* tidAll = n->tids->setunionMany(tidL.begin(), tidL.end());
                tidAll->thisAndNot(*tids);
                tidsFiltered = std::shared_ptr<dbitvec>(tidAll);
                if (tids->getPopCount() < CONST::MINSUP*Pattern::nrow ||
                    CONST::MAXOVERLAP*(tids->getPopCount() + tidsFiltered->getPopCount()) < tidsFiltered->getPopCount())
                    return NULL;
            } else if (p2->root->op == NONE) {


                const uint64_t nOverlap = p1->root->tids->intersectionCount(*p2->root->tids) +
                p1->root->tidsFiltered->intersectionCount(*p2->root->tids);
                // std::cout << "\n\n";
                // p1->print(std::cout);
                // std::cout << "\n";
                // p2->print(std::cout);
                // std::cout << "\n";
                const bool testPassed = MDLAlgo::detail::xorTest(Pattern::nrow, p1->root->tids->getPopCount() + p1->root->tidsFiltered->getPopCount(), p2->root->tids->getPopCount(), nOverlap);
                if (!testPassed)
                {
                    return NULL;
                }
                // retrieve tid list for joined n-ary xor
                dbitvec* tidRes = n->tids->setsymdiff(*p2->root->tids);
                tidRes->thisAndNot(*n->tidsFiltered);
                tids = std::shared_ptr<dbitvec>(tidRes);
                // retrieve new filtered tid list
                std::vector<BitCollection*> tidL = {n->tids.get(), n->tidsFiltered.get(), p2->root->tids.get()};
                dbitvec* tidAll = n->tids->setunionMany(tidL.begin(), tidL.end());
                tidAll->thisAndNot(*tids);
                tidsFiltered = std::shared_ptr<dbitvec>(tidAll);
                if (tids->getPopCount() < CONST::MINSUP*Pattern::nrow ||
                    CONST::MAXOVERLAP*(tids->getPopCount() + tidsFiltered->getPopCount()) < tidsFiltered->getPopCount())
                    return NULL;
            } else {
                assert(false);
            }
            break;
        }

        case AND:
            if (p2->root->op == NONE)
            {
                if (n->tids->getPopCount() < p2->root->tids->getPopCount() * CONST::MINOVERLAP ||
                    p2->root->tids->getPopCount() < n->tids->getPopCount() * CONST::MINOVERLAP)
                    return NULL;
            } else {
                for (const auto& ch : p2->root->children)
                {
                    if (n->tids->getPopCount() < ch->tids->getPopCount() * CONST::MINOVERLAP ||
                        ch->tids->getPopCount() < n->tids->getPopCount() * CONST::MINOVERLAP)
                        return NULL;
                }
            }
            tids = std::shared_ptr<dbitvec>(n->tids->setinter(*p2->root->tids));
            if (tids->getPopCount() < CONST::MINSUP*Pattern::nrow)
                return NULL;
            for (const auto& ch : n->children)
            {
                if (n->tids->getPopCount() < CONST::MINOVERLAP*ch->tids->getPopCount())
                    return NULL;
            }
            if (p2->root->op == NONE)
            {
                if (tids->getPopCount() < CONST::MINOVERLAP * std::max(n->tids->getPopCount(), p2->root->tids->getPopCount()))
                    return NULL;
            } else {
                for (const auto& ch : p2->root->children)
                {
                    if (n->tids->getPopCount() < CONST::MINOVERLAP*ch->tids->getPopCount())
                        return NULL;
                }
            }
            break;

        case NONE:
            if (p2->root->op == XOR)
            {

                if (p1->root->tids->getPopCount() <= CONST::MINCOVER * Pattern::nrow)
                {
                    return NULL;
                }
                const uint64_t nOverlap = p2->root->tids->intersectionCount(*p1->root->tids) +
                p2->root->tidsFiltered->intersectionCount(*p1->root->tids);
                // std::cout << "\n\n";
                // p1->print(std::cout);
                // std::cout << "\n";
                // p2->print(std::cout);
                // std::cout << "\n";
                const bool testPassed = MDLAlgo::detail::xorTest(Pattern::nrow, p2->root->tids->getPopCount() + p2->root->tidsFiltered->getPopCount(), p1->root->tids->getPopCount(), nOverlap);
                if (!testPassed)
                {
                    return NULL;
                }
                // retrieve tid list for joined n-ary xor
                dbitvec* tidRes = n->tids->setsymdiff(*p2->root->tids);
                tidRes->thisAndNot(*p2->root->tidsFiltered);
                tids = std::shared_ptr<dbitvec>(tidRes);
                // retrieve new filtered tid list
                std::vector<BitCollection*> tidL = {n->tids.get(), p2->root->tids.get(), p2->root->tidsFiltered.get()};
                dbitvec* tidAll = n->tids->setunionMany(tidL.begin(), tidL.end());
                tidAll->thisAndNot(*tids);
                tidsFiltered = std::shared_ptr<dbitvec>(tidAll);
                if (tids->getPopCount() < CONST::MINSUP*Pattern::nrow ||
                CONST::MAXOVERLAP*(tids->getPopCount() + tidsFiltered->getPopCount()) < tidsFiltered->getPopCount())
                    return NULL;

            } else if (p2->root->op == AND)
            {
                for (const auto& ch : p2->root->children)
                {
                    if (n->tids->getPopCount() < ch->tids->getPopCount() * CONST::MINOVERLAP ||
                        ch->tids->getPopCount() < n->tids->getPopCount() * CONST::MINOVERLAP)
                        return NULL;
                }
                tids = std::shared_ptr<dbitvec>(n->tids->setinter(*p2->root->tids));
                if (tids->getPopCount() < CONST::MINSUP*Pattern::nrow ||
                    tids->getPopCount() < CONST::MINOVERLAP * n->tids->getPopCount())
                    return NULL;
                for (const auto& ch : p2->root->children)
                {
                    if (tids->getPopCount() < CONST::MINOVERLAP * ch->tids->getPopCount())
                        return NULL;
                }

            } else {
                assert(false);
            }
            break;

        default:
            assert(false);
            break;
    }
    if (n->op == NONE)
    {
        n->children.push_back(n->copy());
        n->op = p2->root->op;
    }
    if (p2->root->op == NONE)
    {
        n->children.push_back(p2->root->copy());

    } else {

        for (auto ch : p2->root->children)
        {
            n->children.push_back(ch);
        }
    }
    n->items = std::shared_ptr<SparseBitvec>(n->items->setunion(*p2->root->items));
    n->tids = tids;
    n->tidsFiltered = tidsFiltered;
    if (pos != -1)
    {
        p->root->children[pos] = std::move(n);
        p->root->items->thisUnion(*p2->root->items);
        std::vector<const dbitvec*> childTids;
        for (const auto& ch : p->root->children)
        {
            childTids.push_back(ch->tids.get());
        }
        if (p->root->op == AND)
        {
            p->root->tids->interMany(childTids);
            if (p->root->tids->getPopCount() < CONST::MINSUP*Pattern::nrow)
                return NULL;
            for (const auto& ch : p->root->children)
            {
                if (p->root->tids->getPopCount() < ch->tids->getPopCount() * CONST::MINOVERLAP)
                    return NULL;
            }
        } else {

            std::shared_ptr<dbitvec>filt (p->root->tids->symdiffMany(childTids));
            p->root->tidsFiltered = filt;
            if (CONST::MAXOVERLAP*(p->root->tids->getPopCount() + p->root->tidsFiltered->getPopCount()) < p->root->tidsFiltered->getPopCount())
                return NULL;
        }
        if (p->root->tids->getPopCount() < CONST::MINSUP*Pattern::nrow)
            return NULL;
    } else {
        p->root = std::move(n);
    }
    p->depth = computeDepth(p->root);
    return p;
}



std::shared_ptr<Pattern> Pattern::mergeVertical(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2, const int pos, Operator op)
{
    // retrieve node of pattern 1 into which we want to merge
    std::shared_ptr<Pattern> p = p1->copy();
    std::shared_ptr<Node> n = p->root;
    if (pos != -1)
    {
        assert(n->children.size() > pos);
        n = n->children[pos];
    }
    std::shared_ptr<dbitvec> tids;
    std::shared_ptr<dbitvec> tidsFiltered;
    switch (op)
    {
        case XOR:
        {
            tids = std::shared_ptr<dbitvec>(n->tids->setsymdiff(*p2->root->tids));
            dbitvec* res = n->tids->setunion(*p2->root->tids);
            res->thisAndNot(*tids);
            tidsFiltered = std::shared_ptr<dbitvec>(res);
            if (tids->getPopCount() < CONST::MINSUP*Pattern::nrow ||
                CONST::MAXOVERLAP*(tids->getPopCount() + tidsFiltered->getPopCount()) < tidsFiltered->getPopCount())
                return NULL;
            break;
        }
        case AND:
            if (n->tids->getPopCount() < p2->root->tids->getPopCount() * CONST::MINOVERLAP ||
                p2->root->tids->getPopCount() < n->tids->getPopCount() * CONST::MINOVERLAP)
                return NULL;
            tids = std::shared_ptr<dbitvec>(n->tids->setinter(*p2->root->tids));
            if (tids->getPopCount() < CONST::MINSUP*Pattern::nrow ||
                tids->getPopCount() < CONST::MINOVERLAP * std::max(n->tids->getPopCount(), p2->root->tids->getPopCount()))
                return NULL;
            break;

        default:
            break;
    }
    // reassign to new merged node
    n = std::shared_ptr<Node>(
        new Node(
            op,
            std::shared_ptr<SparseBitvec>(n->items->setunion(*p2->root->items)),
            {n, p2->root},
            tids,
            tidsFiltered
        )
    );
    if (pos != -1)
    {
        p->root->children[pos] = std::move(n);
        p->root->items->thisUnion(*p2->root->items);
        std::vector<const dbitvec*> childTids;
        for (const auto& ch : p->root->children)
        {
            childTids.push_back(ch->tids.get());
        }
        if (p->root->op == AND)
        {
            p->root->tids->interMany(childTids);
            if (p->root->tids->getPopCount() < CONST::MINSUP*Pattern::nrow)
                return NULL;
            for (const auto& ch : p->root->children)
            {
                if (p->root->tids->getPopCount() < ch->tids->getPopCount() * CONST::MINOVERLAP)
                    return NULL;
            }
        } else {

            std::shared_ptr<dbitvec>filt (p->root->tids->symdiffMany(childTids));
            p->root->tidsFiltered = filt;
            if (CONST::MAXOVERLAP*(p->root->tids->getPopCount() + p->root->tidsFiltered->getPopCount()) < p->root->tidsFiltered->getPopCount())
                return NULL;
        }
        if (p->root->tids->getPopCount() < CONST::MINSUP*Pattern::nrow)
            return NULL;
    } else {
        p->root = std::move(n);
    }
    p->depth = computeDepth(p->root);
    return p;
}



std::shared_ptr<Pattern> Pattern::mergeBelow(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2)
{
    std::shared_ptr<Pattern> p = p1->copy();
    switch (p->root->op)
    {
        case XOR:
        {
            p->root->tids = std::shared_ptr<dbitvec>(p->root->tids->setsymdiff(*p2->root->tids));
            dbitvec* res = p1->root->tids->setunion(*p2->root->tids);
            res->thisAndNot(*p->root->tids);
            p->root->tidsFiltered = std::shared_ptr<dbitvec>(res);
            if (p->root->tids->getPopCount() < CONST::MINSUP*Pattern::nrow ||
                CONST::MAXOVERLAP*(p->root->tids->getPopCount() + p->root->tidsFiltered->getPopCount()) < p->root->tidsFiltered->getPopCount())
                return NULL;

            const uint64_t nOverlap = p1->root->tids->intersectionCount(*p2->root->tids) +
            p1->root->tidsFiltered->intersectionCount(*p2->root->tids);
            const bool testPassed = MDLAlgo::detail::xorTest(Pattern::nrow, p1->root->tids->getPopCount() + p1->root->tidsFiltered->getPopCount(), p2->root->tids->getPopCount(), nOverlap);
            if (!testPassed)
            {
                return NULL;
            }
            break;
        }
        case AND:
            if (p->root->tids->getPopCount() < p2->root->tids->getPopCount() * CONST::MINOVERLAP ||
                p2->root->tids->getPopCount() < p->root->tids->getPopCount() * CONST::MINOVERLAP)
                return NULL;
            p->root->tids->thisInter(*p2->root->tids);
            if (p->root->tids->getPopCount() < CONST::MINSUP*Pattern::nrow)
                return NULL;
            if (p->root->tids->getPopCount() < CONST::MINOVERLAP*p2->root->tids->getPopCount())
                return NULL;
            for (const auto& ch : p->root->children)
            {
                if (p->root->tids->getPopCount() < CONST::MINOVERLAP*ch->tids->getPopCount())
                    return NULL;
            }
            break;

        default:
            break;
    }
    p->root->children.push_back(p2->root);
    p->root->items->thisUnion(*p2->root->items);
    p->depth = computeDepth(p->root);
    return p;
}



unsigned int Pattern::computeDepth(std::shared_ptr<Node>& p)
{
    unsigned int depth = 0;
    for (std::shared_ptr<Node>& n : p->children)
    {
        depth = std::max(depth, computeDepth(n) + 1);
    }
    return depth;
}


std::list<Pattern::Candidate> Pattern::genCandidates(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2, const std::vector<std::shared_ptr<Pattern> >& sts, const std::vector<std::shared_ptr<Pattern> >& stsBase, const uint32_t numPat, const std::unordered_map<uint32_t, std::list<std::shared_ptr<Pattern> > >& stMap)
{

    std::list<Pattern::Candidate> cands;
    if (p1->root->items->intersectionCount(*p2->root->items) > 0)
    {
        return cands;
    }

    bool noXor = false;
    if (p1->root->tids->getPopCount() <= CONST::MINCOVER * Pattern::nrow ||
        p2->root->tids->getPopCount() <= CONST::MINCOVER * Pattern::nrow)
    {
        noXor = true;
    }
    // test merge of roots
    if (p1->root->op == p2->root->op)
    {
        // case of both singletons
        if (p1->root->op == NONE)
        {
            if (!noXor)
            {
                const auto newPatXor = Pattern::mergeNewOpRoot(p1, p2, XOR);
                insertIfGain(cands, p1, p2, newPatXor, sts, stsBase, numPat, stMap);
            }
            const auto newPatAnd = Pattern::mergeNewOpRoot(p1, p2, AND);
            insertIfGain(cands, p1, p2, newPatAnd, sts, stsBase, numPat, stMap);

        } else {

            // check for overlap in items
            // if (p1->root->items->intersectionCount(*p2->root->items) == 0)
            // {

                const auto newPatJoin = Pattern::mergeHorizontal(p1, p2, -1);
                insertIfGain(cands, p1, p2, newPatJoin, sts, stsBase, numPat, stMap);
                if (p1->depth == 1 && p2->depth == 1)
                {
                    // check that we have alternating operators in tree layers
                    if (!noXor && p1->root->op != XOR)
                    {
                        const auto newPatXor = Pattern::mergeNewOpRoot(p1, p2, XOR);
                        insertIfGain(cands, p1, p2, newPatXor, sts, stsBase, numPat, stMap);
                    }
                    if (p1->root->op != AND)
                    {
                        const auto newPatAnd = Pattern::mergeNewOpRoot(p1, p2, AND);
                        insertIfGain(cands, p1, p2, newPatAnd, sts, stsBase, numPat, stMap);
                    }
                }
            // }
        }
    // case that p2 is singleton
    } else if (p2->root->op == NONE)
    {
        const auto newPatJoin = Pattern::mergeHorizontal(p1, p2, -1);
        insertIfGain(cands, p1, p2, newPatJoin, sts, stsBase, numPat, stMap);
        if (p1->depth < 2)
        {
            if (!noXor && p1->root->op != XOR)
            {
                const auto newPatXor = Pattern::mergeNewOpRoot(p1, p2, XOR);
                insertIfGain(cands, p1, p2, newPatXor, sts, stsBase, numPat, stMap);
            } else if (p1->root->op != AND)
            {
                const auto newPatAnd = Pattern::mergeNewOpRoot(p1, p2, AND);
                insertIfGain(cands, p1, p2, newPatAnd, sts, stsBase, numPat, stMap);
            }
        }
    }

    // case of glueing together (p2 is of depth1 and different operator or vice versa)
    if (p1->depth >= 1 && p2->depth == 1 && p2->root->op != p1->root->op)
    {
        const auto newPat = Pattern::mergeBelow(p1, p2);
        insertIfGain(cands, p1, p2, newPat, sts, stsBase, numPat, stMap);
    }

    if (p1->depth == 1 && p2->depth >= 1 && p2->root->op != p1->root->op)
    {
        const auto newPat = Pattern::mergeBelow(p2, p1);
        insertIfGain(cands, p1, p2, newPat, sts, stsBase, numPat, stMap);
    }

    // fix p1, test for all merges of p2 into nodes of p1
    for (uint32_t l1 = 0; l1 < p1->root->children.size(); ++l1)
    {

        // case that child is leaf node
        if (p1->root->children[l1]->op == NONE)
        {
            // second is singleton
            if (p2->root->op == NONE)
            {
                if (p1->root->op != AND)
                {
                    const auto newPatAnd = Pattern::mergeVertical(p1, p2, l1, AND);
                    insertIfGain(cands, p1, p2, newPatAnd, sts, stsBase, numPat, stMap);
                }
                // else if (p1->root->op != XOR)
                // {
                //     const auto newPatXor = Pattern::mergeVertical(p1, p2, l1, XOR);
                //     insertIfGain(cands, p1, p2, newPatXor, sts, stsBase, numPat, stMap);
                // }

            } else if (p1->root->op != p2->root->op && p2->depth < 2)
            {
                const auto newPatJoin = Pattern::mergeHorizontal(p1, p2, l1);
                insertIfGain(cands, p1, p2, newPatJoin, sts, stsBase, numPat, stMap);

            }

        } else if (p1->root->children[l1]->op == p2->root->op && p2->depth < 2) {

            const auto newPatJoin = Pattern::mergeHorizontal(p1, p2, l1);
            insertIfGain(cands, p1, p2, newPatJoin, sts, stsBase, numPat, stMap);

        } else if (p2->root->op == NONE) {

            const auto newPatJoin = Pattern::mergeHorizontal(p1, p2, l1);
            insertIfGain(cands, p1, p2, newPatJoin, sts, stsBase, numPat, stMap);
        }
    }

    // fix p2, test for all merges of p1 into nodes of p2
    for (uint32_t l1 = 0; l1 < p2->root->children.size(); ++l1)
    {
        // check if one is subset of other
        if (p2->root->children[l1]->items->intersectionCount(*p1->root->items) ==
            std::min(p2->root->children[l1]->items->getPopCount(), p1->root->items->getPopCount()))
        {
            continue;
        }
        // case that child is leaf node
        if (p2->root->children[l1]->op == NONE)
        {
            // second is singleton
            if (p1->root->op == NONE)
            {
                if (p2->root->op != AND )
                {
                    const auto newPatAnd = Pattern::mergeVertical(p2, p1, l1, AND);
                    insertIfGain(cands, p2, p1, newPatAnd, sts, stsBase, numPat, stMap);
                }
                if (p2->root->op != XOR)
                {
                    const auto newPatXor = Pattern::mergeVertical(p2, p1, l1, XOR);
                    insertIfGain(cands, p2, p1, newPatXor, sts, stsBase, numPat, stMap);
                }

            } else if (p2->root->op != p1->root->op && p1->depth < 2)
            {
                const auto newPatJoin = Pattern::mergeHorizontal(p2, p1, l1);
                insertIfGain(cands, p2, p1, newPatJoin, sts, stsBase, numPat, stMap);

            }
            // we cannot join vertically, since p2 would have to be a singleton for this -> other case

        } else if (p2->root->children[l1]->op == p1->root->op && p1->depth < 2) {

            const auto newPatJoin = Pattern::mergeHorizontal(p2, p1, l1);
            insertIfGain(cands, p2, p1, newPatJoin, sts, stsBase, numPat, stMap);

        } else if (p1->root->op == NONE) {

            const auto newPatJoin = Pattern::mergeHorizontal(p2, p1, l1);
            insertIfGain(cands, p2, p1, newPatJoin, sts, stsBase, numPat, stMap);
        }
    }

    return cands;
}



// std::list<Pattern::Candidate> Pattern::genSTCandidates(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2, const std::vector<std::shared_ptr<Pattern> >& sts, const std::vector<std::shared_ptr<Pattern> >& stsBase, const uint32_t numPat, const std::unordered_map<uint32_t, std::list<std::shared_ptr<Pattern> > >& stMap)
// {
//     std::list<Pattern::Candidate> cands;
//     const auto newPatXor = Pattern::mergeNewOpRoot(p1, p2, XOR);
//     insertIfGain(cands, p1, p2, newPatXor, sts, stsBase, numPat, stMap);
//     const auto newPatAnd = Pattern::mergeNewOpRoot(p1, p2, AND);
//     insertIfGain(cands, p1, p2, newPatAnd, sts, stsBase, numPat, stMap);

//     return cands;
// }



std::shared_ptr<Pattern> Pattern::copy() const
{
    std::shared_ptr<Node> r = this->root->copy();
    return std::make_shared<Pattern>(std::move(r));
}



void Pattern::insertIfGain(std::list<Candidate>& cands, const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2, const std::shared_ptr<Pattern>& pNew, const std::vector<std::shared_ptr<Pattern> >& sts, const std::vector<std::shared_ptr<Pattern> >& stsBase, const uint32_t numPat, const std::unordered_map<uint32_t, std::list<std::shared_ptr<Pattern> > >& stMap)
{
    // No new pattern to begin with
    if (pNew == NULL)
    {
        return;
    }
    // check if there is any overlap
    if (pNew->root->tids->isEmpty())
    {
        return;
    }

    double gain = MDLAlgo::dataCLPattern(pNew->root, true) +
                  MDLAlgo::modelCLPattern(pNew->root, true) +
                  MDLAlgo::detail::logBinom(Pattern::ncol, pNew->root->items->getPopCount());

    // TODO
    // std::cout << "Data costs new: " << MDLAlgo::dataCLPattern(pNew->root, true) << "\n";
    // std::cout << "Model costs new: " << MDLAlgo::modelCLPattern(pNew->root, true) + MDLAlgo::detail::logBinom(Pattern::ncol, pNew->root->items->getPopCount()) << "\n";

    // TODO
    // Only deduct if not buffered
    if (p1->depth != 0 || p2->depth != 0)
    {
        gain -= MDLAlgo::detail::univCodeIntegerL(numPat);
    }

    // prepare map for later lookup of leafs in new pattern
    std::unordered_map<uint32_t, std::tuple<uint32_t, std::shared_ptr<dbitvec> > > mapNewPat;
    // For each pattern, we need to deduct the tids of this pattern from the singletons
    // initialize map for this
    for (const auto& ch : pNew->root->children)
    {
        if (ch->op == NONE)
        {
            mapNewPat[ch->items->firstBitId()] = {1, NULL};
        } else {
            for (const auto& chch : ch->children)
            {
                mapNewPat[chch->items->firstBitId()] = {2, ch->tids};
            }
        }
    }

    CandType ct = PATPAT;
    // check for singleton cases
    if (p1->depth == 0)
    {
        ct = PATST;
        const uint32_t stId = p1->root->items->firstBitId();
        const auto& st = sts[stId];
        // tids covered by new pattern
        uint32_t newCount;
        if (std::get<0>(mapNewPat[stId]) == 1)
        {
            newCount = st->root->tids->intersectionCount(*pNew->root->tids);
        } else {
            newCount = st->root->tids->intersectionCountMany({pNew->root->tids.get(), std::get<1>(mapNewPat[stId]).get()});
        }
        // deduct old singleton data costs
        gain -= MDLAlgo::detail::logBinom(Pattern::nrow, st->root->tids->getPopCount());
        if (newCount > 0 && newCount == st->root->tids->getPopCount())
        {
            // "delete" singleton
            gain += -1 - MDLAlgo::detail::paramCompApprox(Pattern::nrow, 2);
            gain += -MDLAlgo::detail::logBinom(Pattern::ncol, 1);
            // std::cout << "Reduced model costs by " << 1 + MDLAlgo::detail::paramCompApprox(Pattern::nrow, 2) + MDLAlgo::detail::logBinom(Pattern::ncol, 1) << "\n";
        } else {

            // new singleton data cost
            gain += MDLAlgo::detail::logBinom(Pattern::nrow, st->root->tids->getPopCount() - newCount);
        }
        // std::cout << "Reduced data costs by " << MDLAlgo::detail::logBinom(Pattern::nrow, st->root->tids->getPopCount()) - MDLAlgo::detail::logBinom(Pattern::nrow, st->root->tids->getPopCount() - newCount) << "\n";

    } else {

        gain += -MDLAlgo::dataCLPattern(p1->root, true) -
                MDLAlgo::modelCLPattern(p1->root, true) -
                MDLAlgo::detail::logBinom(Pattern::ncol, p1->root->items->getPopCount());
        // std::cout << "Reduced model costs by " << MDLAlgo::modelCLPattern(p1->root, true) + MDLAlgo::detail::logBinom(Pattern::ncol, p1->root->items->getPopCount()) << "\n";
        // std::cout << "Reduced data costs by " << MDLAlgo::dataCLPattern(p1->root, true)  << "\n";
        // std::unique_ptr<dbitvec> uncov (p1->root->tids->setdiff(*pNew->root->tids));
        // for each item in pattern, increase st costs
        // if (uncov->getPopCount() > 0)
        // {
        //     // const auto items = p1->root->items->getOnes();
        //     const auto itemIts = p1->root->items->getIts();
        //     // for (uint32_t i : *items)
        //     for (auto it = itemIts.first; it != itemIts.second; ++it)
        //     {
        //         const uint32_t iUncovCount = uncov->intersectionCount(*stsBase[*it]->root->tids);
        //         if (iUncovCount == 0)
        //             continue;
        //         const uint32_t oldStCover = sts[*it]->root->tids->getPopCount();
        //         gain += MDLAlgo::detail::logBinom(Pattern::nrow, iUncovCount + oldStCover);
        //         // in case st was not used before, add model costs
        //         if (oldStCover == 0)
        //         {
        //             gain += 1 + MDLAlgo::detail::paramCompApprox(Pattern::nrow, 2);
        //             std::cout << "Increased model costs by " << 1 + MDLAlgo::detail::paramCompApprox(Pattern::nrow, 2) << "\n";

        //         } else {

        //             gain -= MDLAlgo::detail::logBinom(Pattern::nrow, oldStCover);
        //         }
        //         std::cout << "Reduced data costs by " << MDLAlgo::detail::logBinom(Pattern::nrow, oldStCover) - MDLAlgo::detail::logBinom(Pattern::nrow, iUncovCount + oldStCover) << "\n";
        //     }
        // }
        for (const auto& ch : p1->root->children)
        {
            if (ch->op == NONE)
            {
                const uint32_t stId = ch->items->firstBitId();
                gain += updateGainByST(stId, p1, pNew, sts, stsBase, numPat, stMap, mapNewPat);


            } else {

                for (const auto& chch : ch->children)
                {
                    const uint32_t stId = chch->items->firstBitId();
                    gain += updateGainByST(stId, p1, pNew, sts, stsBase, numPat, stMap, mapNewPat);
                }
            }
        }
    }
    // check for singleton cases
    if (p2->depth == 0)
    {
        if (ct == PATST)
        {
            ct = STST;
            // TODO
            // We do not yet add this to gain because of buffering
            // gain += MDLAlgo::detail::univCodeIntegerL(numPat + 1);
        } else {
            ct = PATST;
            gain += MDLAlgo::detail::univCodeIntegerL(numPat);
        }
        const uint32_t stId = p2->root->items->firstBitId();
        const auto& st = sts[stId];
        // tids covered by new pattern
        uint32_t newCount;
        if (std::get<0>(mapNewPat[stId]) == 1)
        {
            newCount = st->root->tids->intersectionCount(*pNew->root->tids);
        } else {
            newCount = st->root->tids->intersectionCountMany({pNew->root->tids.get(), std::get<1>(mapNewPat[stId]).get()});
        }
        // deduct old singleton data costs
        gain -= MDLAlgo::detail::logBinom(Pattern::nrow, st->root->tids->getPopCount());
        if (newCount > 0 && newCount == st->root->tids->getPopCount())
        {
            // "delete" singleton
            gain += -1 - MDLAlgo::detail::paramCompApprox(Pattern::nrow, 2);
            gain += -MDLAlgo::detail::logBinom(Pattern::ncol, 1);
            // std::cout << "Reduced model costs by " << 1 + MDLAlgo::detail::paramCompApprox(Pattern::nrow, 2) + MDLAlgo::detail::logBinom(Pattern::ncol, 1)  << "\n";
        } else {

            // new singleton data cost
            gain += MDLAlgo::detail::logBinom(Pattern::nrow, st->root->tids->getPopCount() - newCount);
        }
        // std::cout << "Reduced data costs by " << MDLAlgo::detail::logBinom(Pattern::nrow, st->root->tids->getPopCount()) - MDLAlgo::detail::logBinom(Pattern::nrow, st->root->tids->getPopCount() - newCount) << "\n";

    } else {

        if (ct == PATPAT)
        {
            gain += MDLAlgo::detail::univCodeIntegerL(numPat - 1);
        } else {

            gain += MDLAlgo::detail::univCodeIntegerL(numPat);
        }

        gain += -MDLAlgo::dataCLPattern(p2->root, true) -
                MDLAlgo::modelCLPattern(p2->root, true) -
                MDLAlgo::detail::logBinom(Pattern::ncol, p2->root->items->getPopCount());

        for (const auto& ch : p2->root->children)
        {
            if (ch->op == NONE)
            {
                const uint32_t stId = ch->items->firstBitId();
                gain += updateGainByST(stId, p2, pNew, sts, stsBase, numPat, stMap, mapNewPat);

            } else {

                for (const auto& chch : ch->children)
                {
                    const uint32_t stId = chch->items->firstBitId();
                    gain += updateGainByST(stId, p2, pNew, sts, stsBase, numPat, stMap, mapNewPat);
                }
            }
        }
    }

    // std::cout << "\n";
    // p1->print(std::cout);
    // std::cout <<  " ~ ";
    // p2->print(std::cout);
    // std::cout << " / tids ";
    // // p1->root->tids->print(std::cout);
    // std::cout << p1->root->tids->getPopCount() << "\n";
    // std::cout << "  ~  ";
    // // p2->root->tids->print(std::cout);
    // std::cout << p2->root->tids->getPopCount() << "\n";
    // std::cout << "\n ---> ";
    // pNew->print(std::cout);
    // std::cout << " / tid ";
    // // pNew->root->tids->print(std::cout);
    // std::cout << pNew->root->tids->getPopCount() << "\n";
    // if (pNew->root->op == XOR)
    // {
    //     std::cout << "  ||  ";
    //     // pNew->root->tidsFiltered->print(std::cout);
    //     std::cout << pNew->root->tidsFiltered->getPopCount() << "\n";
    // }
    // std::cout << " / " << gain << "\n\n";

    // if (pNew->root->op == XOR)
    // {
    //     std::cout << gain << "\n\n";
    // }

    if (gain < 0)
    {
        const Candidate newCand = {ct, gain, pNew, p1, p2};
        // auto candIt = std::upper_bound(
        //     cands.begin(),
        //     cands.end(),
        //     newCand,
        //     CandComparator
        // );
        // cands.insert(candIt, newCand);

        // TODO: right now just best one as there is no gain estimate involved
        if (cands.size() != 0)
        {
            cands.pop_front();
        }
        cands.push_back(newCand);
    }
}


double Pattern::updateGainByST(const uint32_t stId, const std::shared_ptr<Pattern>& pOld, const std::shared_ptr<Pattern>& pNew, const std::vector<std::shared_ptr<Pattern> >& sts, const std::vector<std::shared_ptr<Pattern> >& stsBase, const uint32_t numPat, const std::unordered_map<uint32_t, std::list<std::shared_ptr<Pattern> > >& stMap, std::unordered_map<uint32_t, std::tuple<uint32_t, std::shared_ptr<dbitvec> > > mapNewPat)
{
    const auto stMapEntry = stMap.find(stId);
    dbitvec* newStCov;
    if (std::get<0>(mapNewPat[stId]) == 1)
    {
        std::unique_ptr<dbitvec> newPatCov (pNew->root->tids->setinter(*stsBase[stId]->root->tids));
        newStCov = stsBase[stId]->root->tids->setdiff(*newPatCov);
    } else {
        std::unique_ptr<dbitvec> newPatCov (pNew->root->tids->setinter(*stsBase[stId]->root->tids));
        newPatCov->thisInter(*std::get<1>(mapNewPat[stId]));
        newStCov = stsBase[stId]->root->tids->setdiff(*newPatCov);
    }
    // TODO: make this more efficient by not checking every node...
    // Loop over all patterns that contain this item and remove transactions of this pattern from singleton transaction list
    for (const auto& p : stMapEntry->second)
    {
        if (p == pOld)
        {
            continue;
        }
        for (const auto& ch : p->root->children)
        {
            if (ch->op == NONE)
            {
                if (ch->items->firstBitId() == stId)
                {
                    // Note that we need to deduce where PATTERN holds, not where the leaf holds!
                    newStCov->diffArg(*p->root->tids);
                    break;
                }
            } else {
                bool isFound = false;
                for (const auto& chch : ch->children)
                {
                    if (chch->items->firstBitId() == stId)
                    {
                        std::unique_ptr<dbitvec> redTids (p->root->tids->setinter(*ch->tids));
                        redTids->thisInter(*chch->tids);
                        newStCov->diffArg(*redTids);
                        isFound = true;
                        break;
                    }
                }
                if (isFound)
                    break;
            }
        }
    }
    double res = MDLAlgo::detail::logBinom(Pattern::nrow, newStCov->getPopCount());

    const uint32_t oldStCount = sts[stId]->root->tids->getPopCount();

    // std::cout << "Reduced data costs by " << MDLAlgo::detail::logBinom(Pattern::nrow, oldStCover) - MDLAlgo::detail::logBinom(Pattern::nrow, oldStCover + (covOld - overlap) - (covNew - overlap)) << "\n";
    if (oldStCount == 0 && newStCov->getPopCount() > 0)
    {
        // parametric complexity for model
        res += MDLAlgo::detail::paramCompApprox(Pattern::nrow, 2);
        // other model costs (which item is this singleton and saying that this is leaf node)
        res += MDLAlgo::detail::logBinom(Pattern::ncol, 1) + 1;
        // std::cout << "Increased model costs by " << 1 + MDLAlgo::detail::paramCompApprox(Pattern::nrow, 2) << "\n";

    } else {

        res -= MDLAlgo::detail::logBinom(Pattern::nrow, oldStCount);
    }
    delete newStCov;
    return res;
}



void Pattern::print(std::ostream& os, const std::vector<uint32_t>& idMap)
{
    this->root->print(os, idMap);
}
void Pattern::print(std::ostream& os)
{
    this->root->print(os);
}