#ifndef PATTERN_H
#define PATTERN_H

#include <vector>
#include <list>
#include <memory>
#include <fstream>

#include "densebitvec.h"
#include "sparsebitvec.h"
#include "codestats.h"

enum Operator {NONE, AND, XOR};

struct Node
{
    Operator op;
    std::shared_ptr<SparseBitvec> items;
    std::vector<std::shared_ptr<Node> > children;
    // tid list of where this subtree holds
    std::shared_ptr<dbitvec> tids;
    // In case this is an XOR node, a list of tids that have already been discarded because of XOR
    std::shared_ptr<dbitvec> tidsFiltered;

    Node(Operator opO, std::shared_ptr<SparseBitvec> it, std::vector<std::shared_ptr<Node> > ch, std::shared_ptr<dbitvec> ts, std::shared_ptr<dbitvec> tsf) :
            op(opO)
        ,   items(it)
        ,   children(ch)
        ,   tids(ts)
        ,   tidsFiltered(tsf)
    {
    }

    std::shared_ptr<Node> copy()
    {
        std::vector<std::shared_ptr<Node> > chs;
        for (const auto & ch : children)
        {
            chs.push_back(ch->copy());
        }
        std::shared_ptr<dbitvec> tidF = NULL;
        if (tidsFiltered != NULL)
        {
            tidF = std::shared_ptr<dbitvec>(tidsFiltered->copy());
        }
        return std::make_shared<Node>(op,
            std::shared_ptr<SparseBitvec>(items->copy()),
            std::move(chs),
            std::shared_ptr<dbitvec>(tids->copy()),
            tidF
        );
    }


    void print(std::ostream& os, const std::vector<uint32_t>& idMap)
    {
        switch(this->op)
        {
            case NONE:
            {
                auto ones = this->items->getOnes();
                os << idMap[(*ones)[0]];
                delete ones;
                break;
            }
            case AND:
                os << "AND (   ";
                for (const auto& n : this->children)
                {
                    n->print(os, idMap);
                    os << " , ";
                }
                os << "   )";
                break;
            case XOR:
                os << "XOR [   ";
                for (const auto& n : this->children)
                {
                    n->print(os, idMap);
                    os << " , ";
                }
                os << "   ]";
                break;
            default:
                std::cerr << "Undefined operator.\n";
                break;
        }
        os << "<" << this->tids->getPopCount() << ">";
    }
    void print(std::ostream& os)
    {
        switch(this->op)
        {
            case NONE:
            {
                auto ones = this->items->getOnes();
                os << (*ones)[0];
                delete ones;
                break;
            }
            case AND:
                os << "AND (   ";
                for (const auto& n : this->children)
                {
                    n->print(os);
                    os << " , ";
                }
                os << "   )";
                break;
            case XOR:
                os << "XOR [   ";
                for (const auto& n : this->children)
                {
                    n->print(os);
                    os << " , ";
                }
                os << "   ]";
                break;
            default:
                std::cerr << "Undefined operator.\n";
                break;
        }
        os << "<" << this->tids->getPopCount() << ">";
    }
};

// FWD declaration
namespace MDLAlgo
{
    double dataCLPattern(const std::shared_ptr<Node>& n, const bool isRoot);
    double modelCLPattern(const std::shared_ptr<Node>& n, const bool isRoot);
}

class Pattern
{
public:

    // Position where to merge two patterns
    // Traverse tree using child indices ch1 and ch2 for first and second level of the tree.
    // If ch1 is -1, merge into root,
    // if ch2 is -1, merge into child ch1 of root
    // otherwise merge into child ch2 of child ch1 of root.
    // struct TreePos
    // {
    //     const int ch1;
    //     const int ch2;
    // };

    Pattern();
    Pattern(unsigned int rootItem, std::shared_ptr<dbitvec> tidL);
    explicit Pattern(std::shared_ptr<Node>&& r);

    // merge two pattern under a new operator (vertical merge under new operator)
    static std::shared_ptr<Pattern> mergeNewOpRoot(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2, Operator op);
    // horizontal merge of two patterns, merging p2 into given position of p1
    static std::shared_ptr<Pattern> mergeHorizontal(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2, const int pos);
    // vertical merge of two patterns, replacing child ch1 of p1 at given position
    // with an operator node having ch1 and p2 as children.
    static std::shared_ptr<Pattern> mergeVertical(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2, const int pos, Operator op);
    // merge one pattern below root of other pattern, inserting a new child under root of p1 which is equal to p2
    static std::shared_ptr<Pattern> mergeBelow(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2);


    struct CandidateGen
    {
        const double estimate;
        const int pos;
    };
    // estimate the potential impact of merging given pattern with another pattern at any position
    // return ordered list of Candidates that yield some gain
    static std::list<Pattern::CandidateGen> estimateMergeWithOther(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2);
    enum CandType {PATPAT, PATST, STST};
    struct Candidate
    {
        const CandType type;
        const double gain;
        const std::shared_ptr<Pattern> newPat;
        const std::shared_ptr<Pattern>& pat1;
        const std::shared_ptr<Pattern>& pat2;
    };
    static struct CandComp
    {
        bool operator() (const Candidate& c1, const Candidate& c2)
        {
            return c1.gain < c2.gain;
        }

    } CandComparator;
    // candidate generation process
    static std::list<Pattern::Candidate> genCandidates(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2, const std::vector<std::shared_ptr<Pattern> >& sts, const std::vector<std::shared_ptr<Pattern> >& stsBase, const uint32_t numPat, const std::unordered_map<uint32_t, std::list<std::shared_ptr<Pattern> > >& stMap);
    // generate from singletons
    // static std::list<Pattern::Candidate> genSTCandidates(const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2, const std::vector<std::shared_ptr<Pattern> >& sts, const std::vector<std::shared_ptr<Pattern> >& stsBase, const uint32_t numPat, const std::unordered_map<uint32_t, std::list<std::shared_ptr<Pattern> > >& stMap);



    inline const std::shared_ptr<Node> getRoot() const
    {
        return this->root;
    }

    inline bool isST() const
    {
        return this->depth == 0;
    }

    void print(std::ostream& os, const std::vector<uint32_t>& idMap);
    void print(std::ostream& os);

    // TODO: init!!
    static uint64_t nrow;
    static uint64_t ncol;
    static std::shared_ptr<dbitvec> fullData;

private:

    std::shared_ptr<Pattern> copy() const;

    static inline void insertIfGain(std::list<Candidate>& cands, const std::shared_ptr<Pattern>& p1, const std::shared_ptr<Pattern>& p2, const std::shared_ptr<Pattern>& pNew, const std::vector<std::shared_ptr<Pattern> >& sts, const std::vector<std::shared_ptr<Pattern> >& stsBase, const uint32_t numPat, const std::unordered_map<uint32_t, std::list<std::shared_ptr<Pattern> > >& stMap);

    static inline double updateGainByST(const uint32_t stId, const std::shared_ptr<Pattern>& pOld, const std::shared_ptr<Pattern>& pNew, const std::vector<std::shared_ptr<Pattern> >& sts, const std::vector<std::shared_ptr<Pattern> >& stsBase, const uint32_t numPat, const std::unordered_map<uint32_t, std::list<std::shared_ptr<Pattern> > >& stMap, std::unordered_map<uint32_t, std::tuple<uint32_t, std::shared_ptr<dbitvec> > > mapNewPat);

    static unsigned int computeDepth(std::shared_ptr<Node>& p);

    std::shared_ptr<Node> root;

    unsigned int depth;
};


#endif // PATTERN_H