
#include <algorithm>

#include "patternset.h"
#include "mdlalgo.h"
#include "const.h"


PatternSet::PatternSet(std::vector<dbitvec*>::const_iterator begin, std::vector<dbitvec*>::const_iterator end, const std::vector<uint32_t>& idM) :
        idMap(idM)
    ,   baselineCL(0)
    ,   currentCL(0)
{
    uint32_t id = 0;
    sts.reserve(std::distance(begin, end));

    std::cout << "Initializing singletons...\n\n";
    for (auto it = begin; it != end; ++it, ++id)
    {
        // generate singletons using simple copies of the database columns
        std::shared_ptr<Pattern> newSt = std::make_shared<Pattern>(id, std::shared_ptr<dbitvec>((*it)->copy()));
        sts.push_back(newSt);
        std::shared_ptr<Pattern> newStBase = std::make_shared<Pattern>(id, std::shared_ptr<dbitvec>((*it)->copy()));
        stsBase.push_back(newStBase);


        std::cout << "Singleton " << id << ":\n" << newSt->getRoot()->tids->getPopCount() << "\t\t";
        newSt->print(std::cout, this->idMap);
        std::cout << "\n";
        stMap[id] = {};
    }
    stsBuf.reserve(sts.size() - 1);
    for (uint32_t i = 0; i < sts.size() - 1; ++i)
    {
        stsBuf.push_back(std::vector<stBuffer>(sts.size() - i - 1, {true, false, NULL}));
    }
    this->computeCL();
    this->baselineCL = this->currentCL;
    std::cout << "\nBaseline codelength: " << this->baselineCL << "\n\n";
}



void PatternSet::commitCandidate(Pattern::Candidate& cand)
{

    // this->removeFromSingletonTids(cand.newPat);
    // if (!cand.pat1->isST())
    // {
    //     this->updateSingletonTids(cand.pat1, cand.newPat);
    // }
    // if (!cand.pat2->isST())
    // {
    //     this->updateSingletonTids(cand.pat2, cand.newPat);
    // }


    // insert new pattern
    auto candIt = std::upper_bound(
        patterns.begin(),
        patterns.end(),
        cand.newPat,
        PatternComparator
    );
    patterns.insert(candIt, cand.newPat);

    const auto itPair = cand.newPat->getRoot()->items->getIts();
    // delete singleton references to old patterns
    for (auto stIt = itPair.first; stIt != itPair.second; ++stIt)
    {
        std::list<std::shared_ptr<Pattern> > refPats = this->stMap[*stIt];
        for (auto pIt = refPats.begin(); pIt != refPats.end();)
        {
            if (*pIt == cand.pat1 || *pIt == cand.pat2)
            {
                pIt = refPats.erase(pIt);
            } else {
                ++pIt;
            }
        }
        refPats.push_back(cand.newPat);
    }

    // delete old patterns
    bool cand1Er = false;
    bool cand2Er = false;
    for (std::list<std::shared_ptr<Pattern> >::iterator it = patterns.begin(); it != patterns.end(); )
    {
        if (!cand1Er && !cand.pat1->isST() && *it == cand.pat1)
        {
            cand1Er = true;
            it = patterns.erase(it);
        } else if (!cand2Er && !cand.pat2->isST() && *it == cand.pat2)
        {
            cand2Er = true;
            it = patterns.erase(it);
        } else {
            ++it;
        }
    }

    // reset singletons that were affected by new pattern to base encoding
    // and reset buffer
    for (auto stIt = itPair.first; stIt != itPair.second; ++stIt)
    {
        sts[*stIt]->getRoot()->tids->copyContent(stsBase[*stIt]->getRoot()->tids.get());
        for (uint32_t i = 0; i < *stIt; ++i)
        {
            stsBuf[i][*stIt -i-1].recompute = true;
        }
        for (uint32_t j = 0; j < Pattern::ncol - *stIt - 1; ++j)
        {
            stsBuf[*stIt][j].recompute = true;
        }
    }

    // Go over all patterns
    for (const auto& p : patterns)
    {
        // For each pattern, deduct the tids from the singletons
        for (const auto& ch : p->getRoot()->children)
        {
            if (ch->op == NONE)
            {
                // deduct
                sts[ch->items->firstBitId()]->getRoot()->tids->diffArg(*p->getRoot()->tids);

            } else {
                // recurse
                for (const auto& chch : ch->children)
                {
                    std::unique_ptr<dbitvec> redTids (p->getRoot()->tids->setinter(*ch->tids));
                    redTids->thisInter(*chch->tids);
                    // deduct
                    sts[chch->items->firstBitId()]->getRoot()->tids->diffArg(*redTids);
                }
            }
        }
    }
}



std::list<Pattern::Candidate> PatternSet::genCandidates()
{
    std::list<Pattern::Candidate> cands;
    double bestGain = 1;
    // generate candidates by singleton merges

    // costs for current number of patterns + 1 (the new pattern added through singletons)
    const auto pNumCostsNew = MDLAlgo::detail::univCodeIntegerL(patterns.size() + 1);
    const auto pNumCostsOld = MDLAlgo::detail::univCodeIntegerL(patterns.size());

#pragma omp parallel for num_threads(CORENUM)
    for (uint32_t i = 0; i < stsBase.size(); ++i)
    {
        const auto& st1 = stsBase[i];
        for (uint32_t j = i + 1; j < stsBase.size(); ++j)
        {
            // const auto& st2 = stsBase[j];
            // const auto mergedCands = Pattern::genCandidates(st1, st2, sts, stsBase, patterns.size(), stMap);
            // Recompute value for this pair
            if (stsBuf[i][j-(i+1)].recompute)
            {
                stsBuf[i][j-(i+1)].recompute = false;
                const auto& st2 = stsBase[j];
                const auto mergedCands = Pattern::genCandidates(st1, st2, sts, stsBase, patterns.size(), stMap);
                // if there is no good candidate, store it as no gain pattern
                if (mergedCands.empty())
                {

                    stsBuf[i][j-(i+1)].noGain = true;

                // update candidate list and buffer with new candidate accordingly
                } else {
                    stsBuf[i][j-(i+1)].noGain = false;
                    stsBuf[i][j-(i+1)].bestCand = std::make_shared<Pattern::Candidate>(mergedCands.front());

#pragma omp critical
{
                    if (mergedCands.front().gain + pNumCostsNew - pNumCostsOld < bestGain)
                    {
                        if (bestGain < 0)
                        {
                            cands.pop_front();
                        }
                        cands.push_back({mergedCands.front().type,
                            mergedCands.front().gain + pNumCostsNew - pNumCostsOld,
                            mergedCands.front().newPat,
                            mergedCands.front().pat1,
                            mergedCands.front().pat2
                            });
                        bestGain = cands.front().gain;
                    }
}
                }

            } else if (stsBuf[i][j-(i+1)].noGain)
            {
                // No candidates with gain for this pair
                continue;

            } else {
                // look up value and return
#pragma omp critical
{
                if (stsBuf[i][j-(i+1)].bestCand->gain + pNumCostsNew - pNumCostsOld < bestGain)
                {
                    if (bestGain < 0)
                    {
                        cands.pop_front();
                    }
                    const auto& nC = *(stsBuf[i][j-(i+1)].bestCand);
                    cands.push_back({nC.type,
                        nC.gain + pNumCostsNew - pNumCostsOld,
                        nC.newPat,
                        nC.pat1,
                        nC.pat2
                        });
                    bestGain = cands.front().gain;
                }
}

            }
        }
    }

    // generate candidates by merging existing patterns
#pragma omp parallel for num_threads(CORENUM)
    for (uint32_t i = 0; i < patterns.size(); ++i)
    {
        auto p1It = patterns.begin();
        std::advance(p1It, i);
        auto p2It = p1It;
        ++p2It;
        for (; p2It != patterns.end(); ++p2It)
        {
            const auto mergedCands = Pattern::genCandidates(*p1It, *p2It, sts, stsBase, patterns.size(), stMap);
            // TODO
            // TODO
            // TODO: Currently only holds best candidate
            if (mergedCands.size() != 0)
            {
#pragma omp critical
{
                if (mergedCands.front().gain < bestGain)
                {
                    if (bestGain < 0)
                    {
                        cands.pop_front();
                    }
                    cands.push_back(mergedCands.front());
                    bestGain = mergedCands.front().gain;
                }
}
            }
        }
        for (const auto& st : stsBase)
        {
            const auto mergedCands = Pattern::genCandidates(*p1It, st, sts, stsBase, patterns.size(), stMap);
            // TODO
            // TODO
            // TODO: Currently only holds best candidate
            if (mergedCands.size() != 0)
            {
#pragma omp critical
{
                if (mergedCands.front().gain < bestGain)
                {
                    if (bestGain < 0)
                    {
                        cands.pop_front();
                    }
                    cands.push_back(mergedCands.front());
                    bestGain = mergedCands.front().gain;
                }
}
            }
        }
    }

    return cands;
}



void PatternSet::printToFile(std::ofstream& ofs) const
{
    for (const auto& p : patterns)
    {
        ofs << p->getRoot()->tids->getPopCount() << "\t\t";
        p->print(ofs, this->idMap);
        ofs << "\n\n";
    }
}



// TODO:
// TODO:
// TODO:
// Assumption that nobody else covers this item!
void PatternSet::updateSingletonTids(const std::shared_ptr<Pattern>& oldPat, const std::shared_ptr<Pattern>& newPat)
{
    // update singleton usages
    const std::unique_ptr<dbitvec> uncov (oldPat->getRoot()->tids->setdiff(*newPat->getRoot()->tids));
    if (uncov->getPopCount() > 0)
    {
        // retrieve items in old pattern
        //const std::unique_ptr<const std::vector<uint32_t> > items (oldPat->getRoot()->items->getOnes());
        const auto itemIts = oldPat->getRoot()->items->getIts();
        //for (uint32_t i : *items)
        for (auto it = itemIts.first; it != itemIts.second; ++it)
        {
            // compute what is now uncovered
            const std::unique_ptr<dbitvec> iUncov (uncov->setinter(*stsBase[*it]->getRoot()->tids));
            if (iUncov->getPopCount() > 0)
            {
                // update singleton tids
                sts[*it]->getRoot()->tids->thisUnion(*iUncov);
            }
        }
    }
}



void PatternSet::removeFromSingletonTids(const std::shared_ptr<Pattern>& newPat)
{
    // retrieve items in new pattern
    const auto itemIts = newPat->getRoot()->items->getIts();
    //for (uint32_t i : *items)
    for (auto it = itemIts.first; it != itemIts.second; ++it)
    {
        // update singleton tids
        sts[*it]->getRoot()->tids->diffArg(*newPat->getRoot()->tids);
    }
}



double PatternSet::computeCL()
{
    double clAcc = 0;
    for (const auto& st : sts)
    {
        if (!st->getRoot()->tids->isEmpty())
            clAcc += MDLAlgo::dataCLPattern(st->getRoot(), true);
    }
    for (const auto& p : patterns)
    {
        clAcc += MDLAlgo::dataCLPattern(p->getRoot(), true);
    }
    this->currentCL = MDLAlgo::modelCL(sts) + MDLAlgo::modelCL(patterns) + clAcc;
    return this->currentCL;
}