#ifndef PATTERNSET_H
#define PATTERNSET_H

#include <list>
#include <unordered_map>
#include <vector>
#include <memory>


#include "bitcollection.h"
#include "pattern.h"

class PatternSet
{

public:

    // ------- Ctors -----------


    PatternSet() = delete;

    // Generate pattern set initialized with singletons corresponding to given columns
    PatternSet(std::vector<dbitvec*>::const_iterator begin, std::vector<dbitvec*>::const_iterator end, const std::vector<uint32_t>& idM);


    // -------------------------

    // Commit to given candidate, adding it to the correct position and deleting "stale" patterns
    void commitCandidate(Pattern::Candidate& cand);

    // Generate candidates
    std::list<Pattern::Candidate> genCandidates();

    // Print the pattern set to given file
    void printToFile(std::ofstream& ofs) const;

    // Compute the current codelength of the pattern set
    double computeCL();


private:

    void updateSingletonTids(const std::shared_ptr<Pattern>& oldPat, const std::shared_ptr<Pattern>& newPat);
    void removeFromSingletonTids(const std::shared_ptr<Pattern>& newPat);

    // Update the singletons by iterating over all contained patterns
    void updateSingletons();

    // Compare two patterns based on their coverage (size of tid sets)
    static struct PatternComp
    {
        bool operator() (const std::shared_ptr<Pattern>& newPat, const std::shared_ptr<Pattern>& exPat)
        {
            return newPat->getRoot()->tids->getPopCount() < exPat->getRoot()->tids->getPopCount();
        }
    } PatternComparator;
    // Set of patterns sorted by tid count
    std::list<std::shared_ptr<Pattern> > patterns;
    // Set of singletons sorted by item index
    std::vector<std::shared_ptr<Pattern> > sts;
    // Set of singletons with baseline tids sorted by item index
    std::vector<std::shared_ptr<Pattern> > stsBase;

    struct stBuffer
    {
        bool recompute;
        bool noGain;
        // The best out of the XOR and AND case
        std::shared_ptr<Pattern::Candidate> bestCand;
    };
    // Buffer to avoid recomputing singleton patterns
    std::vector<std::vector<stBuffer> > stsBuf;

    // Map to patterns where singleton is present
    // singleton id -> pattern
    std::unordered_map<uint32_t, std::list<std::shared_ptr<Pattern> > > stMap;

    const std::vector<uint32_t>& idMap;

    double baselineCL;
    double currentCL;
};


#endif // PATTERNSET_H