# -*- coding: utf-8 -*-

from typing import Dict

import tempfile
import os
import subprocess
import pathlib

import pandas as pd

from prolothar_common.models.eventlog import EventLog

HERE = pathlib.Path(__file__).parent

class Sqs():
    """python interface to the SQS pattern miner"""

    def find_patterns(self, log: EventLog):
        with tempfile.TemporaryDirectory() as temporary_directory:
            dict_activity_id = dict((activity,id) for id,activity
                                    in enumerate(log.compute_activity_set()))
            dict_id_activity = dict((v,k) for k, v in dict_activity_id.items())

            sqs_input_file_path = self.__write_sqs_input_file(
                    log, temporary_directory, dict_activity_id)

            sqs_output_file_path = self.__execute_sqs(sqs_input_file_path,
                                                      temporary_directory)

            return self.__parse_sqs_output_file(sqs_output_file_path,
                                                dict_id_activity)

    def __write_sqs_input_file(self, log: EventLog, temporary_directory: str,
                             dict_activity_id: Dict[str,int]) -> str:
        sqs_input_file_path = os.path.join(temporary_directory,
                                           'log_sqs_data.txt')
        write_sqs_input_file(log, sqs_input_file_path, dict_activity_id)
        return sqs_input_file_path

    def __execute_sqs(self, sqs_input_file_path: str,
                      temporary_directory: str) -> str:
        sqs_output_file_path = os.path.abspath(
                os.path.join(temporary_directory, 'sqs_output.txt'))
        sqs_process = subprocess.run([
                str(HERE / 'sqs' / 'sqs'),
                '-i', sqs_input_file_path,
                '-o', sqs_output_file_path,
                '-m', 'search'
        ], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
           universal_newlines=True)
        if sqs_process.returncode != 0:
            print('------------')
            print(sqs_process.stdout)
            print(sqs_process.stderr)
            print('------------')
            raise Exception('sqs returned exit code %d' %
                            sqs_process.returncode)
        return sqs_output_file_path

    def __parse_sqs_output_file(self, sqs_output_file_path: str,
                                dict_id_activity: Dict[int,str]):
        patterns = []
        with open(sqs_output_file_path, 'r') as result_file:
            for line in result_file:
                pattern = {'sequence': []}
                columns = line.split()
                for column in columns[:-3]:
                    pattern['sequence'].append(dict_id_activity[int(column)])
                pattern['???'] = float(columns[-3][1:])
                pattern['worth'] = float(columns[-2])
                pattern['usage'] = int(columns[-1][:-1])
                patterns.append(pattern)
        return pd.DataFrame(patterns)

def write_sqs_input_file(log: EventLog, filepath: str,
                         dict_activity_id: Dict[str,int],
                         symbol_separator: str = ' ',
                         end_of_trace: str = '-1 '):
    with open(os.path.abspath(filepath), 'w') as data_file:
        for trace in log.traces:
            for event in trace.events:
                data_file.write(str(dict_activity_id[event.activity_name]) +
                                symbol_separator)
            data_file.write(end_of_trace)
