# -*- coding: utf-8 -*-

from typing import Dict, List, Union, Tuple

import tempfile
import os
import subprocess
import pathlib

from prolothar_common.models.eventlog import EventLog
from prolothar_pattern_mining.thirdparty.sqs import write_sqs_input_file

Pattern = Union[str, Tuple[str], List[str], List[Tuple[str]]]

HERE = pathlib.Path(__file__).parent

class Squish():
    """python interface to the Squish pattern miner"""

    def __init__(self, verbose: bool = False):
        self.__verbose = verbose

    def find_patterns(self, log: EventLog) -> List[Pattern]:
        with tempfile.TemporaryDirectory() as temporary_directory:
            dict_activity_id = dict((activity,id) for id,activity
                                    in enumerate(log.compute_activity_set()))
            dict_id_activity = dict((str(v),k) for k, v in dict_activity_id.items())

            squish_input_file_path = self.__write_squish_input_file(
                    log, temporary_directory, dict_activity_id)

            squish_output = self.__execute_squish(squish_input_file_path,
                                                  temporary_directory)

            return self.__parse_squish_output(squish_output, dict_id_activity)

    def __write_squish_input_file(
            self, log: EventLog, temporary_directory: str,
            dict_activity_id: Dict[str,int]) -> str:
        squish_input_file_path = os.path.join(temporary_directory,
                                           'log_squish_data.txt')
        write_sqs_input_file(log, squish_input_file_path, dict_activity_id,
                             symbol_separator=',')
        return squish_input_file_path

    def __execute_squish(
            self, sqs_input_file_path: str, temporary_directory: str) -> str:
        sqs_process = subprocess.run([
                str(HERE / 'squish' / 'squish'),
                sqs_input_file_path,
        ], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
           universal_newlines=True)
        if sqs_process.returncode != 0:
            print('------------')
            print(sqs_process.stdout)
            print(sqs_process.stderr)
            print('------------')
            raise Exception('squish returned exit code %d' %
                            sqs_process.returncode)
        return sqs_process.stdout

    def __parse_squish_output(
            self, stdout: str,
            dict_id_activity: Dict[str,str]) -> List[Pattern]:
        if self.__verbose:
            print(stdout)
        patterns_started = False
        parsed_patterns = []
        for line in stdout.split('\n'):
            if patterns_started:
                if '-----' in line:
                    break
                if not line.startswith('codelen:'):
                   parsed_patterns.append(
                           self.__parse_pattern(line, dict_id_activity))
            else:
                patterns_started = 'All choice patterns' in line
        return parsed_patterns

    def __parse_pattern(self, line: str,
                        dict_id_activity: Dict[str,str]) -> Pattern:
        line = line.replace('{', '')
        line = line.replace('}', '')
        parsed_sequence_element_list = []
        for sequence_element in line.split(' - '):
            if sequence_element:
                parsed_sequence_element_list.append(
                        self.__parse_sequence_element(sequence_element,
                                                      dict_id_activity))
        if len(parsed_sequence_element_list) == 1:
            return parsed_sequence_element_list[0]
        else:
            return parsed_sequence_element_list

    def __parse_sequence_element(self, sequence_element: str,
                                 dict_id_activity: Dict[str,str]) -> Pattern:
        if '|' in sequence_element:
            return (
                dict_id_activity[a]
                for a in sequence_element.split(' | ') if a
            )
        else:
            return dict_id_activity[sequence_element]
