"""
This module contains methods to plot histograms of discrete or continuous values
"""
from typing import Iterable, Union
from collections import Counter
from math import sqrt

import matplotlib.pyplot as plt
from matplotlib.axes import Axes

def plot_counts(values: Iterable, ax: Axes = None, order_by: str = 'value'):
    """
    plots the frequency of discrete values

    Parameters
    ----------
    values : Iterable
        a container of discete values whose frequency will be plotted
    ax : Axes
        matplotlib axis to plot. must only be set if multiple plots are supposed
        to be combined in one plot by using subplots.
    order_by : str
        can be 'value', 'frequency_asc' or 'frequency_desc'. determines the order
        on the x axis
    """
    if not ax:
        ax = plt.gca()

    counter = Counter()
    for v in values:
        counter[v] += 1

    if order_by == 'value':
        x, y = zip(*sorted(counter.items()))
    elif order_by == 'frequency_asc':
        x, y = zip(*sorted(counter.items(), key=lambda x: x[1]))
    elif order_by == 'frequency_desc':
        x, y = zip(*sorted(counter.items(), key=lambda x: -x[1]))
    else:
        raise ValueError('unknown value for order_by: "%s"' % order_by)
    ax.bar(x, y, 1)

def plot_histogram(
        values: Iterable[Union[float]], nr_of_bins: int = None,
        ax: Axes = None):
    if not ax:
        ax = plt.gca()
    if not nr_of_bins:
        nr_of_bins = sqrt(len(values))

    ax.hist(values, bins=nr_of_bins)