"""
contains a class for "plot context", that means auto showable plots, such that
boilerplate code for plot creation is reduced.
"""

from io import BytesIO
import base64

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes

class PlotContext():
    """
    a class that facilitates plotting and reduced boilerplate code, e.g. by
    automatically calling plt.show() or savefig()
    """

    def __init__(self, title: str = None, show: bool = True,
                 filepath: str = None, use_tight_layout: bool = False):
        self.__title = title
        self.__show = show
        self.__filepath = filepath
        self.__figure: Figure = None
        self.__axes: Axes = None
        self.__use_tight_layout = use_tight_layout

    def get_figure(self) -> Figure:
        return self.__figure

    def get_axes(self) -> Axes:
        return self.__axes

    def __enter__(self):
        self.__figure, self.__axes = plt.subplots()
        self.__figure.suptitle(self.__title)
        return self

    def __exit__(self, type, value, traceback):
        if self.__use_tight_layout:
            plt.tight_layout()
        if self.__filepath is not None:
            self.__figure.savefig(self.__filepath)
        if self.__show:
            plt.show()
        else:
            plt.close()

    def to_base_64(self, image_format: str = 'png') -> str:
        """
        returns a base64 encoded version of the figure in this plot context
        """
        out_img = BytesIO()
        self.get_figure().savefig(out_img, format=image_format)
        out_img.seek(0)  # rewind file
        return base64.b64encode(out_img.read()).decode("ascii").replace("\n", "")
