# -*- coding: utf-8 -*-

import os
import pathlib
import pickle
import json
import pandas as pd

from prolothar_common.models.dataset import TargetSequenceDataset
from prolothar_common.models.dataset.instance import TargetSequenceInstance

HERE = pathlib.Path(__file__).parent

def _categorize_incidence_value(incidence: float) -> str:
    if incidence <= 0:
        category = '0'
    elif incidence < 35:
        category = '0-35'
    elif incidence < 50:
        category = '35-50'
    elif incidence < 100:
        category = '50-100'
    elif incidence < 165:
        category = '100-165'
    elif incidence < 200:
        category = '165-200'
    else:
        category = '200+'
    return category

def load_seven_day_incidence_county_dataset() -> TargetSequenceDataset:
    path_to_dataset = os.path.join(HERE, 'german_seven_day_incidence_county_dataset.pickle')

    if os.path.exists(path_to_dataset):
        with open(path_to_dataset, 'rb') as f:
            dataset = pickle.load(f)
    else:
        dataset = TargetSequenceDataset(
            ['is_district_town', 'state'],
            ['population', 'longitude', 'latitude']
        )

        incidences = pd.read_csv(str(HERE / '7di-rki-by-ags.csv'), sep=',')
        incidences.drop(['germany_7di'], inplace=True, axis=1)
        incidences.columns = pd.Series(incidences.columns.str.replace('_7di', ''))

        with open(HERE / 'german_counties.json') as f:
            counties = json.load(f)

        for county_id, county_data in counties.items():
            if 'population' in county_data:
                dataset.add_instance(TargetSequenceInstance(
                    county_data['name'],
                    {
                        'is_district_town': 'yes' if county_data['name'].startswith('SK ') else 'no',
                        'state': county_data['state'],
                        'population': county_data['population'],
                        'longitude': county_data['lon'],
                        'latitude': county_data['lat'],
                    },
                    [
                        _categorize_incidence_value(incidence) for incidence in incidences[county_id]
                    ]
                ))

        try:
            with open(path_to_dataset, 'wb') as f:
                pickle.dump(dataset, f)
        except PermissionError as e:
            print('No permission to write german 7 day incidence dataset to disk for caching:')
            print(e)

    return dataset
